"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const constants_1 = require("../../../lib/common-libs/constants");
const GenericParser_1 = require("./GenericParser");
const common_1 = require("../../../lib/common");
const index_1 = require("../../../lib/common-libs/index");
const BlockDTO_1 = require("../../dto/BlockDTO");
class BlockParser extends GenericParser_1.GenericParser {
    constructor() {
        super([
            { prop: "version", regexp: constants_1.CommonConstants.BLOCK.VERSION },
            { prop: "type", regexp: constants_1.CommonConstants.BLOCK.TYPE },
            { prop: "currency", regexp: constants_1.CommonConstants.BLOCK.CURRENCY },
            { prop: "number", regexp: constants_1.CommonConstants.BLOCK.BNUMBER },
            { prop: "powMin", regexp: constants_1.CommonConstants.BLOCK.POWMIN },
            { prop: "time", regexp: constants_1.CommonConstants.BLOCK.TIME },
            { prop: "medianTime", regexp: constants_1.CommonConstants.BLOCK.MEDIAN_TIME },
            { prop: "dividend", regexp: constants_1.CommonConstants.BLOCK.UD },
            { prop: "unitbase", regexp: constants_1.CommonConstants.BLOCK.UNIT_BASE },
            { prop: "issuer", regexp: constants_1.CommonConstants.BLOCK.BLOCK_ISSUER },
            { prop: "issuersFrame", regexp: constants_1.CommonConstants.BLOCK.BLOCK_ISSUERS_FRAME },
            { prop: "issuersFrameVar", regexp: constants_1.CommonConstants.BLOCK.BLOCK_ISSUERS_FRAME_VAR },
            { prop: "issuersCount", regexp: constants_1.CommonConstants.BLOCK.DIFFERENT_ISSUERS_COUNT },
            { prop: "parameters", regexp: constants_1.CommonConstants.BLOCK.PARAMETERS },
            { prop: "previousHash", regexp: constants_1.CommonConstants.BLOCK.PREV_HASH },
            { prop: "previousIssuer", regexp: constants_1.CommonConstants.BLOCK.PREV_ISSUER },
            { prop: "membersCount", regexp: constants_1.CommonConstants.BLOCK.MEMBERS_COUNT },
            { prop: "identities", regexp: /Identities:\n([\s\S]*)Joiners/, parser: splitAndMatch('\n', constants_1.CommonConstants.IDENTITY.INLINE) },
            { prop: "joiners", regexp: /Joiners:\n([\s\S]*)Actives/, parser: splitAndMatch('\n', constants_1.CommonConstants.BLOCK.JOINER) },
            { prop: "actives", regexp: /Actives:\n([\s\S]*)Leavers/, parser: splitAndMatch('\n', constants_1.CommonConstants.BLOCK.ACTIVE) },
            { prop: "leavers", regexp: /Leavers:\n([\s\S]*)Excluded/, parser: splitAndMatch('\n', constants_1.CommonConstants.BLOCK.LEAVER) },
            { prop: "revoked", regexp: /Revoked:\n([\s\S]*)Excluded/, parser: splitAndMatch('\n', constants_1.CommonConstants.BLOCK.REVOCATION) },
            { prop: "excluded", regexp: /Excluded:\n([\s\S]*)Certifications/, parser: splitAndMatch('\n', constants_1.CommonConstants.PUBLIC_KEY) },
            { prop: "certifications", regexp: /Certifications:\n([\s\S]*)Transactions/, parser: splitAndMatch('\n', constants_1.CommonConstants.CERT.OTHER.INLINE) },
            { prop: "transactions", regexp: /Transactions:\n([\s\S]*)/, parser: extractTransactions },
            { prop: "inner_hash", regexp: constants_1.CommonConstants.BLOCK.INNER_HASH },
            { prop: "nonce", regexp: constants_1.CommonConstants.BLOCK.NONCE }
        ], index_1.rawer.getBlock);
    }
    _clean(obj) {
        obj.identities = obj.identities || [];
        obj.joiners = obj.joiners || [];
        obj.actives = obj.actives || [];
        obj.leavers = obj.leavers || [];
        obj.revoked = obj.revoked || [];
        obj.excluded = obj.excluded || [];
        obj.certifications = obj.certifications || [];
        obj.transactions = obj.transactions || [];
        obj.version = obj.version || '';
        obj.type = obj.type || '';
        obj.hash = common_1.hashf(index_1.rawer.getBlockInnerHashAndNonceWithSignature(obj)).toUpperCase();
        obj.inner_hash = obj.inner_hash || '';
        obj.currency = obj.currency || '';
        obj.nonce = obj.nonce || '';
        obj.number = obj.number || '';
        obj.time = obj.time || '';
        obj.medianTime = obj.medianTime || '';
        obj.dividend = obj.dividend || null;
        obj.unitbase = obj.unitbase || '';
        obj.issuer = obj.issuer || '';
        obj.parameters = obj.parameters || '';
        obj.previousHash = obj.previousHash || '';
        obj.previousIssuer = obj.previousIssuer || '';
        obj.membersCount = obj.membersCount || '';
        obj.transactions.map((tx) => {
            tx.currency = obj.currency;
            tx.hash = common_1.hashf(index_1.rawer.getTransaction(tx)).toUpperCase();
        });
        obj.len = BlockDTO_1.BlockDTO.getLen(obj);
    }
    ;
    _verify(obj) {
        let err = null;
        const codes = {
            'BAD_VERSION': 150,
            'BAD_CURRENCY': 151,
            'BAD_NUMBER': 152,
            'BAD_TYPE': 153,
            'BAD_NONCE': 154,
            'BAD_RECIPIENT_OF_NONTRANSFERT': 155,
            'BAD_PREV_HASH_PRESENT': 156,
            'BAD_PREV_HASH_ABSENT': 157,
            'BAD_PREV_ISSUER_PRESENT': 158,
            'BAD_PREV_ISSUER_ABSENT': 159,
            'BAD_DIVIDEND': 160,
            'BAD_TIME': 161,
            'BAD_MEDIAN_TIME': 162,
            'BAD_INNER_HASH': 163,
            'BAD_MEMBERS_COUNT': 164,
            'BAD_UNITBASE': 165,
            'BAD_ISSUER': 166
        };
        if (!err) {
            // Version
            if (!obj.version || !obj.version.match(constants_1.CommonConstants.DOCUMENTS_BLOCK_VERSION_REGEXP))
                err = { code: codes.BAD_VERSION, message: "Version unknown" };
        }
        if (!err) {
            // Type
            if (!obj.type || !obj.type.match(/^Block$/))
                err = { code: codes.BAD_TYPE, message: "Not a Block type" };
        }
        if (!err) {
            // Nonce
            if (!obj.nonce || !obj.nonce.match(constants_1.CommonConstants.INTEGER))
                err = { code: codes.BAD_NONCE, message: "Nonce must be an integer value" };
        }
        if (!err) {
            // Number
            if (!obj.number || !obj.number.match(constants_1.CommonConstants.INTEGER))
                err = { code: codes.BAD_NUMBER, message: "Incorrect Number field" };
        }
        if (!err) {
            // Time
            if (!obj.time || !obj.time.match(constants_1.CommonConstants.INTEGER))
                err = { code: codes.BAD_TIME, message: "Time must be an integer" };
        }
        if (!err) {
            // MedianTime
            if (!obj.medianTime || !obj.medianTime.match(constants_1.CommonConstants.INTEGER))
                err = { code: codes.BAD_MEDIAN_TIME, message: "MedianTime must be an integer" };
        }
        if (!err) {
            if (obj.dividend && !obj.dividend.match(constants_1.CommonConstants.INTEGER))
                err = { code: codes.BAD_DIVIDEND, message: "Incorrect UniversalDividend field" };
        }
        if (!err) {
            if (obj.unitbase && !obj.unitbase.match(constants_1.CommonConstants.INTEGER))
                err = { code: codes.BAD_UNITBASE, message: "Incorrect UnitBase field" };
        }
        if (!err) {
            if (!obj.issuer || !obj.issuer.match(constants_1.CommonConstants.BASE58))
                err = { code: codes.BAD_ISSUER, message: "Incorrect Issuer field" };
        }
        if (!err) {
            // MembersCount
            if (!obj.nonce || !obj.nonce.match(constants_1.CommonConstants.INTEGER))
                err = { code: codes.BAD_MEMBERS_COUNT, message: "MembersCount must be an integer value" };
        }
        if (!err) {
            // InnerHash
            if (!obj.inner_hash || !obj.inner_hash.match(constants_1.CommonConstants.FINGERPRINT))
                err = { code: codes.BAD_INNER_HASH, message: "InnerHash must be a hash value" };
        }
        return err && err.message;
    }
    ;
}
exports.BlockParser = BlockParser;
function splitAndMatch(separator, regexp) {
    return function (raw) {
        const lines = raw.split(new RegExp(separator));
        const kept = [];
        lines.forEach(function (line) {
            if (line.match(regexp))
                kept.push(line);
        });
        return kept;
    };
}
function extractTransactions(raw) {
    const regexps = {
        "issuers": constants_1.CommonConstants.TRANSACTION.SENDER,
        "inputs": constants_1.CommonConstants.TRANSACTION.SOURCE_V3,
        "unlocks": constants_1.CommonConstants.TRANSACTION.UNLOCK,
        "outputs": constants_1.CommonConstants.TRANSACTION.TARGET,
        "comments": constants_1.CommonConstants.TRANSACTION.INLINE_COMMENT,
        "signatures": constants_1.CommonConstants.SIG
    };
    const transactions = [];
    const lines = raw.split(/\n/);
    for (let i = 0; i < lines.length; i++) {
        const line = lines[i];
        // On each header
        if (line.match(constants_1.CommonConstants.TRANSACTION.HEADER)) {
            // Parse the transaction
            const currentTX = { raw: line + '\n' };
            const sp = line.split(':');
            const version = parseInt(sp[1]);
            const nbIssuers = parseInt(sp[2]);
            const nbInputs = parseInt(sp[3]);
            const nbUnlocks = parseInt(sp[4]);
            const nbOutputs = parseInt(sp[5]);
            const hasComment = parseInt(sp[6]);
            const start = 2;
            currentTX.version = version;
            currentTX.blockstamp = lines[i + 1];
            currentTX.raw += currentTX.blockstamp + '\n';
            currentTX.locktime = parseInt(sp[7]);
            const linesToExtract = {
                issuers: {
                    start: start,
                    end: (start - 1) + nbIssuers
                },
                inputs: {
                    start: start + nbIssuers,
                    end: (start - 1) + nbIssuers + nbInputs
                },
                unlocks: {
                    start: start + nbIssuers + nbInputs,
                    end: (start - 1) + nbIssuers + nbInputs + nbUnlocks
                },
                outputs: {
                    start: start + nbIssuers + nbInputs + nbUnlocks,
                    end: (start - 1) + nbIssuers + nbInputs + nbUnlocks + nbOutputs
                },
                comments: {
                    start: start + nbIssuers + nbInputs + nbUnlocks + nbOutputs,
                    end: (start - 1) + nbIssuers + nbInputs + nbUnlocks + nbOutputs + hasComment
                },
                signatures: {
                    start: start + nbIssuers + nbInputs + nbUnlocks + nbOutputs + hasComment,
                    end: (start - 1) + 2 * nbIssuers + nbInputs + nbUnlocks + nbOutputs + hasComment
                }
            };
            ['issuers', 'inputs', 'unlocks', 'outputs', 'comments', 'signatures'].forEach((prop) => {
                currentTX[prop] = currentTX[prop] || [];
                for (let j = linesToExtract[prop].start; j <= linesToExtract[prop].end; j++) {
                    const line = lines[i + j];
                    if (line.match(regexps[prop])) {
                        currentTX.raw += line + '\n';
                        currentTX[prop].push(line);
                    }
                }
            });
            // Comment
            if (hasComment) {
                currentTX.comment = currentTX.comments[0];
            }
            else {
                currentTX.comment = '';
            }
            currentTX.hash = common_1.hashf(index_1.rawer.getTransaction(currentTX)).toUpperCase();
            // Add to txs array
            transactions.push(currentTX);
            i = i + 1 + 2 * nbIssuers + nbInputs + nbUnlocks + nbOutputs + hasComment;
        }
        else {
            // Not a transaction header, stop reading
            i = lines.length;
        }
    }
    return transactions;
}
//# sourceMappingURL=block.js.map