"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const GenericParser_1 = require("./GenericParser");
const constants_1 = require("../../../lib/common-libs/constants");
const index_1 = require("../../../lib/common-libs/index");
class PeerParser extends GenericParser_1.GenericParser {
    constructor() {
        super([
            { prop: "version", regexp: /Version: (.*)/ },
            { prop: "currency", regexp: /Currency: (.*)/ },
            { prop: "pubkey", regexp: /PublicKey: (.*)/ },
            { prop: "block", regexp: constants_1.CommonConstants.PEER.BLOCK },
            {
                prop: "endpoints", regexp: /Endpoints:\n([\s\S]*)/, parser: (str) => str.split("\n")
            }
        ], index_1.rawer.getPeer);
    }
    _clean(obj) {
        obj.endpoints = obj.endpoints || [];
        // Removes trailing space
        if (obj.endpoints.length > 0)
            obj.endpoints.splice(obj.endpoints.length - 1, 1);
        obj.getBMAOrNull = function () {
            let bma = null;
            obj.endpoints.forEach((ep) => {
                let matches = !bma && ep.match(constants_1.CommonConstants.BMA_REGEXP);
                if (matches) {
                    bma = {
                        "dns": matches[2] || '',
                        "ipv4": matches[4] || '',
                        "ipv6": matches[6] || '',
                        "port": matches[8] || 9101
                    };
                }
            });
            return bma || null;
        };
    }
    _verify(obj) {
        let err = null;
        const codes = {
            'BAD_VERSION': 150,
            'BAD_CURRENCY': 151,
            'BAD_DNS': 152,
            'BAD_IPV4': 153,
            'BAD_IPV6': 154,
            'BAD_PORT': 155,
            'BAD_FINGERPRINT': 156,
            'BAD_BLOCK': 157,
            'NO_IP_GIVEN': 158,
            'TOO_LONG_ENDPOINT': 159
        };
        if (!err) {
            // Version
            if (!obj.version || !obj.version.match(constants_1.CommonConstants.DOCUMENTS_VERSION_REGEXP))
                err = { code: codes.BAD_VERSION, message: "Version unknown" };
        }
        if (!err) {
            // PublicKey
            if (!obj.pubkey || !obj.pubkey.match(constants_1.CommonConstants.BASE58))
                err = { code: codes.BAD_FINGERPRINT, message: "Incorrect PublicKey field" };
        }
        if (!err) {
            // Block
            if (!obj.block)
                err = { code: codes.BAD_BLOCK, message: "Incorrect Block field" };
        }
        if (!err) {
            // Endpoint length
            for (const ep of (obj.endpoints || [])) {
                if (!err && ep.length > 255) {
                    err = { code: codes.TOO_LONG_ENDPOINT, message: "An endpoint has maximum 255 characters length." };
                }
            }
        }
        // Basic Merkled API requirements
        let bma = obj.getBMAOrNull();
        if (bma) {
            if (!err) {
                // DNS
                if (bma.dns && !bma.dns.match(/^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])$/))
                    err = { code: codes.BAD_DNS, message: "Incorrect Dns field" };
            }
            if (!err) {
                // IPv4
                if (bma.ipv4 && !bma.ipv4.match(/^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$/))
                    err = { code: codes.BAD_IPV4, message: "Incorrect IPv4 field" };
            }
            if (!err) {
                // IPv6
                if (bma.ipv6 && !bma.ipv6.match(/^((([0-9A-Fa-f]{1,4}:){7}[0-9A-Fa-f]{1,4})|(([0-9A-Fa-f]{1,4}:){6}:[0-9A-Fa-f]{1,4})|(([0-9A-Fa-f]{1,4}:){5}:([0-9A-Fa-f]{1,4}:)?[0-9A-Fa-f]{1,4})|(([0-9A-Fa-f]{1,4}:){4}:([0-9A-Fa-f]{1,4}:){0,2}[0-9A-Fa-f]{1,4})|(([0-9A-Fa-f]{1,4}:){3}:([0-9A-Fa-f]{1,4}:){0,3}[0-9A-Fa-f]{1,4})|(([0-9A-Fa-f]{1,4}:){2}:([0-9A-Fa-f]{1,4}:){0,4}[0-9A-Fa-f]{1,4})|(([0-9A-Fa-f]{1,4}:){6}((b((25[0-5])|(1d{2})|(2[0-4]d)|(d{1,2}))b).){3}(b((25[0-5])|(1d{2})|(2[0-4]d)|(d{1,2}))b))|(([0-9A-Fa-f]{1,4}:){0,5}:((b((25[0-5])|(1d{2})|(2[0-4]d)|(d{1,2}))b).){3}(b((25[0-5])|(1d{2})|(2[0-4]d)|(d{1,2}))b))|(::([0-9A-Fa-f]{1,4}:){0,5}((b((25[0-5])|(1d{2})|(2[0-4]d)|(d{1,2}))b).){3}(b((25[0-5])|(1d{2})|(2[0-4]d)|(d{1,2}))b))|([0-9A-Fa-f]{1,4}::([0-9A-Fa-f]{1,4}:){0,5}[0-9A-Fa-f]{1,4})|(::([0-9A-Fa-f]{1,4}:){0,6}[0-9A-Fa-f]{1,4})|(([0-9A-Fa-f]{1,4}:){1,7}:))$/))
                    err = { code: codes.BAD_IPV6, message: "Incorrect IPv6 field" };
            }
            if (!err) {
                // IP
                if (!bma.dns && !bma.ipv4 && !bma.ipv6)
                    err = { code: codes.NO_IP_GIVEN, message: "It must be given at least DNS or one IP, either v4 or v6" };
            }
            if (!err) {
                // Port
                if (bma.port && !(bma.port + "").match(/^\d+$/))
                    err = { code: codes.BAD_PORT, message: "Port must be provided and match an integer format" };
            }
        }
        return err && err.message;
    }
    ;
}
exports.PeerParser = PeerParser;
//# sourceMappingURL=peer.js.map