"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const constants_1 = require("../constants");
const GenericParser_1 = require("./GenericParser");
const index_1 = require("../../../lib/common-libs/index");
const txunlock_1 = require("../txunlock");
class TransactionParser extends GenericParser_1.GenericParser {
    constructor() {
        super([
            { prop: "version", regexp: /Version: (.*)/ },
            { prop: "currency", regexp: /Currency: (.*)/ },
            { prop: "issuers", regexp: /Issuers:\n([\s\S]*)Inputs/, parser: extractIssuers },
            { prop: "inputs", regexp: /Inputs:\n([\s\S]*)Unlocks/, parser: extractInputs },
            { prop: "unlocks", regexp: /Unlocks:\n([\s\S]*)Outputs/, parser: extractUnlocks },
            { prop: "outputs", regexp: /Outputs:\n([\s\S]*)/, parser: extractOutputs },
            { prop: "comment", regexp: constants_1.CommonConstants.TRANSACTION.COMMENT },
            { prop: "locktime", regexp: constants_1.CommonConstants.TRANSACTION.LOCKTIME },
            { prop: "blockstamp", regexp: constants_1.CommonConstants.TRANSACTION.BLOCKSTAMP },
            { prop: "signatures", regexp: /Outputs:\n([\s\S]*)/, parser: extractSignatures }
        ], index_1.rawer.getTransaction);
    }
    _clean(obj) {
        obj.comment = obj.comment || "";
        obj.locktime = parseInt(obj.locktime) || 0;
        obj.signatures = obj.signatures || [];
        obj.issuers = obj.issuers || [];
        obj.inputs = obj.inputs || [];
        obj.outputs = obj.outputs || [];
        obj.signatures.push(obj.signature);
        const compactSize = 2 // Header + blockstamp
            + obj.issuers.length
            + obj.inputs.length
            + obj.unlocks.length
            + obj.outputs.length
            + (obj.comment ? 1 : 0)
            + obj.signatures;
        if (compactSize > 100) {
            throw 'A transaction has a maximum size of 100 lines';
        }
    }
    _verify(obj) {
        let err = null;
        const codes = {
            'BAD_VERSION': 150,
            'NO_BLOCKSTAMP': 151
        };
        if (!err) {
            // Version
            if (!obj.version || !obj.version.match(constants_1.CommonConstants.DOCUMENTS_TRANSACTION_VERSION_REGEXP))
                err = { code: codes.BAD_VERSION, message: "Version unknown" };
            // Blockstamp
            if (!obj.blockstamp || !obj.blockstamp.match(constants_1.CommonConstants.BLOCKSTAMP_REGEXP))
                err = { code: codes.BAD_VERSION, message: "Blockstamp is required" };
        }
        return err && err.message;
    }
}
exports.TransactionParser = TransactionParser;
function extractIssuers(raw) {
    const issuers = [];
    const lines = raw.split(/\n/);
    for (const line of lines) {
        if (line.match(constants_1.CommonConstants.TRANSACTION.SENDER)) {
            issuers.push(line);
        }
        else {
            // Not a pubkey, stop reading
            break;
        }
    }
    return issuers;
}
function extractInputs(raw) {
    const inputs = [];
    const lines = raw.split(/\n/);
    for (const line of lines) {
        if (line.match(constants_1.CommonConstants.TRANSACTION.SOURCE_V3)) {
            inputs.push(line);
        }
        else {
            // Not a transaction input, stop reading
            break;
        }
    }
    return inputs;
}
function extractUnlocks(raw) {
    const unlocks = [];
    const lines = raw.split(/\n/);
    for (const line of lines) {
        if (line.match(constants_1.CommonConstants.TRANSACTION.UNLOCK)) {
            unlocks.push(line);
        }
        else {
            // Not a transaction unlock, stop reading
            break;
        }
    }
    return unlocks;
}
function extractOutputs(raw) {
    const outputs = [];
    const lines = raw.split(/\n/);
    for (const line of lines) {
        if (line.match(constants_1.CommonConstants.TRANSACTION.TARGET)) {
            outputs.push(line);
            const unlocked = txunlock_1.checkGrammar(line.split(':')[2]);
            if (unlocked === null) {
                throw Error("Wrong output format");
            }
        }
        else {
            // Not a transaction input, stop reading
            break;
        }
    }
    return outputs;
}
function extractSignatures(raw) {
    const signatures = [];
    const lines = raw.split(/\n/);
    for (const line of lines) {
        if (line.match(constants_1.CommonConstants.SIG)) {
            signatures.push(line);
        }
    }
    return signatures;
}
//# sourceMappingURL=transaction.js.map