"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const BlockDTO_1 = require("../dto/BlockDTO");
const DuniterBlockchain_1 = require("../blockchain/DuniterBlockchain");
const underscore_1 = require("../common-libs/underscore");
const errors_1 = require("../common-libs/errors");
const indexer = require('../indexer').Indexer;
const constants = require('../constants');
class BlockchainContext {
    constructor() {
        this.HEADrefreshed = Promise.resolve();
    }
    /**
     * Refresh the virtual HEAD value for determined variables of the next coming block, avoiding to recompute them
     * each time a new block arrives to check if the values are correct. We can know and store them early on, in vHEAD.
     */
    refreshHead() {
        this.HEADrefreshed = (async () => {
            this.vHEAD_1 = await this.dal.head(1);
            // We suppose next block will have same version #, and no particular data in the block (empty index)
            let block;
            // But if no HEAD_1 exist, we must initialize a block with default values
            if (!this.vHEAD_1) {
                block = {
                    version: constants.BLOCK_GENERATED_VERSION,
                    time: Math.round(Date.now() / 1000),
                    powMin: this.conf.powMin || 0,
                    powZeros: 0,
                    powRemainder: 0,
                    avgBlockSize: 0
                };
            }
            else {
                block = { version: this.vHEAD_1.version };
            }
            this.vHEAD = await indexer.completeGlobalScope(BlockDTO_1.BlockDTO.fromJSONObject(block), this.conf, [], this.dal);
        })();
        return this.HEADrefreshed;
    }
    /**
     * Gets a copy of vHEAD, extended with some extra properties.
     * @param props The extra properties to add.
     */
    async getvHeadCopy(props = {}) {
        if (!this.vHEAD) {
            await this.refreshHead();
        }
        const copy = {};
        const keys = Object.keys(this.vHEAD);
        for (const k of keys) {
            copy[k] = this.vHEAD[k];
        }
        underscore_1.Underscore.extend(copy, props);
        return copy;
    }
    /**
     * Get currently written HEAD.
     */
    async getvHEAD_1() {
        if (!this.vHEAD) {
            await this.refreshHead();
        }
        return this.vHEAD_1;
    }
    /**
     * Utility method: gives the personalized difficulty level of a given issuer for next block.
     * @param issuer The issuer we want to get the difficulty level.
     */
    async getIssuerPersonalizedDifficulty(issuer) {
        const local_vHEAD = await this.getvHeadCopy({ issuer });
        await indexer.preparePersonalizedPoW(local_vHEAD, this.vHEAD_1, (n, m, p = "") => this.dal.range(n, m, p), this.conf);
        return local_vHEAD.issuerDiff;
    }
    setConfDAL(newConf, newDAL) {
        this.dal = newDAL;
        this.conf = newConf;
        this.logger = require('../logger').NewLogger(this.dal.profile);
    }
    async checkBlock(block, withPoWAndSignature) {
        return DuniterBlockchain_1.DuniterBlockchain.checkBlock(block, withPoWAndSignature, this.conf, this.dal);
    }
    async addBlock(obj, index = null, HEAD = null, trim) {
        const block = await DuniterBlockchain_1.DuniterBlockchain.pushTheBlock(obj, index, HEAD, this.conf, this.dal, this.logger, trim);
        this.vHEAD_1 = this.vHEAD = null;
        return block;
    }
    async addSideBlock(obj) {
        const dbb = await DuniterBlockchain_1.DuniterBlockchain.pushSideBlock(obj, this.dal, this.logger);
        return dbb.toBlockDTO();
    }
    async revertCurrentBlock() {
        const head_1 = await this.dal.bindexDAL.head(1);
        this.logger.debug('Reverting block #%s...', head_1.number);
        const block = await this.dal.getAbsoluteValidBlockInForkWindow(head_1.number, head_1.hash);
        if (!block) {
            throw errors_1.DataErrors[errors_1.DataErrors.BLOCK_TO_REVERT_NOT_FOUND];
        }
        await DuniterBlockchain_1.DuniterBlockchain.revertBlock(head_1.number, head_1.hash, this.dal, block);
        // Invalidates the head, since it has changed.
        await this.refreshHead();
        return block;
    }
    async revertCurrentHead() {
        const head_1 = await this.dal.bindexDAL.head(1);
        this.logger.debug('Reverting HEAD~1... (b#%s)', head_1.number);
        await DuniterBlockchain_1.DuniterBlockchain.revertBlock(head_1.number, head_1.hash, this.dal);
        // Invalidates the head, since it has changed.
        await this.refreshHead();
    }
    async applyNextAvailableFork() {
        const current = await this.current();
        this.logger.debug('Find next potential block #%s...', current.number + 1);
        const forks = await this.dal.getForkBlocksFollowing(current);
        if (!forks.length) {
            throw constants.ERRORS.NO_POTENTIAL_FORK_AS_NEXT;
        }
        const block = forks[0];
        await this.checkAndAddBlock(BlockDTO_1.BlockDTO.fromJSONObject(block));
        this.logger.debug('Applied block #%s', block.number);
    }
    async checkAndAddBlock(block, trim = true) {
        const { index, HEAD } = await this.checkBlock(block, constants.WITH_SIGNATURES_AND_POW);
        return await this.addBlock(block, index, HEAD, trim);
    }
    current() {
        return this.dal.getCurrentBlockOrNull();
    }
    async checkHaveEnoughLinks(target, newLinks) {
        const links = await this.dal.getValidLinksTo(target);
        let count = links.length;
        if (newLinks[target] && newLinks[target].length) {
            count += newLinks[target].length;
        }
        if (count < this.conf.sigQty) {
            throw 'Key ' + target + ' does not have enough links (' + count + '/' + this.conf.sigQty + ')';
        }
    }
}
exports.BlockchainContext = BlockchainContext;
//# sourceMappingURL=BlockchainContext.js.map