"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const other_constants_1 = require("../../other_constants");
const directory_1 = require("../../system/directory");
const sqlite3 = require("sqlite3").verbose();
const MEMORY_PATH = ':memory:';
class SQLiteDriver {
    constructor(path) {
        this.path = path;
        this.dbPromise = null;
        this.logger = require('../../logger').NewLogger('driver');
    }
    getDB() {
        if (!this.dbPromise) {
            this.dbPromise = (async () => {
                this.logger.debug('Opening SQLite database "%s"...', this.path);
                let sqlite = new sqlite3.Database(this.path);
                await new Promise((resolve) => sqlite.once('open', resolve));
                // Database is opened
                if (other_constants_1.OtherConstants.SQL_TRACES) {
                    sqlite.on('trace', (trace) => {
                        this.logger.trace(trace);
                    });
                }
                // Force case sensitiveness on LIKE operator
                const sql = 'PRAGMA case_sensitive_like=ON';
                await new Promise((resolve, reject) => sqlite.exec(sql, (err) => {
                    if (err)
                        return reject(Error('SQL error "' + err.message + '" on INIT queries "' + sql + '"'));
                    return resolve();
                }));
                // Database is ready
                return sqlite;
            })();
        }
        return this.dbPromise;
    }
    async executeAll(sql, params) {
        const db = await this.getDB();
        return new Promise((resolve, reject) => db.all(sql, params, (err, rows) => {
            if (err) {
                return reject(Error('SQL error "' + err.message + '" on query "' + sql + '"'));
            }
            else {
                return resolve(rows);
            }
        }));
    }
    async executeSql(sql) {
        const db = await this.getDB();
        return new Promise((resolve, reject) => db.exec(sql, (err) => {
            if (err) {
                return reject(Error('SQL error "' + err.message + '" on query "' + sql + '"'));
            }
            else {
                return resolve();
            }
        }));
    }
    async destroyDatabase() {
        this.logger.debug('Removing SQLite database...');
        await this.closeConnection();
        if (this.path !== MEMORY_PATH) {
            await directory_1.RealFS().fsUnlink(this.path);
        }
        this.logger.debug('Database removed');
    }
    async closeConnection() {
        if (!this.dbPromise) {
            return;
        }
        const db = await this.getDB();
        if (process.platform === 'win32') {
            db.open; // For an unknown reason, we need this line.
        }
        await new Promise((resolve, reject) => {
            this.logger.debug('Trying to close SQLite...');
            db.on('close', () => {
                this.logger.info('Database closed.');
                this.dbPromise = null;
                resolve();
            });
            db.on('error', (err) => {
                if (err && err.message === 'SQLITE_MISUSE: Database is closed') {
                    this.dbPromise = null;
                    return resolve();
                }
                reject(err);
            });
            try {
                db.close();
            }
            catch (e) {
                this.logger.error(e);
                throw e;
            }
        });
    }
}
exports.SQLiteDriver = SQLiteDriver;
//# sourceMappingURL=SQLiteDriver.js.map