"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
class DividendDaoHandler {
    static getNewDividendEntry(pub) {
        return { pub, member: true, availables: [], dividends: [], consumed: [], consumedUDs: [] };
    }
    static produceDividend(r, blockNumber, dividend, unitbase, dividends = []) {
        r.availables.push(blockNumber);
        r.dividends.push({ amount: dividend, base: unitbase });
        dividends.push({
            srcType: 'D',
            amount: dividend,
            base: unitbase,
            conditions: 'SIG(' + r.pub + ')',
            op: 'CREATE',
            identifier: r.pub,
            pos: blockNumber
        });
    }
    static consume(m, dividendToConsume) {
        const index = m.availables.indexOf(dividendToConsume.pos);
        // We add it to the consumption history
        m.consumed.push(dividendToConsume.writtenOn); // `writtenOn` is the date (block#) of consumption
        m.consumedUDs.push({
            dividendNumber: dividendToConsume.pos,
            dividend: m.dividends[index],
            txCreatedOn: dividendToConsume.created_on,
            txLocktime: dividendToConsume.locktime,
            txHash: dividendToConsume.tx,
        });
        // We remove it from available dividends
        m.availables.splice(index, 1);
        m.dividends.splice(index, 1);
    }
    static udSources(member) {
        return member.availables.map(pos => this.toUDSource(member, pos));
    }
    static toUDSource(entry, pos) {
        const index = entry.availables.indexOf(pos);
        if (index === -1) {
            return null;
        }
        const src = entry.dividends[index];
        return {
            consumed: false,
            pos,
            amount: src.amount,
            base: src.base,
        };
    }
    static getUDSourceByIdPosAmountBase(member, identifier, pos, amount, base) {
        let src = null;
        if (member) {
            const udSrc = this.toUDSource(member, pos);
            if (udSrc && udSrc.amount === amount && udSrc.base === base) {
                src = udSrc;
            }
        }
        return [{
                written_time: 0,
                conditions: 'SIG(' + identifier + ')',
                consumed: !src,
                amount,
                base
            }];
    }
    static getUDSource(member, identifier, pos) {
        let src = null;
        if (member) {
            src = this.toUDSource(member, pos);
        }
        if (!src) {
            return null;
        }
        return {
            written_time: 0,
            conditions: 'SIG(' + identifier + ')',
            consumed: !src,
            amount: src.amount,
            base: src.base
        };
    }
    static getWrittenOnUDs(m, number, res) {
        const s = this.toUDSource(m, number);
        res.push({
            srcType: 'D',
            op: 'CREATE',
            conditions: 'SIG(' + m.pub + ')',
            amount: s.amount,
            base: s.base,
            identifier: m.pub,
            pos: s.pos
        });
    }
    static removeDividendsProduced(m, number, createdUDsDestroyedByRevert) {
        const index = m.availables.indexOf(number);
        const src = m.dividends[index];
        createdUDsDestroyedByRevert.push({
            conditions: 'SIG(' + m.pub + ')',
            pos: number,
            identifier: m.pub,
            amount: src.amount,
            base: src.base,
            srcType: 'D',
            op: 'CREATE'
        });
        m.availables.splice(index, 1);
        m.dividends.splice(index, 1);
    }
    static unconsumeDividends(m, number, consumedUDsRecoveredByRevert) {
        let index;
        do {
            index = m.consumed.indexOf(number);
            if (index !== -1) {
                const src = m.consumedUDs[index].dividend;
                consumedUDsRecoveredByRevert.push({
                    conditions: 'SIG(' + m.pub + ')',
                    pos: m.consumedUDs[index].dividendNumber,
                    identifier: m.pub,
                    amount: src.amount,
                    base: src.base,
                    srcType: 'D',
                    op: 'CREATE'
                });
                // We put it back as available
                m.availables.push(m.consumedUDs[index].dividendNumber);
                m.dividends.push(m.consumedUDs[index].dividend);
                // We remove it from consumed
                m.consumed.splice(index, 1);
                m.consumedUDs.splice(index, 1);
            }
        } while (index !== -1);
    }
    static trimConsumed(m, belowNumber) {
        let updated = false;
        for (let i = 0; i < m.consumed.length; i++) {
            const consumedBlockNumber = m.consumed[i];
            if (consumedBlockNumber < belowNumber) {
                // We trim this entry as it can't be reverted now
                m.consumed.splice(i, 1);
                m.consumedUDs.splice(i, 1);
                i--; // The array changed, we loop back before i++
                updated = true;
            }
        }
        return updated;
    }
    static toDump(rows) {
        const entries = [];
        for (const m of rows) {
            // Generate for unspent UDs
            for (let i = 0; i < m.availables.length; i++) {
                const writtenOn = m.availables[i];
                const ud = m.dividends[i];
                entries.push({
                    op: 'CREATE',
                    index: 'SINDEX',
                    srcType: 'D',
                    tx: null,
                    identifier: m.pub,
                    writtenOn,
                    pos: writtenOn,
                    created_on: 'NULL',
                    written_on: writtenOn + '',
                    written_time: 0,
                    amount: ud.amount,
                    base: ud.base,
                    locktime: null,
                    consumed: false,
                    conditions: 'SIG(' + m.pub + ')',
                    unlock: null,
                    txObj: null,
                    age: 0,
                });
            }
            // Generate for spent UDs
            for (let i = 0; i < m.consumed.length; i++) {
                const writtenOn = m.consumed[i];
                const ud = m.consumedUDs[i];
                entries.push({
                    op: 'CREATE',
                    index: 'SINDEX',
                    srcType: 'D',
                    tx: null,
                    identifier: m.pub,
                    writtenOn: ud.dividendNumber,
                    pos: ud.dividendNumber,
                    created_on: 'NULL',
                    written_on: writtenOn + '',
                    written_time: 0,
                    amount: ud.dividend.amount,
                    base: ud.dividend.base,
                    locktime: null,
                    consumed: false,
                    conditions: 'SIG(' + m.pub + ')',
                    unlock: null,
                    txObj: null,
                    age: 0,
                });
                entries.push({
                    op: 'UPDATE',
                    index: 'SINDEX',
                    srcType: 'D',
                    tx: ud.txHash,
                    identifier: m.pub,
                    writtenOn,
                    pos: ud.dividendNumber,
                    created_on: ud.txCreatedOn,
                    written_on: writtenOn + '',
                    written_time: 0,
                    amount: ud.dividend.amount,
                    base: ud.dividend.base,
                    locktime: ud.txLocktime,
                    consumed: true,
                    conditions: 'SIG(' + m.pub + ')',
                    unlock: null,
                    txObj: null,
                    age: 0,
                });
            }
        }
        return entries;
    }
}
exports.DividendDaoHandler = DividendDaoHandler;
//# sourceMappingURL=DividendDaoHandler.js.map