"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
Object.defineProperty(exports, "__esModule", { value: true });
const MonitorExecutionTime_1 = require("../../../debug/MonitorExecutionTime");
const indexer_1 = require("../../../indexer");
const LevelDBTable_1 = require("./LevelDBTable");
const underscore_1 = require("../../../common-libs/underscore");
const pint_1 = require("../../../common-libs/pint");
const reduce_1 = require("../../../common-libs/reduce");
const LevelDBWrittenOnIndexer_1 = require("./indexers/LevelDBWrittenOnIndexer");
const LevelMIndexRevokesOnIndexer_1 = require("./indexers/LevelMIndexRevokesOnIndexer");
const LevelMIndexExpiresOnIndexer_1 = require("./indexers/LevelMIndexExpiresOnIndexer");
class LevelDBMindex extends LevelDBTable_1.LevelDBTable {
    constructor(getLevelDB) {
        super('level_mindex', getLevelDB);
        this.getLevelDB = getLevelDB;
    }
    /**
     * TECHNICAL
     */
    async init() {
        await super.init();
        this.indexForExpiresOn = new LevelMIndexExpiresOnIndexer_1.LevelMIndexExpiresOnIndexer('level_mindex/expiresOn', this.getLevelDB);
        this.indexForRevokesOn = new LevelMIndexRevokesOnIndexer_1.LevelMIndexRevokesOnIndexer('level_mindex/revokesOn', this.getLevelDB);
        this.indexForWrittenOn = new LevelDBWrittenOnIndexer_1.LevelDBWrittenOnIndexer('level_mindex/writtenOn', this.getLevelDB, i => i.pub);
        await this.indexForExpiresOn.init();
        await this.indexForRevokesOn.init();
        await this.indexForWrittenOn.init();
    }
    async close() {
        await super.close();
        await this.indexForExpiresOn.close();
        await this.indexForRevokesOn.close();
        await this.indexForWrittenOn.close();
    }
    /**
     * INSERT
     */
    async insert(record) {
        await this.insertBatch([record]);
    }
    async insertBatch(records) {
        // Database insertion
        let prevRecords = [];
        const recordsByPub = reduce_1.reduceGroupBy(records, 'pub');
        await Promise.all(underscore_1.Underscore.keys(recordsByPub).map(String).map(async (pub) => {
            const existing = (await this.getOrNull(pub)) || [];
            prevRecords = prevRecords.concat(existing);
            await this.put(pub, existing.concat(recordsByPub[pub]));
        }));
        // Indexation
        await this.indexForExpiresOn.onInsert(records, prevRecords);
        await this.indexForRevokesOn.onInsert(records, prevRecords);
        await this.indexForWrittenOn.onInsert(records);
    }
    /**
     * Reduceable DAO
     */
    async trimRecords(belowNumber) {
        // Trim writtenOn: we remove from the index the blocks below `belowNumber`, and keep track of the deleted values
        const pubkeys = underscore_1.Underscore.uniq((await this.indexForWrittenOn.deleteBelow(belowNumber)));
        // For each entry, we trim the records of our INDEX
        await Promise.all(pubkeys.map(async (pub) => {
            const oldEntries = await this.get(pub);
            const newEntries = indexer_1.reduceForDBTrimming(oldEntries, belowNumber);
            await this.put(pub, newEntries);
        }));
        await this.indexForExpiresOn.onTrimming(belowNumber);
        await this.indexForRevokesOn.onTrimming(belowNumber);
    }
    /**
     * Generic DAO
     */
    async findRawWithOrder(criterion, sort) {
        const rows = (await this.findAllValues()).reduce(reduce_1.reduceConcat, []);
        return underscore_1.Underscore.sortBy(rows, r => `${String(r.writtenOn).padStart(10, '0')}-${r.pub}`);
    }
    async getWrittenOn(blockstamp) {
        const ids = (await this.indexForWrittenOn.getWrittenOnKeys(pint_1.pint(blockstamp))) || [];
        return (await Promise.all(ids.map(id => this.get(id)))).reduce(reduce_1.reduceConcat, []).filter(e => e.written_on === blockstamp);
    }
    async removeBlock(blockstamp) {
        // Trim writtenOn: we remove from the index the blocks below `belowNumber`, and keep track of the deleted values
        let newStateRecords = [];
        const writteOn = pint_1.pint(blockstamp);
        const pubkeys = underscore_1.Underscore.uniq((await this.indexForWrittenOn.deleteAt(writteOn)));
        let removedRecords = [];
        // For each entry, we trim the records of our INDEX
        await Promise.all(pubkeys.map(async (pub) => {
            const records = await this.get(pub);
            const keptRecords = records.filter(e => e.written_on !== blockstamp);
            removedRecords = removedRecords.concat(records.filter(e => e.written_on === blockstamp));
            newStateRecords = newStateRecords.concat(keptRecords);
            await this.put(pub, keptRecords);
        }));
        // Update indexes
        await this.indexForExpiresOn.onRemove(removedRecords, newStateRecords);
        await this.indexForRevokesOn.onRemove(removedRecords, newStateRecords);
    }
    //------------- DAO QUERIES --------------
    async findByPubAndChainableOnGt(pub, medianTime) {
        return (await this.reducable(pub)).filter(e => e.chainable_on && e.chainable_on > medianTime);
    }
    async findExpiresOnLteAndRevokesOnGt(medianTime) {
        return this.indexForExpiresOn.findExpiresOnLte(medianTime);
    }
    async findPubkeysThatShouldExpire(medianTime) {
        const results = [];
        const pubkeys = await this.findExpiresOnLteAndRevokesOnGt(medianTime);
        for (const pub of pubkeys) {
            const MS = await this.getReducedMS(pub); // We are sure because `memberships` already comes from the MINDEX
            const hasRenewedSince = MS.expires_on > medianTime;
            if (!MS.expired_on && !hasRenewedSince) {
                results.push({
                    pub: MS.pub,
                    created_on: MS.created_on,
                });
            }
        }
        return results;
    }
    async findRevokesOnLteAndRevokedOnIsNull(medianTime) {
        return this.indexForRevokesOn.findRevokesOnLte(medianTime);
    }
    async getReducedMS(pub) {
        const reducable = await this.reducable(pub);
        return indexer_1.reduceOrNull(reducable);
    }
    async getReducedMSForImplicitRevocation(pub) {
        return this.getReducedMS(pub);
    }
    async getRevokedPubkeys() {
        return this.findWhereTransform(v => !!indexer_1.reduce(v).revoked_on, kv => kv.key);
    }
    async reducable(pub) {
        return (await this.getOrNull(pub)) || [];
    }
}
__decorate([
    MonitorExecutionTime_1.MonitorExecutionTime()
], LevelDBMindex.prototype, "insert", null);
__decorate([
    MonitorExecutionTime_1.MonitorExecutionTime()
], LevelDBMindex.prototype, "insertBatch", null);
exports.LevelDBMindex = LevelDBMindex;
//# sourceMappingURL=LevelDBMindex.js.map