"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
class LevelDBTable {
    constructor(name, getLevelDB) {
        this.name = name;
        this.getLevelDB = getLevelDB;
    }
    cleanCache() {
    }
    triggerInit() {
    }
    async close() {
        await this.db.close();
    }
    async init() {
        this.db = await this.getLevelDB(`${this.name}`);
    }
    async get(k) {
        const data = await this.db.get(k);
        return JSON.parse(String(data));
    }
    async getOrNull(k) {
        try {
            const data = await this.db.get(k);
            return JSON.parse(String(data));
        }
        catch (e) {
            if (!e || e.type !== 'NotFoundError') {
                throw Error(e);
            }
            return null;
        }
    }
    async del(k) {
        return await this.db.del(k);
    }
    async put(k, record) {
        return await this.db.put(k, JSON.stringify(record));
    }
    async batchInsert(records, key) {
        const batch = records.map(r => {
            return {
                type: 'put',
                key: r[key],
                value: JSON.stringify(r)
            };
        });
        await this.db.batch(batch);
    }
    async batchInsertWithKeyComputing(records, keyComputing) {
        const batch = records.map(r => {
            return {
                type: 'put',
                key: keyComputing(r),
                value: JSON.stringify(r)
            };
        });
        await this.db.batch(batch);
    }
    async count(options) {
        let count = 0;
        await new Promise(res => {
            this.db.createReadStream(options)
                .on('data', () => count++)
                .on('close', res);
        });
        return count;
    }
    async readAll(callback, options) {
        await new Promise(res => {
            this.db.createReadStream(options)
                .on('data', data => callback(JSON.parse(String(data.value))))
                .on('close', res);
        });
    }
    async readAllKeyValue(callback, options) {
        await new Promise(res => {
            this.db.createReadStream(options)
                .on('data', data => callback({
                key: String(data.key),
                value: JSON.parse(String(data.value))
            }))
                .on('close', res);
        });
    }
    async applyAllKeyValue(callback, options) {
        const ops = [];
        await new Promise(res => {
            this.db.createReadStream(options)
                .on('data', data => ops.push(callback({
                key: String(data.key),
                value: JSON.parse(String(data.value))
            })))
                .on('close', res);
        });
        await Promise.all(ops);
    }
    async deleteWhere(options) {
        const deletedKv = [];
        await this.applyAllKeyValue(async (kv) => {
            deletedKv.push(kv);
            await this.del(kv.key);
        }, options);
        return deletedKv;
    }
    async findAllKeys(options) {
        const data = [];
        await this.readAllKeyValue(kv => {
            data.push(kv.key);
        }, options);
        return data;
    }
    async findAllValues(options) {
        const data = [];
        await this.readAllKeyValue(kv => {
            data.push(kv.value);
        }, options);
        return data;
    }
    async findWhere(filter) {
        return this.findWhereTransform(filter, t => t.value);
    }
    async findWhereTransform(filter, transform) {
        const data = [];
        await this.readAllKeyValue(kv => {
            if (!filter || filter(kv.value)) {
                data.push(transform(kv));
            }
        }, {});
        return data;
    }
    async dump(dumpValue = (v) => v) {
        let count = 0;
        await this.readAllKeyValue(entry => {
            console.log(entry.key, dumpValue({
                key: entry.key,
                value: entry.value
            }));
            count++;
        });
        return count;
    }
}
exports.LevelDBTable = LevelDBTable;
//# sourceMappingURL=LevelDBTable.js.map