"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const LevelDBDataIndex_1 = require("../generic/LevelDBDataIndex");
const indexer_1 = require("../../../../indexer");
const reduce_1 = require("../../../../common-libs/reduce");
const pint_1 = require("../../../../common-libs/pint");
const underscore_1 = require("../../../../common-libs/underscore");
class LevelMIndexExpiresOnIndexer extends LevelDBDataIndex_1.LevelDBDataIndex {
    async onInsert(records, prevState) {
        const prevStateByPub = reduce_1.reduceGroupBy(prevState, 'pub');
        // Case 1: expires_on change (when MS JOIN|RENEW)
        const byExpiresOn = reduce_1.reduceGroupBy(records.filter(e => e.expires_on), "expires_on");
        await Promise.all(underscore_1.Underscore.keys(byExpiresOn)
            .map(async (expiresOn) => {
            const pubkeys = byExpiresOn[expiresOn].map(e => e.pub);
            // 1. If the key had a previous revokes_on, we remove it
            const reducedWhosExpiresOnChanges = pubkeys.filter(p => prevStateByPub[p])
                .map(p => indexer_1.reduce(prevStateByPub[p]))
                .filter(r => r.expires_on && !r.expired_on);
            for (const reduced of reducedWhosExpiresOnChanges) {
                await this.removeAllKeysFromExpiresOn(reduced.expires_on, [reduced.pub]);
            }
            // 2. We put the new value
            await this.addAllKeysToExpiresOn(pint_1.pint(expiresOn), byExpiresOn[expiresOn].map(e => e.pub));
        }));
        // Case 2: expiration occurs
        const pubkeysToexpire = underscore_1.Underscore.uniq(records.filter(e => e.expired_on).map(r => r.pub));
        const prevStateFM = underscore_1.Underscore.values(prevStateByPub).map(indexer_1.reduce);
        const byExpiresOnPrevState = reduce_1.reduceGroupBy(prevStateFM.filter(r => pubkeysToexpire.includes(r.pub)), 'expires_on');
        await Promise.all(underscore_1.Underscore.keys(byExpiresOnPrevState)
            .map(async (expiresOn) => this.removeAllKeysFromExpiresOn(pint_1.pint(expiresOn), byExpiresOnPrevState[expiresOn].map(e => e.pub))));
    }
    async onRemove(records, newState) {
        const newStateByPub = reduce_1.reduceGroupBy(newState, 'pub');
        // Case 1: expires_on change REVERT
        const byExpiresOn = reduce_1.reduceGroupBy(records.filter(e => e.expires_on), "expires_on");
        await Promise.all(underscore_1.Underscore.keys(byExpiresOn)
            .map(async (expiresOn) => {
            const pubkeys = byExpiresOn[expiresOn].map(e => e.pub);
            // 1. Remove the existing value
            await this.removeAllKeysFromExpiresOn(pint_1.pint(expiresOn), pubkeys);
            // 2. Put back the old one if it exists
            const reduced = pubkeys
                .filter(p => newStateByPub[p])
                .map(p => newStateByPub[p])
                .map(indexer_1.reduce)
                .filter(r => r.expires_on);
            for (const r of reduced) {
                await this.addAllKeysToExpiresOn(r.expires_on, [r.pub]);
            }
        }));
        // Case 2: expiration REVERT
        const values = underscore_1.Underscore.values(newStateByPub).map(entries => indexer_1.reduce(entries));
        const byExpiredOn = reduce_1.reduceGroupBy(values, "expired_on");
        await Promise.all(underscore_1.Underscore.keys(byExpiredOn)
            .map(async (expiresOn) => this.addAllKeysToExpiresOn(pint_1.pint(expiresOn), byExpiredOn[expiresOn].map(e => e.pub))));
    }
    async addAllKeysToExpiresOn(expiresOn, pubkeys) {
        const key = LevelMIndexExpiresOnIndexer.trimKey(expiresOn);
        let entry = await this.getOrNull(key);
        if (!entry) {
            entry = [];
        }
        for (const pub of pubkeys) {
            entry.push(pub);
        }
        await this.put(key, entry);
    }
    async removeAllKeysFromExpiresOn(expiresOn, pubkeys) {
        // We remove the "expires_on" indexed values
        const key = LevelMIndexExpiresOnIndexer.trimKey(expiresOn);
        const entry = await this.get(key);
        for (const pub of pubkeys) {
            if (entry.includes(pub)) {
                entry.splice(entry.indexOf(pub), 1);
            }
        }
        if (entry.length) {
            // Some expirations left
            await this.put(key, entry); // TODO: test this, can occur, probably not covered
        }
        else {
            // No more expirations left
            await this.del(key);
        }
    }
    async findExpiresOnLte(medianTime) {
        return (await this.findAllValues({ lte: LevelMIndexExpiresOnIndexer.trimKey(medianTime) })).reduce(reduce_1.reduceConcat, []);
    }
    static trimKey(expiresOn) {
        return String(expiresOn).padStart(10, '0');
    }
}
exports.LevelMIndexExpiresOnIndexer = LevelMIndexExpiresOnIndexer;
//# sourceMappingURL=LevelMIndexExpiresOnIndexer.js.map