"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const LevelDBDataIndex_1 = require("../generic/LevelDBDataIndex");
const indexer_1 = require("../../../../indexer");
const reduce_1 = require("../../../../common-libs/reduce");
const underscore_1 = require("../../../../common-libs/underscore");
const pint_1 = require("../../../../common-libs/pint");
class LevelMIndexRevokesOnIndexer extends LevelDBDataIndex_1.LevelDBDataIndex {
    async onInsert(records, prevState) {
        const prevStateByPub = reduce_1.reduceGroupBy(prevState, 'pub');
        // Case 1: revokes_on change (when MS JOIN|RENEW)
        const byRevokesOn = reduce_1.reduceGroupBy(records.filter(e => e.revokes_on), "revokes_on");
        await Promise.all(underscore_1.Underscore.keys(byRevokesOn)
            .map(async (revokesOn) => {
            const pubkeys = byRevokesOn[revokesOn].map(e => e.pub);
            // 1. If the key had a previous revokes_on, we remove it
            const reducedWhosRevokesOnChanges = pubkeys.filter(p => prevStateByPub[p])
                .map(p => indexer_1.reduce(prevStateByPub[p]))
                .filter(r => r.revokes_on && !r.revoked_on);
            for (const reduced of reducedWhosRevokesOnChanges) {
                await this.removeAllKeysFromRevokesOn(reduced.revokes_on, [reduced.pub]);
            }
            // 2. We put the new value
            await this.addAllKeysToRevokesOn(pint_1.pint(revokesOn), byRevokesOn[revokesOn].map(e => e.pub));
        }));
        // Case 2: revocation occurs
        const pubkeysToRevoke = underscore_1.Underscore.uniq(records.filter(e => e.revoked_on).map(r => r.pub));
        const prevStateFM = underscore_1.Underscore.values(prevStateByPub).map(indexer_1.reduce);
        const byRevokesOnPrevState = reduce_1.reduceGroupBy(prevStateFM.filter(r => pubkeysToRevoke.includes(r.pub)), 'revokes_on');
        await Promise.all(underscore_1.Underscore.keys(byRevokesOnPrevState)
            .map(async (revokesOn) => this.removeAllKeysFromRevokesOn(pint_1.pint(revokesOn), byRevokesOnPrevState[revokesOn].map(e => e.pub))));
    }
    async onRemove(records, newState) {
        const newStateByPub = reduce_1.reduceGroupBy(newState, 'pub');
        // Case 1: revokes_on change REVERT
        const byRevokesOn = reduce_1.reduceGroupBy(records.filter(e => e.revokes_on), "revokes_on");
        await Promise.all(underscore_1.Underscore.keys(byRevokesOn)
            .map(async (revokesOn) => {
            const pubkeys = byRevokesOn[revokesOn].map(e => e.pub);
            // 1. Remove the existing value
            await this.removeAllKeysFromRevokesOn(pint_1.pint(revokesOn), pubkeys);
            // 2. Put back the old one if it exists
            const reduced = pubkeys
                .filter(p => newStateByPub[p])
                .map(p => newStateByPub[p])
                .map(indexer_1.reduce)
                .filter(r => r.revokes_on);
            for (const r of reduced) {
                await this.addAllKeysToRevokesOn(r.revokes_on, [r.pub]);
            }
        }));
        // Case 2: revocation REVERT
        const values = underscore_1.Underscore.values(newStateByPub).map(entries => indexer_1.reduce(entries));
        const byExpiredOn = reduce_1.reduceGroupBy(values, "revoked_on");
        await Promise.all(underscore_1.Underscore.keys(byExpiredOn)
            .map(async (revokesOn) => this.addAllKeysToRevokesOn(pint_1.pint(revokesOn), byExpiredOn[revokesOn].map(e => e.pub))));
    }
    async addAllKeysToRevokesOn(revokesOn, pubkeys) {
        const key = LevelMIndexRevokesOnIndexer.trimKey(revokesOn);
        let entry = await this.getOrNull(key);
        if (!entry) {
            entry = [];
        }
        for (const pub of pubkeys) {
            entry.push(pub);
        }
        await this.put(key, entry);
    }
    async removeAllKeysFromRevokesOn(revokesOn, pubkeys) {
        // We remove the "revokes_on" indexed values
        const key = LevelMIndexRevokesOnIndexer.trimKey(revokesOn);
        const entry = await this.get(key);
        for (const pub of pubkeys) {
            if (entry.includes(pub)) {
                entry.splice(entry.indexOf(pub), 1);
            }
        }
        if (entry.length) {
            // Some revocations left
            await this.put(key, entry); // TODO: test this, can occur, probably not covered
        }
        else {
            // No more revocations left
            await this.del(key);
        }
    }
    async findRevokesOnLte(revokesOn) {
        return (await this.findAllValues({ lte: LevelMIndexRevokesOnIndexer.trimKey(revokesOn) })).reduce(reduce_1.reduceConcat, []);
    }
    static trimKey(revokesOn) {
        return String(revokesOn).padStart(10, '0');
    }
}
exports.LevelMIndexRevokesOnIndexer = LevelMIndexRevokesOnIndexer;
//# sourceMappingURL=LevelMIndexRevokesOnIndexer.js.map