"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const AbstractSQLite_1 = require("./AbstractSQLite");
const SandBox_1 = require("./SandBox");
const IdentityDTO_1 = require("../../dto/IdentityDTO");
const constants = require('../../constants');
class DBIdentity {
    constructor() {
        this.certs = [];
        this.signed = [];
    }
    clone() {
        return DBIdentity.copyFromExisting(this);
    }
    getTargetHash() {
        return IdentityDTO_1.IdentityDTO.getTargetHash({
            pub: this.pubkey,
            created_on: this.buid,
            uid: this.uid
        });
    }
    json() {
        const others = [];
        this.certs.forEach((cert) => {
            others.push({
                "pubkey": cert.from,
                "meta": {
                    "block_number": cert.block_number,
                    "block_hash": cert.block_hash
                },
                "uids": cert.uids,
                "isMember": cert.isMember,
                "wasMember": cert.wasMember,
                "signature": cert.sig
            });
        });
        const uids = [{
                "uid": this.uid,
                "meta": {
                    "timestamp": this.buid
                },
                "revoked": this.revoked,
                "revoked_on": parseInt(String(this.revoked_on)),
                "revocation_sig": this.revocation_sig,
                "self": this.sig,
                "others": others
            }];
        const signed = [];
        this.signed.forEach((cert) => {
            signed.push({
                "uid": cert.idty.uid,
                "pubkey": cert.idty.pubkey,
                "meta": {
                    "timestamp": cert.idty.buid
                },
                "cert_time": {
                    "block": cert.block_number,
                    "block_hash": cert.block_hash
                },
                "isMember": cert.idty.member,
                "wasMember": cert.idty.wasMember,
                "signature": cert.sig
            });
        });
        return {
            "pubkey": this.pubkey,
            "uids": uids,
            "signed": signed
        };
    }
    static copyFromExisting(idty) {
        return new ExistingDBIdentity(idty);
    }
}
exports.DBIdentity = DBIdentity;
class NewDBIdentity extends DBIdentity {
    constructor(pubkey, sig, buid, uid, hash) {
        super();
        this.pubkey = pubkey;
        this.sig = sig;
        this.buid = buid;
        this.uid = uid;
        this.hash = hash;
        this.revoked = false;
        this.currentMSN = null;
        this.currentINN = null;
        this.member = false;
        this.kick = false;
        this.leaving = false;
        this.wasMember = false;
        this.revocation_sig = null;
        this.written = false;
        this.wotb_id = null;
        this.revoked_on = null;
        this.expires_on = 0;
    }
}
exports.NewDBIdentity = NewDBIdentity;
class ExistingDBIdentity extends DBIdentity {
    constructor(idty) {
        super();
        this.pubkey = idty.pubkey;
        this.sig = idty.sig;
        this.buid = idty.buid;
        this.uid = idty.uid;
        this.hash = idty.hash;
        this.revoked = idty.revoked;
        this.currentMSN = idty.currentMSN;
        this.currentINN = idty.currentINN;
        this.member = idty.member;
        this.kick = idty.kick;
        this.leaving = idty.leaving;
        this.wasMember = idty.wasMember;
        this.revocation_sig = idty.revocation_sig;
        this.written = idty.written;
        this.wotb_id = idty.wotb_id;
        this.revoked_on = idty.revoked_on;
        this.expires_on = idty.expires_on;
        this.certs = idty.certs || [];
        this.signed = idty.signed || [];
    }
}
exports.ExistingDBIdentity = ExistingDBIdentity;
class IdentityDAL extends AbstractSQLite_1.AbstractSQLite {
    constructor(driver) {
        super(driver, 'idty', 
        // PK fields
        ['pubkey', 'uid', 'hash'], 
        // Fields
        [
            'revoked',
            'revoked_on',
            'revocation_sig',
            'currentMSN',
            'currentINN',
            'buid',
            'member',
            'kick',
            'leaving',
            'wasMember',
            'pubkey',
            'uid',
            'sig',
            'hash',
            'written',
            'wotb_id',
            'expired',
            'expires_on',
            'removed'
        ], 
        // Arrays
        [], 
        // Booleans
        ['revoked', 'member', 'kick', 'leaving', 'wasMember', 'written', 'removed'], 
        // BigIntegers
        [], 
        // Transient
        ['certsCount', 'ref_block']);
        this.sandbox = new SandBox_1.SandBox(constants.SANDBOX_SIZE_IDENTITIES, this.getSandboxIdentities.bind(this), (compared, reference) => {
            if (compared.certsCount < reference.certsCount) {
                return -1;
            }
            else if (compared.certsCount > reference.certsCount) {
                return 1;
            }
            else if (compared.ref_block < reference.ref_block) {
                return -1;
            }
            else if (compared.ref_block > reference.ref_block) {
                return 1;
            }
            else {
                return 0;
            }
        });
    }
    async init() {
        await this.exec('BEGIN;' +
            'CREATE TABLE IF NOT EXISTS ' + this.table + ' (' +
            'revoked BOOLEAN NOT NULL,' +
            'currentMSN INTEGER NULL,' +
            'currentINN INTEGER NULL,' +
            'buid VARCHAR(100) NOT NULL,' +
            'member BOOLEAN NOT NULL,' +
            'kick BOOLEAN NOT NULL,' +
            'leaving BOOLEAN NULL,' +
            'wasMember BOOLEAN NOT NULL,' +
            'pubkey VARCHAR(50) NOT NULL,' +
            'uid VARCHAR(255) NOT NULL,' +
            'sig VARCHAR(100) NOT NULL,' +
            'revocation_sig VARCHAR(100) NULL,' +
            'hash VARCHAR(64) NOT NULL,' +
            'written BOOLEAN NULL,' +
            'wotb_id INTEGER NULL,' +
            'expires_on INTEGER NULL,' +
            'PRIMARY KEY (pubkey,uid,hash)' +
            ');' +
            'CREATE INDEX IF NOT EXISTS idx_idty_pubkey ON idty (pubkey);' +
            'CREATE INDEX IF NOT EXISTS idx_idty_uid ON idty (uid);' +
            'CREATE INDEX IF NOT EXISTS idx_idty_kick ON idty (kick);' +
            'CREATE INDEX IF NOT EXISTS idx_idty_member ON idty (member);' +
            'CREATE INDEX IF NOT EXISTS idx_idty_wasMember ON idty (wasMember);' +
            'CREATE INDEX IF NOT EXISTS idx_idty_hash ON idty (hash);' +
            'CREATE INDEX IF NOT EXISTS idx_idty_written ON idty (written);' +
            'CREATE INDEX IF NOT EXISTS idx_idty_currentMSN ON idty (currentMSN);' +
            'CREATE INDEX IF NOT EXISTS idx_idty_currentINN ON idty (currentINN);' +
            'COMMIT;');
    }
    revokeIdentity(pubkey) {
        return this.exec('DELETE FROM ' + this.table + ' WHERE pubkey = \'' + pubkey + '\'');
    }
    removeUnWrittenWithPubkey(pubkey) {
        return this.sqlRemoveWhere({
            pubkey: pubkey,
            written: false
        });
    }
    removeUnWrittenWithUID(uid) {
        return this.sqlRemoveWhere({
            uid: uid,
            written: false
        });
    }
    setRevoked(pubkey) {
        return this.query('UPDATE ' + this.table + ' SET revoked = ? WHERE pubkey = ?', [true, pubkey]);
    }
    getByHash(hash) {
        return this.sqlFindOne({
            hash: hash
        });
    }
    saveIdentity(idty) {
        return this.saveEntity(idty);
    }
    async deleteByHash(hash) {
        await this.exec('UPDATE ' + this.table + ' SET removed = 1 where hash = \'' + hash + '\'');
    }
    getToRevoke() {
        return this.sqlFind({
            revocation_sig: { $null: false },
            revoked: false,
            wasMember: true
        });
    }
    getPendingIdentities() {
        return this.sqlFind({
            revocation_sig: { $null: true },
            revoked: false
        });
    }
    searchThoseMatching(search) {
        return this.sqlFindLikeAny({
            pubkey: "%" + search + "%",
            uid: "%" + search + "%"
        });
    }
    async trimExpiredIdentities(medianTime) {
        await this.exec('DELETE FROM ' + this.table + ' WHERE (expires_on IS NULL AND revocation_sig IS NULL) OR expires_on < ' + medianTime);
    }
    /**************************
     * SANDBOX STUFF
     */
    getSandboxIdentities() {
        return this.query('SELECT * FROM sandbox_idty LIMIT ' + (this.sandbox.maxSize), []);
    }
    getSandboxRoom() {
        return this.sandbox.getSandboxRoom();
    }
    setSandboxSize(maxSize) {
        this.sandbox.maxSize = maxSize;
    }
}
exports.IdentityDAL = IdentityDAL;
//# sourceMappingURL=IdentityDAL.js.map