"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const common_1 = require("../common");
const keyring_1 = require("../common-libs/crypto/keyring");
class InputDTO {
    constructor(amount, base, type, identifier, pos, raw) {
        this.amount = amount;
        this.base = base;
        this.type = type;
        this.identifier = identifier;
        this.pos = pos;
        this.raw = raw;
    }
}
exports.InputDTO = InputDTO;
class OutputDTO {
    constructor(amount, base, conditions, raw) {
        this.amount = amount;
        this.base = base;
        this.conditions = conditions;
        this.raw = raw;
    }
}
exports.OutputDTO = OutputDTO;
class TxSignatureResultImpl {
    constructor(issuers) {
        this.sigs = issuers.map(k => {
            return { k, ok: false };
        });
    }
    get allMatching() {
        return this.sigs.reduce((ok, s) => ok && s.ok, true);
    }
}
exports.TxSignatureResultImpl = TxSignatureResultImpl;
class TransactionDTO {
    constructor(version, currency, locktime, hash, blockstamp, blockstampTime, issuers, inputs, outputs, unlocks, signatures, comment) {
        this.version = version;
        this.currency = currency;
        this.locktime = locktime;
        this.hash = hash;
        this.blockstamp = blockstamp;
        this.blockstampTime = blockstampTime;
        this.issuers = issuers;
        this.inputs = inputs;
        this.outputs = outputs;
        this.unlocks = unlocks;
        this.signatures = signatures;
        this.comment = comment;
        // Compute the hash if not given
        if (!hash) {
            this.hash = this.getHash();
        }
    }
    clone() {
        return TransactionDTO.fromJSONObject(this);
    }
    get signature() {
        return this.signatures[0];
    }
    get output_amount() {
        return this.outputs.reduce((maxBase, output) => Math.max(maxBase, parseInt(output.split(':')[0])), 0);
    }
    get output_base() {
        return this.outputs.reduce((sum, output) => sum + parseInt(output.split(':')[1]), 0);
    }
    get blockNumber() {
        return parseInt(this.blockstamp);
    }
    get block_hash() {
        return this.blockstamp.split('-')[1];
    }
    getLen() {
        return 2 // header + blockstamp
            + this.issuers.length * 2 // issuers + signatures
            + this.inputs.length * 2 // inputs + unlocks
            + (this.comment ? 1 : 0)
            + this.outputs.length;
    }
    getHash() {
        const raw = TransactionDTO.toRAW(this);
        return common_1.hashf(raw);
    }
    getRawTxNoSig() {
        return TransactionDTO.toRAW(this, true);
    }
    inputsAsObjects() {
        return this.inputs.map(input => {
            const [amount, base, type, identifier, pos] = input.split(':');
            return new InputDTO(parseInt(amount), parseInt(base), type, identifier, parseInt(pos), input);
        });
    }
    outputsAsObjects() {
        return this.outputs.map(output => {
            const [amount, base, conditions] = output.split(':');
            return new OutputDTO(parseInt(amount), parseInt(base), conditions, output);
        });
    }
    outputsAsRecipients() {
        return this.outputs.map((out) => {
            const recipent = out.match('SIG\\((.*)\\)');
            return (recipent && recipent[1]) || 'UNKNOWN';
        });
    }
    getRaw() {
        let raw = "";
        raw += "Version: " + (this.version) + "\n";
        raw += "Type: Transaction\n";
        raw += "Currency: " + this.currency + "\n";
        raw += "Blockstamp: " + this.blockstamp + "\n";
        raw += "Locktime: " + this.locktime + "\n";
        raw += "Issuers:\n";
        (this.issuers || []).forEach((issuer) => {
            raw += issuer + '\n';
        });
        raw += "Inputs:\n";
        this.inputs.forEach((input) => {
            raw += input + '\n';
        });
        raw += "Unlocks:\n";
        this.unlocks.forEach((unlock) => {
            raw += unlock + '\n';
        });
        raw += "Outputs:\n";
        this.outputs.forEach((output) => {
            raw += output + '\n';
        });
        raw += "Comment: " + (this.comment || "") + "\n";
        this.signatures.forEach((signature) => {
            raw += signature + '\n';
        });
        return raw;
    }
    getCompactVersion() {
        let issuers = this.issuers;
        let raw = ["TX", this.version, issuers.length, this.inputs.length, this.unlocks.length, this.outputs.length, this.comment ? 1 : 0, this.locktime || 0].join(':') + '\n';
        raw += this.blockstamp + "\n";
        (issuers || []).forEach((issuer) => {
            raw += issuer + '\n';
        });
        (this.inputs || []).forEach((input) => {
            raw += input + '\n';
        });
        (this.unlocks || []).forEach((input) => {
            raw += input + '\n';
        });
        (this.outputs || []).forEach((output) => {
            raw += output + '\n';
        });
        if (this.comment)
            raw += this.comment + '\n';
        (this.signatures || []).forEach((signature) => {
            raw += signature + '\n';
        });
        return raw;
    }
    computeAllHashes() {
        this.hash = common_1.hashf(this.getRaw()).toUpperCase();
    }
    json() {
        return {
            'version': this.version,
            'currency': this.currency,
            'issuers': this.issuers,
            'inputs': this.inputs,
            'unlocks': this.unlocks,
            'outputs': this.outputs,
            'comment': this.comment,
            'locktime': this.locktime,
            'blockstamp': this.blockstamp,
            'blockstampTime': this.blockstampTime,
            'signatures': this.signatures,
            'raw': this.getRaw(),
            'hash': this.hash
        };
    }
    getTransactionSigResult() {
        const sigResult = new TxSignatureResultImpl(this.issuers.slice());
        let i = 0;
        const raw = this.getRawTxNoSig();
        let matching = true;
        while (matching && i < this.signatures.length) {
            const sig = this.signatures[i];
            const pub = this.issuers[i];
            sigResult.sigs[i].ok = matching = keyring_1.verify(raw, sig, pub);
            i++;
        }
        return sigResult;
    }
    checkSignatures() {
        return this.getTransactionSigResult().allMatching;
    }
    static fromJSONObject(obj, currency = "") {
        return new TransactionDTO(obj.version || 10, currency || obj.currency || "", obj.locktime || 0, obj.hash || "", obj.blockstamp || "", obj.blockstampTime || 0, obj.issuers || [], obj.inputs || [], obj.outputs || [], obj.unlocks || [], obj.signatures || [], obj.comment || "");
    }
    static toRAW(json, noSig = false) {
        let raw = "";
        raw += "Version: " + (json.version) + "\n";
        raw += "Type: Transaction\n";
        raw += "Currency: " + json.currency + "\n";
        raw += "Blockstamp: " + json.blockstamp + "\n";
        raw += "Locktime: " + json.locktime + "\n";
        raw += "Issuers:\n";
        (json.issuers || []).forEach((issuer) => {
            raw += issuer + '\n';
        });
        raw += "Inputs:\n";
        (json.inputs || []).forEach((input) => {
            raw += input + '\n';
        });
        raw += "Unlocks:\n";
        (json.unlocks || []).forEach((unlock) => {
            raw += unlock + '\n';
        });
        raw += "Outputs:\n";
        (json.outputs || []).forEach((output) => {
            raw += output + '\n';
        });
        raw += "Comment: " + (json.comment || "") + "\n";
        if (!noSig) {
            (json.signatures || []).forEach((signature) => {
                raw += signature + '\n';
            });
        }
        return raw;
    }
    static outputObj2Str(o) {
        return [o.amount, o.base, o.conditions].join(':');
    }
    static inputObj2Str(i) {
        return [i.amount, i.base, i.type, i.identifier, i.pos].join(':');
    }
    static outputStr2Obj(outputStr) {
        const sp = outputStr.split(':');
        return {
            amount: parseInt(sp[0]),
            base: parseInt(sp[1]),
            conditions: sp[2],
            raw: outputStr
        };
    }
    static inputStr2Obj(inputStr) {
        const sp = inputStr.split(':');
        return {
            amount: sp[0],
            base: sp[1],
            type: sp[2],
            identifier: sp[3],
            pos: parseInt(sp[4]),
            raw: inputStr
        };
    }
    static unlock2params(unlock) {
        const match = unlock.match(/^\d+:(.*)$/);
        if (match) {
            return match[1].split(' ');
        }
        return [];
    }
    static mock() {
        return new TransactionDTO(1, "", 0, "", "", 0, [], [], [], [], [], "");
    }
}
exports.TransactionDTO = TransactionDTO;
//# sourceMappingURL=TransactionDTO.js.map