"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const stream = require("stream");
const BlockDTO_1 = require("../dto/BlockDTO");
const RevocationDTO_1 = require("../dto/RevocationDTO");
const IdentityDTO_1 = require("../dto/IdentityDTO");
const CertificationDTO_1 = require("../dto/CertificationDTO");
const MembershipDTO_1 = require("../dto/MembershipDTO");
const TransactionDTO_1 = require("../dto/TransactionDTO");
const PeerDTO_1 = require("../dto/PeerDTO");
const constants_1 = require("../common-libs/constants");
const request = require('request');
const constants = require('../../lib/constants');
const logger = require('../logger').NewLogger('multicaster');
const WITH_ISOLATION = true;
class Multicaster extends stream.Transform {
    constructor(conf = null, timeout = 0) {
        super({ objectMode: true });
        this.conf = conf;
        this.timeout = timeout;
        this.on('identity', (data, peers) => this.idtyForward(data, peers));
        this.on('cert', (data, peers) => this.certForward(data, peers));
        this.on('revocation', (data, peers) => this.revocationForward(data, peers));
        this.on('block', (data, peers) => this.blockForward(data, peers));
        this.on('transaction', (data, peers) => this.txForward(data, peers));
        this.on('peer', (data, peers) => this.peerForward(data, peers));
        this.on('membership', (data, peers) => this.msForward(data, peers));
    }
    async blockForward(doc, peers) {
        return this.forward({
            transform: (b) => BlockDTO_1.BlockDTO.fromJSONObject(b),
            type: 'Block',
            uri: '/blockchain/block',
            getObj: (block) => {
                return {
                    "block": block.getRawSigned()
                };
            },
            getDocID: (block) => 'block#' + block.number
        })(doc, peers);
    }
    async idtyForward(doc, peers) {
        return this.forward({
            transform: (obj) => IdentityDTO_1.IdentityDTO.fromJSONObject(obj),
            type: 'Identity',
            uri: '/wot/add',
            getObj: (idty) => {
                return {
                    "identity": idty.getRawSigned()
                };
            },
            getDocID: (idty) => 'with ' + (idty.certs || []).length + ' certs'
        })(doc, peers);
    }
    async certForward(doc, peers) {
        return this.forward({
            transform: (obj) => CertificationDTO_1.CertificationDTO.fromJSONObject(obj),
            type: 'Cert',
            uri: '/wot/certify',
            getObj: (cert) => {
                return {
                    "cert": cert.getRawSigned()
                };
            },
            getDocID: (idty) => 'with ' + (idty.certs || []).length + ' certs'
        })(doc, peers);
    }
    async revocationForward(doc, peers) {
        return this.forward({
            transform: (json) => RevocationDTO_1.RevocationDTO.fromJSONObject(json),
            type: 'Revocation',
            uri: '/wot/revoke',
            getObj: (revocation) => {
                return {
                    "revocation": revocation.getRaw()
                };
            }
        })(doc, peers);
    }
    async txForward(doc, peers) {
        return this.forward({
            transform: (obj) => TransactionDTO_1.TransactionDTO.fromJSONObject(obj),
            type: 'Transaction',
            uri: '/tx/process',
            getObj: (transaction) => {
                return {
                    "transaction": transaction.getRaw(),
                    "signature": transaction.signature
                };
            }
        })(doc, peers);
    }
    async peerForward(doc, peers) {
        return this.forward({
            type: 'Peer',
            uri: '/network/peering/peers',
            transform: (obj) => PeerDTO_1.PeerDTO.fromJSONObject(obj),
            getObj: (peering) => {
                return {
                    peer: peering.getRawSigned()
                };
            },
            getDocID: (doc) => doc.keyID() + '#' + doc.blockNumber(),
            withIsolation: WITH_ISOLATION,
            onError: (resJSON, peering, to) => {
                if (resJSON.ucode !== undefined && resJSON.ucode !== constants_1.CommonConstants.ERRORS.NEWER_PEER_DOCUMENT_AVAILABLE.uerr.ucode) {
                    if (resJSON.ucode == constants_1.CommonConstants.ERRORS.DOCUMENT_BEING_TREATED.uerr.ucode || resJSON.ucode == constants.ERRORS.PEER_DOCUMENT_ALREADY_KNOWN.uerr.ucode) {
                        return Promise.resolve();
                    }
                    else {
                        throw Error(resJSON.message);
                    }
                }
                else {
                    // Handle possibly outdated peering document
                    const sentPeer = PeerDTO_1.PeerDTO.fromJSONObject(peering);
                    if (PeerDTO_1.PeerDTO.blockNumber(resJSON.peer.block) > sentPeer.blockNumber()) {
                        this.push({ outdated: true, peer: resJSON.peer });
                        logger.warn('Outdated peer document (%s) sent to %s', sentPeer.keyID() + '#' + sentPeer.blockNumber(), to);
                    }
                    return Promise.resolve();
                }
            }
        })(doc, peers);
    }
    async msForward(doc, peers) {
        return this.forward({
            transform: (obj) => MembershipDTO_1.MembershipDTO.fromJSONObject(obj),
            type: 'Membership',
            uri: '/blockchain/membership',
            getObj: (membership) => {
                return {
                    "membership": membership.getRaw(),
                    "signature": membership.signature
                };
            }
        })(doc, peers);
    }
    _write(obj, enc, done) {
        this.emit(obj.type, obj.obj, obj.peers);
        done();
    }
    sendBlock(toPeer, block) {
        return this.blockForward(block, [toPeer]);
    }
    sendPeering(toPeer, peer) {
        return this.peerForward(peer, [toPeer]);
    }
    forward(params) {
        return async (doc, peers) => {
            try {
                if (!params.withIsolation || !(this.conf && this.conf.isolate)) {
                    let theDoc = params.transform ? params.transform(doc) : doc;
                    if (params.getDocID) {
                        logger.info('POST %s %s to %s peers', params.type, params.getDocID(theDoc), peers.length);
                    }
                    else {
                        logger.info('POST %s to %s peers', params.type, peers.length);
                    }
                    // Parallel treatment for superfast propagation
                    await Promise.all(peers.map(async (p) => {
                        let peer = PeerDTO_1.PeerDTO.fromJSONObject(p);
                        const namedURL = peer.getNamedURL();
                        try {
                            await this.post(peer, params.uri, params.getObj(theDoc));
                        }
                        catch (e) {
                            if (params.onError) {
                                try {
                                    const json = JSON.parse(e.body);
                                    await params.onError(json, doc, namedURL);
                                }
                                catch (ex) {
                                    logger.warn('Could not reach %s, reason: %s', namedURL, (ex && ex.message || ex));
                                }
                            }
                        }
                    }));
                }
                else {
                    logger.debug('[ISOLATE] Prevent --> new Peer to be sent to %s peer(s)', peers.length);
                }
            }
            catch (err) {
                logger.error(err);
            }
        };
    }
    post(peer, uri, data) {
        if (!peer.isReachable()) {
            return Promise.resolve();
        }
        return new Promise((resolve, reject) => {
            const postReq = request.post({
                "uri": protocol(peer.getPort()) + '://' + peer.getURL() + uri,
                "timeout": this.timeout || constants.NETWORK.DEFAULT_TIMEOUT
            }, (err, res) => {
                if (err) {
                    this.push({ unreachable: true, peer: { pubkey: peer.pubkey } });
                    logger.warn(err.message || err);
                }
                if (res && res.statusCode != 200) {
                    return reject(res);
                }
                resolve(res);
            });
            postReq.form(data);
        });
    }
}
exports.Multicaster = Multicaster;
function protocol(port) {
    return port == 443 ? 'https' : 'http';
}
//# sourceMappingURL=multicaster.js.map