"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const fs = require("fs");
const SQLiteDriver_1 = require("../dal/drivers/SQLiteDriver");
const CFSCore_1 = require("../dal/fileDALs/CFSCore");
const wot_1 = require("../wot");
const programOptions_1 = require("../common-libs/programOptions");
const LevelDBDriver_1 = require("../dal/drivers/LevelDBDriver");
const opts = programOptions_1.cliprogram;
const qfs = require('q-io/fs');
const DEFAULT_DOMAIN = "duniter_default";
const DEFAULT_HOME = (process.platform == 'win32' ? process.env.USERPROFILE : process.env.HOME) + '/.config/duniter/';
const getLogsPath = (profile, directory = null) => path.join(getHomePath(profile, directory), 'duniter.log');
const getHomePath = (profile, directory = null) => path.normalize(getUserHome(directory) + '/') + getDomain(profile);
const getUserHome = (directory = null) => (directory || DEFAULT_HOME);
const getDomain = (profile = null) => (profile || DEFAULT_DOMAIN);
class QioFileSystem {
    constructor(qio, isMemory = false) {
        this.qio = qio;
        this.isMemory = isMemory;
    }
    async fsExists(file) {
        return this.qio.exists(file);
    }
    async fsReadFile(file) {
        return this.qio.read(file);
    }
    async fsUnlink(file) {
        return this.qio.remove(file);
    }
    async fsList(dir) {
        if (!(await this.qio.exists(dir))) {
            return [];
        }
        return this.qio.list(dir);
    }
    fsWrite(file, content) {
        return this.qio.write(file, content);
    }
    async fsStreamTo(file, iterator) {
        if (this.isMemory) {
            for (const line of iterator) {
                await this.qio.append(file, line);
            }
        }
        else {
            // Use NodeJS streams for faster writing
            let wstream = fs.createWriteStream(file);
            await new Promise(async (res, rej) => {
                // When done, return
                wstream.on('close', (err) => {
                    if (err)
                        return rej(err);
                    res();
                });
                // Write each line
                for (const line of iterator) {
                    wstream.write(line + "\n");
                }
                // End the writing
                wstream.end();
            });
        }
    }
    fsMakeDirectory(dir) {
        return this.qio.makeTree(dir);
    }
    async fsRemoveTree(dir) {
        return this.qio.removeTree(dir);
    }
}
exports.RealFS = () => {
    return new QioFileSystem(qfs);
};
exports.MemFS = (initialTree = {}) => {
    return new QioFileSystem(require('q-io/fs-mock')(initialTree), true);
};
exports.Directory = {
    DATA_FILES: ['mindex.db', 'c_mindex.db', 'iindex.db', 'cindex.db', 'sindex.db', 'wallet.db', 'dividend.db', 'txs.db', 'peers.db'],
    WW_FILES: ['wotwizard-export_0.db', 'wotwizard-export.db'],
    DATA_DIRS: ['data'],
    INSTANCE_NAME: getDomain(opts.mdb),
    INSTANCE_HOME: getHomePath(opts.mdb, opts.home),
    GET_FILE_PATH: (fileSubPath, home = '') => path.join(home || exports.Directory.INSTANCE_HOME, fileSubPath),
    INSTANCE_HOMELOG_FILE: getLogsPath(opts.mdb, opts.home),
    DUNITER_DB_NAME: 'duniter',
    LOKI_DB_DIR: 'loki',
    WOTB_FILE: 'wotb.bin',
    getHome: (profile = null, directory = null) => getHomePath(profile, directory),
    getHomeDB: async (isMemory, dbName, home = '') => {
        // Memory
        if (isMemory) {
            return new SQLiteDriver_1.SQLiteDriver(':memory:');
        }
        // Or file
        const sqlitePath = exports.Directory.GET_FILE_PATH(dbName, home);
        return new SQLiteDriver_1.SQLiteDriver(sqlitePath);
    },
    getHomeLevelDB: async (isMemory, dbName, home = '') => {
        // Memory
        if (isMemory) {
            return LevelDBDriver_1.LevelDBDriver.newMemoryInstance();
        }
        // Or file
        const levelDbRootPath = path.join(home, 'data', 'leveldb');
        await exports.RealFS().fsMakeDirectory(levelDbRootPath);
        const levelDBPath = path.join(levelDbRootPath, dbName);
        return LevelDBDriver_1.LevelDBDriver.newFileInstance(levelDBPath);
    },
    getHomeFS: async (isMemory, theHome, makeTree = true) => {
        const home = theHome || exports.Directory.getHome();
        const params = {
            home: home,
            fs: isMemory ? exports.MemFS() : exports.RealFS()
        };
        if (makeTree) {
            await params.fs.fsMakeDirectory(home);
        }
        return params;
    },
    getHomeParams: async (isMemory, theHome) => {
        const params = await exports.Directory.getHomeFS(isMemory, theHome);
        const home = params.home;
        let dbf;
        let wotbf;
        if (isMemory) {
            // Memory DB
            dbf = () => new SQLiteDriver_1.SQLiteDriver(':memory:');
            wotbf = () => wot_1.WoTBObject.memoryInstance();
        }
        else {
            // File DB
            const sqlitePath = path.join(home, exports.Directory.DUNITER_DB_NAME + '.db');
            dbf = () => new SQLiteDriver_1.SQLiteDriver(sqlitePath);
            const wotbFilePath = path.join(home, exports.Directory.WOTB_FILE);
            let existsFile = await qfs.exists(wotbFilePath);
            if (!existsFile) {
                fs.closeSync(fs.openSync(wotbFilePath, 'w'));
            }
            wotbf = () => wot_1.WoTBObject.fileInstance(wotbFilePath);
        }
        return {
            home: params.home,
            fs: params.fs,
            dbf,
            wotbf
        };
    },
    createHomeIfNotExists: async (fileSystem, theHome) => {
        const fsHandler = new CFSCore_1.CFSCore(theHome, fileSystem);
        return fsHandler.makeTree('');
    }
};
//# sourceMappingURL=directory.js.map