"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const stream = require("stream");
const network_1 = require("./lib/network");
const upnp_1 = require("./lib/upnp");
const constants_1 = require("./lib/constants");
const limiter_1 = require("./lib/limiter");
const PeerDTO_1 = require("../../lib/dto/PeerDTO");
const underscore_1 = require("../../lib/common-libs/underscore");
const bma_1 = require("./lib/bma");
const Q = require('q');
const rp = require('request-promise');
const async = require('async');
const dtos = require('./lib/dtos');
const inquirer = require('inquirer');
let networkWizardDone = false;
exports.BmaDependency = {
    duniter: {
        cliOptions: [
            { value: '--upnp', desc: 'Use UPnP to open remote port.' },
            { value: '--noupnp', desc: 'Do not use UPnP to open remote port.' },
            { value: '--bma', desc: 'Enables BMA API and its crawlers.' },
            { value: '--nobma', desc: 'Disables BMA API and its crawlers.' },
            { value: '--bma-with-crawler', desc: 'Enables BMA Crawler.' },
            { value: '--bma-without-crawler', desc: 'Disable BMA Crawler.' },
            { value: '-p, --port <port>', desc: 'Port to listen for requests', parser: (val) => parseInt(val) },
            { value: '--ipv4 <address>', desc: 'IPv4 interface to listen for requests' },
            { value: '--ipv6 <address>', desc: 'IPv6 interface to listen for requests' },
            { value: '--remoteh <host>', desc: 'Remote interface others may use to contact this node' },
            { value: '--remote4 <host>', desc: 'Remote interface for IPv4 access' },
            { value: '--remote6 <host>', desc: 'Remote interface for IPv6 access' },
            { value: '--remotep <port>', desc: 'Remote port others may use to contact this node' },
        ],
        wizard: {
            'network': async (conf, program, logger) => {
                await Q.nbind(networkConfiguration, null, conf, logger)();
                conf.nobma = false;
                networkWizardDone = true;
            },
            'network-reconfigure': async (conf, program, logger) => {
                if (!networkWizardDone) {
                    // This step can only be launched lonely
                    await Q.nbind(networkReconfiguration, null)(conf, program.autoconf, logger, program.noupnp);
                }
            }
        },
        config: {
            onLoading: async (conf, program, logger) => {
                // If the usage of BMA hasn't been defined yet
                if (conf.nobma === undefined) {
                    // Do we have an existing BMA conf?
                    if (conf.port !== undefined
                        || conf.ipv4 !== undefined
                        || conf.ipv6 !== undefined
                        || conf.remoteport !== undefined
                        || conf.remotehost !== undefined
                        || conf.remoteipv4 !== undefined
                        || conf.remoteipv6 !== undefined) {
                        conf.nobma = false;
                    }
                    else {
                        conf.nobma = true;
                    }
                }
                // If bmaWithCrawler hasn't been defined yet
                if (conf.bmaWithCrawler === undefined) {
                    conf.bmaWithCrawler = false;
                }
                if (program.port !== undefined)
                    conf.port = parseInt(program.port);
                if (program.ipv4 !== undefined)
                    conf.ipv4 = program.ipv4;
                if (program.ipv6 !== undefined)
                    conf.ipv6 = program.ipv6;
                if (program.remoteh !== undefined)
                    conf.remotehost = program.remoteh;
                if (program.remote4 !== undefined)
                    conf.remoteipv4 = program.remote4;
                if (program.remote6 !== undefined)
                    conf.remoteipv6 = program.remote6;
                if (program.remotep !== undefined)
                    conf.remoteport = parseInt(program.remotep);
                if (program.bma !== undefined)
                    conf.nobma = false;
                if (program.nobma !== undefined)
                    conf.nobma = true;
                if (program.bmaWithCrawler !== undefined)
                    conf.bmaWithCrawler = true;
                if (program.bmaWithoutCrawler !== undefined)
                    conf.bmaWithCrawler = false;
                if (!conf.ipv4)
                    delete conf.ipv4;
                if (!conf.ipv6)
                    delete conf.ipv6;
                if (!conf.remoteipv4)
                    delete conf.remoteipv4;
                if (!conf.remoteipv6)
                    delete conf.remoteipv6;
                // Default remoteipv6: same as local if defined
                if (!conf.remoteipv6 && conf.ipv6) {
                    conf.remoteipv6 = conf.ipv6;
                }
                // Fix #807: default remoteipv4: same as local ipv4 if no removeipv4 is not defined AND no DNS nor IPv6
                if (conf.ipv4 && !(conf.remoteipv4 || conf.remotehost || conf.remoteipv6)) {
                    conf.remoteipv4 = conf.ipv4;
                }
                if (!conf.remoteport && conf.port) {
                    conf.remoteport = conf.port;
                }
                // Network autoconf
                if (program.autoconf) {
                    await Q.nbind(networkReconfiguration, null)(conf, true, logger, program.noupnp);
                }
                // Default value
                if (conf.upnp === undefined || conf.upnp === null) {
                    conf.upnp = true; // Defaults to true
                }
                if (!conf.dos) {
                    conf.dos = { whitelist: ['127.0.0.1'] };
                    conf.dos.maxcount = 50;
                    conf.dos.burst = 20;
                    conf.dos.limit = conf.dos.burst * 2;
                    conf.dos.maxexpiry = 10;
                    conf.dos.checkinterval = 1;
                    conf.dos.trustProxy = true;
                    conf.dos.includeUserAgent = true;
                    conf.dos.errormessage = 'Error';
                    conf.dos.testmode = false;
                    conf.dos.silent = false;
                    conf.dos.silentStart = false;
                    conf.dos.responseStatus = 429;
                }
                // UPnP
                if (program.noupnp === true) {
                    conf.upnp = false;
                }
                if (program.upnp === true) {
                    conf.upnp = true;
                }
            },
            beforeSave: async (conf, program) => {
                if (!conf.ipv4)
                    delete conf.ipv4;
                if (!conf.ipv6)
                    delete conf.ipv6;
                if (!conf.remoteipv4)
                    delete conf.remoteipv4;
                if (!conf.remoteipv6)
                    delete conf.remoteipv6;
                conf.dos.whitelist = underscore_1.Underscore.uniq(conf.dos.whitelist);
            }
        },
        service: {
            input: (server, conf, logger) => {
                // Configuration errors
                if (!conf.nobma) {
                    if (!conf.ipv4 && !conf.ipv6) {
                        throw new Error("BMA: no interface to listen to. Provide ipv4/ipv6 interface or deactivate BMA");
                    }
                    if (!conf.remoteipv4 && !conf.remoteipv6 && !conf.remotehost) {
                        throw new Error('BMA: no interface for remote contact.');
                    }
                    if (!conf.remoteport) {
                        throw new Error('BMA: no port for remote contact.');
                    }
                }
                if (!conf.nobma) {
                    server.addEndpointsDefinitions(() => Promise.resolve(getEndpoint(conf)));
                    server.addWrongEndpointFilter((endpoints) => getWrongEndpoints(endpoints, server.conf.pair.pub));
                }
                return new BMAPI(server, conf, logger);
            }
        },
        methods: {
            noLimit: () => limiter_1.BMALimitation.noLimit(),
            bma: async (server, interfaces = null, httpLogs = false, logger) => bma_1.bma(server, interfaces, httpLogs, logger),
            dtos,
            getMainEndpoint: (conf) => Promise.resolve(getEndpoint(conf))
        }
    }
};
async function getWrongEndpoints(endpoints, selfPubkey) {
    const wrongs = [];
    await Promise.all(endpoints.map(async (theEndpoint) => {
        let remote = PeerDTO_1.PeerDTO.endpoint2host(theEndpoint);
        try {
            // We test only BMA APIs, because other may exist and we cannot judge against them
            if (theEndpoint.startsWith('BASIC_MERKLED_API')) {
                let answer = await rp('http://' + remote + '/network/peering', { json: true });
                if (!answer || answer.pubkey != selfPubkey) {
                    throw Error("Not same pubkey as local instance");
                }
            }
        }
        catch (e) {
            wrongs.push(theEndpoint);
        }
    }));
    return wrongs;
}
class BMAPI extends stream.Transform {
    constructor(server, conf, logger) {
        super({ objectMode: true });
        this.server = server;
        this.conf = conf;
        this.logger = logger;
        this.startService = async () => {
            if (this.conf.nobma) {
                // Disable BMA
                return Promise.resolve();
            }
            this.bmapi = await bma_1.bma(this.server, null, this.conf.httplogs, this.logger);
            await this.bmapi.openConnections();
            /***************
             *    UPnP
             **************/
            if (this.upnpAPI) {
                this.upnpAPI.stopRegular();
            }
            if (this.server.conf.upnp) {
                try {
                    this.upnpAPI = await upnp_1.Upnp(this.server.conf.port, this.server.conf.remoteport, this.logger, this.server.conf);
                    this.upnpAPI.startRegular();
                    const gateway = await this.upnpAPI.findGateway();
                    if (gateway) {
                        if (this.bmapi.getDDOS().params.whitelist.indexOf(gateway) === -1) {
                            this.bmapi.getDDOS().params.whitelist.push(gateway);
                        }
                    }
                }
                catch (e) {
                    this.logger.warn(e);
                }
            }
        };
        this.stopService = async () => {
            if (this.conf.nobma) {
                // Disable BMA
                return Promise.resolve();
            }
            if (this.bmapi) {
                await this.bmapi.closeConnections();
            }
            if (this.upnpAPI) {
                this.upnpAPI.stopRegular();
            }
        };
    }
}
exports.BMAPI = BMAPI;
function getEndpoint(theConf) {
    let endpoint = 'BASIC_MERKLED_API';
    if (theConf.remotehost) {
        if (theConf.remotehost.match(constants_1.BMAConstants.HOST_ONION_REGEX)) {
            endpoint = 'BMATOR';
        }
        endpoint += ' ' + theConf.remotehost;
    }
    if (theConf.remoteipv4) {
        endpoint += ' ' + theConf.remoteipv4;
    }
    if (theConf.remoteipv6) {
        endpoint += ' ' + theConf.remoteipv6;
    }
    if (theConf.remoteport) {
        endpoint += ' ' + theConf.remoteport;
    }
    return endpoint;
}
function networkReconfiguration(conf, autoconf, logger, noupnp, done) {
    async.waterfall([
        upnpResolve.bind(null, noupnp, logger),
        function (upnpSuccess, upnpConf, next) {
            // Default values
            conf.port = conf.port || constants_1.BMAConstants.DEFAULT_PORT;
            conf.remoteport = conf.remoteport || constants_1.BMAConstants.DEFAULT_PORT;
            const localOperations = getLocalNetworkOperations(conf, autoconf);
            const remoteOpertions = getRemoteNetworkOperations(conf, upnpConf.remoteipv4);
            const dnsOperations = getHostnameOperations(conf, logger, autoconf);
            const useUPnPOperations = getUseUPnPOperations(conf, logger, autoconf);
            if (upnpSuccess) {
                underscore_1.Underscore.extend(conf, upnpConf);
                const local = [conf.ipv4, conf.port].join(':');
                const remote = [conf.remoteipv4, conf.remoteport].join(':');
                if (autoconf) {
                    conf.ipv6 = conf.remoteipv6 = network_1.Network.getBestLocalIPv6();
                    logger.info('IPv6: %s', conf.ipv6 || "");
                    logger.info('Local IPv4: %s', local);
                    logger.info('Remote IPv4: %s', remote);
                    // Use proposed local + remote with UPnP binding
                    return async.waterfall(useUPnPOperations
                        .concat(dnsOperations), next);
                }
                choose("UPnP is available: duniter will be bound: \n  from " + local + "\n  to " + remote + "\nKeep this configuration?", true, function () {
                    // Yes: not network changes
                    conf.ipv6 = conf.remoteipv6 = network_1.Network.getBestLocalIPv6();
                    async.waterfall(useUPnPOperations
                        .concat(dnsOperations), next);
                }, function () {
                    // No: want to change
                    async.waterfall(localOperations
                        .concat(remoteOpertions)
                        .concat(useUPnPOperations)
                        .concat(dnsOperations), next);
                });
            }
            else {
                conf.upnp = false;
                if (autoconf) {
                    // Yes: local configuration = remote configuration
                    return async.waterfall(localOperations
                        .concat(getHostnameOperations(conf, logger, autoconf))
                        .concat([function (confDone) {
                            conf.remoteipv4 = conf.ipv4;
                            conf.remoteipv6 = conf.ipv6;
                            conf.remoteport = conf.port;
                            logger.info('Local & Remote IPv4: %s', [conf.ipv4, conf.port].join(':'));
                            logger.info('Local & Remote IPv6: %s', [conf.ipv6, conf.port].join(':'));
                            confDone();
                        }]), next);
                }
                choose("UPnP is *not* available: is this a public server (like a VPS)?", true, function () {
                    // Yes: local configuration = remote configuration
                    async.waterfall(localOperations
                        .concat(getHostnameOperations(conf, logger))
                        .concat([function (confDone) {
                            conf.remoteipv4 = conf.ipv4;
                            conf.remoteipv6 = conf.ipv6;
                            conf.remoteport = conf.port;
                            confDone();
                        }]), next);
                }, function () {
                    // No: must give all details
                    async.waterfall(localOperations
                        .concat(remoteOpertions)
                        .concat(dnsOperations), next);
                });
            }
        }
    ], done);
}
exports.networkReconfiguration = networkReconfiguration;
async function upnpResolve(noupnp, logger, done) {
    try {
        let conf = await network_1.Network.upnpConf(noupnp, logger);
        done(null, true, conf);
    }
    catch (err) {
        done(null, false, {});
    }
}
function networkConfiguration(conf, logger, done) {
    async.waterfall([
        upnpResolve.bind(null, !conf.upnp, logger),
        function (upnpSuccess, upnpConf, next) {
            let operations = getLocalNetworkOperations(conf)
                .concat(getRemoteNetworkOperations(conf, upnpConf.remoteipv4));
            if (upnpSuccess) {
                operations = operations.concat(getUseUPnPOperations(conf, logger));
            }
            async.waterfall(operations.concat(getHostnameOperations(conf, logger, false)), next);
        }
    ], done);
}
function getLocalNetworkOperations(conf, autoconf = false) {
    return [
        function (next) {
            const osInterfaces = network_1.Network.listInterfaces();
            const interfaces = [{ name: "None", value: null }];
            osInterfaces.forEach(function (netInterface) {
                const addresses = netInterface.addresses;
                const filtered = underscore_1.Underscore.where(addresses, { family: 'IPv4' });
                filtered.forEach(function (addr) {
                    interfaces.push({
                        name: [netInterface.name, addr.address].join(' '),
                        value: addr.address
                    });
                });
            });
            if (autoconf) {
                conf.ipv4 = network_1.Network.getBestLocalIPv4();
                return next();
            }
            inquirer.prompt([{
                    type: "list",
                    name: "ipv4",
                    message: "IPv4 interface",
                    default: conf.ipv4,
                    choices: interfaces
                }]).then((answers) => {
                conf.ipv4 = answers.ipv4;
                next();
            });
        },
        function (next) {
            const osInterfaces = network_1.Network.listInterfaces();
            const interfaces = [{ name: "None", value: null }];
            osInterfaces.forEach(function (netInterface) {
                const addresses = netInterface.addresses;
                const filtered = underscore_1.Underscore.where(addresses, { family: 'IPv6' });
                filtered.forEach(function (addr) {
                    let address = addr.address;
                    if (addr.scopeid)
                        address += "%" + netInterface.name;
                    let nameSuffix = "";
                    if (addr.scopeid == 0 && !addr.internal) {
                        nameSuffix = " (Global)";
                    }
                    interfaces.push({
                        name: [netInterface.name, address, nameSuffix].join(' '),
                        internal: addr.internal,
                        scopeid: addr.scopeid,
                        value: address
                    });
                });
            });
            interfaces.sort((addr1, addr2) => {
                if (addr1.value === null)
                    return -1;
                if (addr1.internal && !addr2.internal)
                    return 1;
                if (addr1.scopeid && !addr2.scopeid)
                    return 1;
                return 0;
            });
            if (autoconf || !conf.ipv6) {
                conf.ipv6 = conf.remoteipv6 = network_1.Network.getBestLocalIPv6();
            }
            if (autoconf) {
                return next();
            }
            inquirer.prompt([{
                    type: "list",
                    name: "ipv6",
                    message: "IPv6 interface",
                    default: conf.ipv6,
                    choices: interfaces
                }]).then((answers) => {
                conf.ipv6 = conf.remoteipv6 = answers.ipv6;
                next();
            });
        },
        autoconf ? (done) => {
            conf.port = network_1.Network.getRandomPort(conf);
            done();
        } : async.apply(simpleInteger, "Port", "port", conf)
    ];
}
function getRemoteNetworkOperations(conf, remoteipv4) {
    return [
        function (next) {
            if (!conf.ipv4) {
                conf.remoteipv4 = null;
                return next(null, {});
            }
            const choices = [{ name: "None", value: null }];
            // Local interfaces
            const osInterfaces = network_1.Network.listInterfaces();
            osInterfaces.forEach(function (netInterface) {
                const addresses = netInterface.addresses;
                const filtered = underscore_1.Underscore.where(addresses, { family: 'IPv4' });
                filtered.forEach(function (addr) {
                    choices.push({
                        name: [netInterface.name, addr.address].join(' '),
                        value: addr.address
                    });
                });
            });
            if (conf.remoteipv4) {
                choices.push({ name: conf.remoteipv4, value: conf.remoteipv4 });
            }
            if (remoteipv4 && remoteipv4 != conf.remoteipv4) {
                choices.push({ name: remoteipv4, value: remoteipv4 });
            }
            choices.push({ name: "Enter new one", value: "new" });
            inquirer.prompt([{
                    type: "list",
                    name: "remoteipv4",
                    message: "Remote IPv4",
                    default: conf.remoteipv4 || conf.ipv4 || null,
                    choices: choices,
                    validate: function (input) {
                        return !!(input && input.toString().match(constants_1.BMAConstants.IPV4_REGEXP));
                    }
                }]).then((answers) => {
                if (answers.remoteipv4 == "new") {
                    inquirer.prompt([{
                            type: "input",
                            name: "remoteipv4",
                            message: "Remote IPv4",
                            default: conf.remoteipv4 || conf.ipv4,
                            validate: function (input) {
                                return !!(input && input.toString().match(constants_1.BMAConstants.IPV4_REGEXP));
                            }
                        }]).then((answers) => next(null, answers));
                }
                else {
                    next(null, answers);
                }
            });
        },
        async function (answers, next) {
            conf.remoteipv4 = answers.remoteipv4;
            try {
                if (conf.remoteipv4 || conf.remotehost) {
                    await new Promise((resolve, reject) => {
                        const getPort = async.apply(simpleInteger, "Remote port", "remoteport", conf);
                        getPort((err) => {
                            if (err)
                                return reject(err);
                            resolve();
                        });
                    });
                }
                else if (conf.remoteipv6) {
                    conf.remoteport = conf.port;
                }
                next();
            }
            catch (e) {
                next(e);
            }
        }
    ];
}
function getHostnameOperations(conf, logger, autoconf = false) {
    return [function (next) {
            if (!conf.ipv4) {
                conf.remotehost = null;
                return next();
            }
            if (autoconf) {
                logger.info('DNS: %s', conf.remotehost || 'No');
                return next();
            }
            choose("Does this server has a DNS name?", !!conf.remotehost, function () {
                // Yes
                simpleValue("DNS name:", "remotehost", "", conf, function () { return true; }, next);
            }, function () {
                conf.remotehost = null;
                next();
            });
        }];
}
function getUseUPnPOperations(conf, logger, autoconf = false) {
    return [function (next) {
            if (!conf.ipv4) {
                conf.upnp = false;
                return next();
            }
            if (autoconf) {
                logger.info('UPnP: %s', 'Yes');
                conf.upnp = true;
                return next();
            }
            choose("UPnP is available: use automatic port mapping? (easier)", conf.upnp, function () {
                conf.upnp = true;
                next();
            }, function () {
                conf.upnp = false;
                next();
            });
        }];
}
function choose(question, defaultValue, ifOK, ifNotOK) {
    inquirer.prompt([{
            type: "confirm",
            name: "q",
            message: question,
            default: defaultValue
        }]).then((answer) => {
        answer.q ? ifOK() : ifNotOK();
    });
}
function simpleValue(question, property, defaultValue, conf, validation, done) {
    inquirer.prompt([{
            type: "input",
            name: property,
            message: question,
            default: conf[property],
            validate: validation
        }]).then((answers) => {
        conf[property] = answers[property];
        done();
    });
}
function simpleInteger(question, property, conf, done) {
    simpleValue(question, property, conf[property], conf, function (input) {
        return input && input.toString().match(/^[0-9]+$/) ? true : false;
    }, done);
}
//# sourceMappingURL=index.js.map