"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const AbstractController_1 = require("./AbstractController");
const parameters_1 = require("../parameters");
const constants_1 = require("../constants");
const TransactionDTO_1 = require("../../../../lib/dto/TransactionDTO");
const underscore_1 = require("../../../../lib/common-libs/underscore");
const http2raw = require('../http2raw');
class TransactionBinding extends AbstractController_1.AbstractController {
    async parseTransaction(req) {
        const res = await this.pushEntity(req, http2raw.transaction, (raw) => this.server.writeRawTransaction(raw));
        return {
            version: res.version,
            currency: res.currency,
            issuers: res.issuers,
            inputs: res.inputs,
            outputs: res.outputs,
            unlocks: res.unlocks,
            signatures: res.signatures,
            comment: res.comment,
            locktime: res.locktime,
            hash: res.hash,
            written_block: res.blockNumber,
            raw: res.getRaw()
        };
    }
    async getSources(req) {
        const pubkey = await parameters_1.ParametersService.getPubkeyP(req);
        const sources = await this.server.dal.getAvailableSourcesByPubkey(pubkey);
        return {
            currency: this.conf.currency,
            pubkey,
            sources
        };
    }
    async getByHash(req) {
        const hash = parameters_1.ParametersService.getHash(req);
        const tx = await this.server.dal.getTxByHash(hash);
        if (!tx) {
            throw constants_1.BMAConstants.ERRORS.TX_NOT_FOUND;
        }
        tx.inputs = tx.inputs.map((i) => i.raw || i);
        tx.outputs = tx.outputs.map((o) => o.raw || o);
        return {
            version: tx.version,
            currency: tx.currency,
            locktime: tx.locktime,
            // blockstamp: tx.blockstamp,
            // blockstampTime: tx.blockstampTime,
            issuers: tx.issuers,
            inputs: tx.inputs,
            outputs: tx.outputs,
            unlocks: tx.unlocks,
            signatures: tx.signatures,
            comment: tx.comment,
            hash: tx.hash,
            // time: tx.time,
            // block_number: tx.block_number,
            written_block: tx.block_number,
            // received: tx.received,
            raw: ""
        };
    }
    async getHistory(req) {
        const pubkey = await parameters_1.ParametersService.getPubkeyP(req);
        return this.getFilteredHistory(pubkey, (results) => results);
    }
    async getHistoryBetweenBlocks(req) {
        const pubkey = await parameters_1.ParametersService.getPubkeyP(req);
        const from = await parameters_1.ParametersService.getFromP(req);
        const to = await parameters_1.ParametersService.getToP(req);
        return this.getFilteredHistory(pubkey, (res) => {
            const histo = res.history;
            histo.sent = underscore_1.Underscore.filter(histo.sent, function (tx) { return tx && tx.block_number >= from && tx.block_number <= to; });
            histo.received = underscore_1.Underscore.filter(histo.received, function (tx) { return tx && tx.block_number >= from && tx.block_number <= to; });
            underscore_1.Underscore.extend(histo, { sending: [], receiving: [] });
            return res;
        });
    }
    async getHistoryBetweenTimes(req) {
        const pubkey = await parameters_1.ParametersService.getPubkeyP(req);
        const from = await parameters_1.ParametersService.getFromP(req);
        const to = await parameters_1.ParametersService.getToP(req);
        return this.getFilteredHistory(pubkey, (res) => {
            const histo = res.history;
            histo.sent = underscore_1.Underscore.filter(histo.sent, function (tx) { return tx && tx.time >= from && tx.time <= to; });
            histo.received = underscore_1.Underscore.filter(histo.received, function (tx) { return tx && tx.time >= from && tx.time <= to; });
            underscore_1.Underscore.extend(histo, { sending: [], receiving: [] });
            return res;
        });
    }
    async getPendingForPubkey(req) {
        const pubkey = await parameters_1.ParametersService.getPubkeyP(req);
        return this.getFilteredHistory(pubkey, function (res) {
            const histo = res.history;
            underscore_1.Underscore.extend(histo, { sent: [], received: [] });
            return res;
        });
    }
    async getPending() {
        const pending = await this.server.dal.getTransactionsPending();
        return {
            currency: this.conf.currency,
            pending: pending.map(t => {
                const tx = TransactionDTO_1.TransactionDTO.fromJSONObject(t);
                return {
                    version: tx.version,
                    issuers: tx.issuers,
                    inputs: tx.inputs,
                    unlocks: tx.unlocks,
                    outputs: tx.outputs,
                    comment: tx.comment,
                    locktime: tx.locktime,
                    blockstamp: tx.blockstamp,
                    blockstampTime: tx.blockstampTime,
                    signatures: tx.signatures,
                    hash: tx.hash
                };
            })
        };
    }
    async getFilteredHistory(pubkey, filter) {
        let history = await this.server.dal.getTransactionsHistory(pubkey);
        let result = {
            "currency": this.conf.currency,
            "pubkey": pubkey,
            "history": {
                sending: history.sending.map(dbtx2HttpTxOfHistory),
                received: history.received.map(dbtx2HttpTxOfHistory),
                receiving: history.receiving.map(dbtx2HttpTxOfHistory),
                sent: history.sent.map(dbtx2HttpTxOfHistory),
                pending: history.pending.map(dbtx2HttpTxOfHistory)
            }
        };
        return filter(result);
    }
}
exports.TransactionBinding = TransactionBinding;
function dbtx2HttpTxOfHistory(tx) {
    return {
        version: tx.version,
        locktime: tx.locktime,
        blockstamp: tx.blockstamp,
        blockstampTime: tx.blockstampTime,
        issuers: tx.issuers,
        inputs: tx.inputs,
        outputs: tx.outputs,
        unlocks: tx.unlocks,
        signatures: tx.signatures,
        comment: tx.comment,
        hash: tx.hash,
        time: tx.time,
        block_number: tx.block_number,
        received: tx.received
    };
}
//# sourceMappingURL=transactions.js.map