"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const AbstractController_1 = require("./AbstractController");
const constants_1 = require("../constants");
const IdentityDAL_1 = require("../../../../lib/dal/sqliteDAL/IdentityDAL");
const IdentityDTO_1 = require("../../../../lib/dto/IdentityDTO");
const underscore_1 = require("../../../../lib/common-libs/underscore");
const http2raw = require('../http2raw');
const constants = require('../../../../lib/constants');
const ParametersService = require('../parameters').ParametersService;
class WOTBinding extends AbstractController_1.AbstractController {
    async lookup(req) {
        // Get the search parameter from HTTP query
        const search = await ParametersService.getSearchP(req);
        // Make the research
        const identities = await this.IdentityService.searchIdentities(search);
        // Entitify each result
        identities.forEach((idty, index) => identities[index] = IdentityDAL_1.DBIdentity.copyFromExisting(idty));
        // Prepare some data to avoid displaying expired certifications
        for (const idty of identities) {
            const certs = await this.server.dal.certsToTarget(idty.pubkey, idty.getTargetHash());
            const validCerts = [];
            for (const cert of certs) {
                const member = await this.server.dal.getWrittenIdtyByPubkeyForUidAndIsMemberAndWasMember(cert.from);
                if (member) {
                    cert.uids = [member.uid];
                    cert.isMember = member.member;
                    cert.wasMember = member.wasMember;
                }
                else {
                    const potentials = await this.IdentityService.getPendingFromPubkey(cert.from);
                    cert.uids = potentials.map(p => p.uid);
                    cert.isMember = false;
                    cert.wasMember = false;
                }
                validCerts.push(cert);
            }
            idty.certs = validCerts;
            const signed = await this.server.dal.certsFrom(idty.pubkey);
            const validSigned = [];
            for (let j = 0; j < signed.length; j++) {
                const cert = underscore_1.Underscore.clone(signed[j]);
                cert.idty = await this.server.dal.getGlobalIdentityByHashForLookup(cert.target);
                if (cert.idty) {
                    validSigned.push(cert);
                }
                else {
                    this.logger.debug('A certification to an unknown identity was found (%s => %s)', cert.from, cert.to);
                }
            }
            idty.signed = validSigned;
        }
        if (identities.length == 0) {
            throw constants_1.BMAConstants.ERRORS.NO_MATCHING_IDENTITY;
        }
        const resultsByPubkey = {};
        identities.forEach((identity) => {
            const copy = IdentityDAL_1.DBIdentity.copyFromExisting(identity);
            const jsoned = copy.json();
            if (!resultsByPubkey[jsoned.pubkey]) {
                // Create the first matching identity with this pubkey in the map
                resultsByPubkey[jsoned.pubkey] = jsoned;
            }
            else {
                // Merge the identity with the existing(s)
                const existing = resultsByPubkey[jsoned.pubkey];
                // We add the UID of the identity to the list of already added UIDs
                existing.uids = existing.uids.concat(jsoned.uids);
                // We do not merge the `signed`: every identity with the same pubkey has the same `signed` because it the *pubkey* which signs, not the identity
            }
        });
        return {
            partial: false,
            results: underscore_1.Underscore.values(resultsByPubkey)
        };
    }
    async members() {
        const identities = await this.server.dal.getMembers();
        const json = {
            results: []
        };
        identities.forEach((identity) => json.results.push({ pubkey: identity.pubkey, uid: identity.uid }));
        return json;
    }
    async certifiersOf(req) {
        const search = await ParametersService.getSearchP(req);
        const idty = (await this.server.dal.getWrittenIdtyByPubkeyOrUIdForHashingAndIsMember(search));
        const certs = await this.server.dal.certsToTarget(idty.pub, IdentityDTO_1.IdentityDTO.getTargetHash(idty));
        const theCerts = [];
        for (const cert of certs) {
            const certifier = await this.server.dal.getWrittenIdtyByPubkeyForUidAndMemberAndCreatedOn(cert.from);
            if (certifier) {
                let certBlock = await this.server.dal.getBlockWeHaveItForSure(cert.block_number);
                theCerts.push({
                    pubkey: cert.from,
                    uid: certifier.uid,
                    isMember: certifier.member,
                    wasMember: true,
                    cert_time: {
                        block: certBlock.number,
                        medianTime: certBlock.medianTime
                    },
                    sigDate: certifier.created_on,
                    written: (cert.written_block !== null && cert.written_hash) ? {
                        number: cert.written_block,
                        hash: cert.written_hash
                    } : null,
                    signature: cert.sig
                });
            }
        }
        return {
            pubkey: idty.pub,
            uid: idty.uid,
            sigDate: idty.created_on,
            isMember: idty.member,
            certifications: theCerts
        };
    }
    async requirements(req) {
        const search = await ParametersService.getSearchP(req);
        const identities = await this.IdentityService.searchIdentities(search);
        const all = await this.BlockchainService.requirementsOfIdentities(identities);
        if (!all || !all.length) {
            throw constants_1.BMAConstants.ERRORS.NO_IDTY_MATCHING_PUB_OR_UID;
        }
        return {
            identities: all
        };
    }
    async requirementsOfPending(req) {
        const minsig = ParametersService.getMinSig(req);
        let identities = (await this.server.dal.idtyDAL.query('SELECT i.*, count(c.sig) as nbSig ' +
            'FROM idty i, cert c ' +
            'WHERE c.target = i.hash group by i.hash having nbSig >= ?', [minsig])).map(i => ({
            hash: i.hash || "",
            member: i.member || false,
            wasMember: i.wasMember || false,
            pubkey: i.pubkey,
            uid: i.uid || "",
            buid: i.buid || "",
            sig: i.sig || "",
            revocation_sig: i.revocation_sig,
            revoked: i.revoked,
            revoked_on: i.revoked_on ? 1 : 0
        }));
        const members = await this.server.dal.findReceiversAbove(minsig);
        identities = identities.concat(members);
        const all = await this.BlockchainService.requirementsOfIdentities(identities, false);
        if (!all || !all.length) {
            throw constants_1.BMAConstants.ERRORS.NO_IDTY_MATCHING_PUB_OR_UID;
        }
        return {
            identities: all
        };
    }
    async certifiedBy(req) {
        const search = await ParametersService.getSearchP(req);
        const idty = (await this.server.dal.getWrittenIdtyByPubkeyOrUIdForHashingAndIsMember(search));
        const certs = await this.server.dal.certsFrom(idty.pub);
        const theCerts = [];
        for (const cert of certs) {
            const certified = await this.server.dal.getWrittenIdtyByPubkeyForUidAndMemberAndCreatedOn(cert.to);
            if (certified) {
                let certBlock = await this.server.dal.getBlockWeHaveItForSure(cert.block_number);
                theCerts.push({
                    pubkey: cert.to,
                    uid: certified.uid,
                    isMember: certified.member,
                    wasMember: true,
                    cert_time: {
                        block: certBlock.number,
                        medianTime: certBlock.medianTime
                    },
                    sigDate: certified.created_on,
                    written: (cert.written_block !== null && cert.written_hash) ? {
                        number: cert.written_block,
                        hash: cert.written_hash
                    } : null,
                    signature: cert.sig
                });
            }
        }
        return {
            pubkey: idty.pub,
            uid: idty.uid,
            sigDate: idty.created_on,
            isMember: idty.member,
            certifications: theCerts
        };
    }
    async identityOf(req) {
        let search = await ParametersService.getSearchP(req);
        const idty = await this.server.dal.getWrittenIdtyByPubkeyOrUIdForHashingAndIsMember(search);
        if (!idty) {
            throw constants.ERRORS.NO_MEMBER_MATCHING_PUB_OR_UID;
        }
        if (!idty.member) {
            throw 'Not a member';
        }
        return {
            pubkey: idty.pub,
            uid: idty.uid,
            sigDate: idty.created_on
        };
    }
    async add(req) {
        const res = await this.pushEntity(req, http2raw.identity, (raw) => this.server.writeRawIdentity(raw));
        return {
            pubkey: res.pubkey,
            uids: [],
            signed: []
        };
    }
    async certify(req) {
        const res = await this.pushEntity(req, http2raw.certification, (raw) => this.server.writeRawCertification(raw));
        const target = {
            issuer: res.idty_issuer,
            uid: res.idty_uid,
            timestamp: res.idty_buid,
            sig: res.idty_sig
        };
        return {
            issuer: res.issuer,
            timestamp: res.buid,
            sig: res.sig,
            target
        };
    }
    async revoke(req) {
        const res = await this.pushEntity(req, http2raw.revocation, (raw) => this.server.writeRawRevocation(raw));
        return {
            result: true
        };
    }
    async pendingMemberships() {
        const memberships = await this.server.dal.findNewcomers();
        const json = {
            memberships: memberships.map((ms) => {
                return {
                    pubkey: ms.issuer,
                    uid: ms.userid,
                    version: 10,
                    currency: this.server.conf.currency,
                    membership: ms.membership,
                    blockNumber: ms.blockNumber,
                    blockHash: ms.blockHash,
                    written: (!ms.written_number && ms.written_number !== 0) ? null : ms.written_number
                };
            })
        };
        json.memberships = underscore_1.Underscore.sortBy(json.memberships, 'blockNumber');
        json.memberships.reverse();
        return json;
    }
}
exports.WOTBinding = WOTBinding;
//# sourceMappingURL=wot.js.map