"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const contacter_1 = require("./lib/contacter");
const crawler_1 = require("./lib/crawler");
const sync_1 = require("./lib/sync");
const req2fwd_1 = require("./lib/req2fwd");
const index_1 = require("../../lib/common-libs/index");
const PeerDTO_1 = require("../../lib/dto/PeerDTO");
const buid_1 = require("../../lib/common-libs/buid");
const BlockDTO_1 = require("../../lib/dto/BlockDTO");
const directory_1 = require("../../lib/system/directory");
const fileDAL_1 = require("../../lib/dal/fileDAL");
const RemoteSynchronizer_1 = require("./lib/sync/RemoteSynchronizer");
const LocalPathSynchronizer_1 = require("./lib/sync/LocalPathSynchronizer");
const constants_1 = require("../../lib/common-libs/constants");
const errors_1 = require("../../lib/common-libs/errors");
const logger_1 = require("../../lib/logger");
const constants_2 = require("./lib/constants");
const exit_codes_1 = require("../../lib/common-libs/exit-codes");
const connect_1 = require("./lib/connect");
const BMARemoteContacter_1 = require("./lib/sync/BMARemoteContacter");
const sandbox_1 = require("./lib/sandbox");
const HOST_PATTERN = /^[^:/]+(:[0-9]{1,5})?(\/.*)?$/;
const FILE_PATTERN = /^(\/.+)$/;
exports.CrawlerDependency = {
    duniter: {
        service: {
            process: (server, conf, logger) => new crawler_1.Crawler(server, conf, logger)
        },
        methods: {
            contacter: (host, port, opts) => new contacter_1.Contacter(host, port, opts),
            pullBlocks: async (server, pubkey = "") => {
                const crawler = new crawler_1.Crawler(server, server.conf, server.logger);
                return crawler.pullBlocks(server, pubkey);
            },
            pullSandbox: async (server) => {
                const crawler = new crawler_1.Crawler(server, server.conf, server.logger);
                return crawler.sandboxPull(server);
            },
            synchronize: (server, onHost, onPort, upTo, chunkLength, allowLocalSync = false) => {
                const strategy = new RemoteSynchronizer_1.RemoteSynchronizer(onHost, onPort, server, chunkLength, undefined, undefined, allowLocalSync);
                const remote = new sync_1.Synchroniser(server, strategy);
                const syncPromise = remote.sync(upTo, chunkLength);
                return {
                    flow: remote,
                    syncPromise
                };
            },
            /**
             * Used by duniter-ui
             * @param {Server} server
             * @param {string} onHost
             * @param {number} onPort
             * @returns {Promise<any>}
             */
            testForSync: (server, onHost, onPort) => {
                return RemoteSynchronizer_1.RemoteSynchronizer.test(onHost, onPort, server.conf.pair);
            }
        },
        cliOptions: [
            { value: '--nointeractive', desc: 'Disable interactive sync UI.' },
            { value: '--nocautious', desc: 'Do not check blocks validity during sync.' },
            { value: '--cautious', desc: 'Check blocks validity during sync (overrides --nocautious option).' },
            { value: '--nopeers', desc: 'Do not retrieve peers during sync.' },
            { value: '--nop2p', desc: 'Disables P2P downloading of blocs during sync.' },
            { value: '--localsync', desc: 'Allow to synchronize on nodes with local network IP address for `sync` command' },
            { value: '--nosources', desc: 'Do not parse sources (UD, TX) during sync (debug purposes).' },
            { value: '--nosbx', desc: 'Do not retrieve sandboxes during sync.' },
            { value: '--onlypeers', desc: 'Will only try to sync peers.' },
            { value: '--slow', desc: 'Download slowly the blokchcain (for low connnections).' },
            { value: '--readfilesystem', desc: 'Also read the filesystem to speed up block downloading.' },
            { value: '--minsig <minsig>', desc: 'Minimum pending signatures count for `crawl-lookup`. Default is 5.' },
        ],
        cli: [{
                name: 'sync [source] [to]',
                desc: 'Synchronize blockchain from a remote Duniter node. [source] is [host][:port]. [to] defaults to remote current block number.',
                preventIfRunning: true,
                onConfiguredExecute: async (server) => {
                    await server.resetData();
                },
                onDatabaseExecute: async (server, conf, program, params) => {
                    const source = params[0];
                    const to = params[1];
                    if (!source || !(source.match(HOST_PATTERN) || source.match(FILE_PATTERN))) {
                        throw 'Source of sync is required. (either a host:port or a file path)';
                    }
                    let cautious;
                    if (program.nocautious) {
                        cautious = false;
                    }
                    if (program.cautious) {
                        cautious = true;
                    }
                    const upTo = parseInt(to);
                    const chunkLength = 0;
                    const interactive = !program.nointeractive;
                    const askedCautious = cautious;
                    const noShufflePeers = program.noshuffle;
                    let otherDAL = undefined;
                    if (program.readfilesystem) {
                        const dbName = program.mdb;
                        const dbHome = program.home;
                        const home = directory_1.Directory.getHome(dbName, dbHome);
                        const params = await directory_1.Directory.getHomeParams(false, home);
                        otherDAL = new fileDAL_1.FileDAL(params, async () => null, async () => null);
                    }
                    let strategy;
                    if (source.match(HOST_PATTERN)) {
                        const sp = source.split(':');
                        const onHost = sp[0];
                        const onPort = parseInt(sp[1] ? sp[1] : '443'); // Defaults to 443
                        strategy = new RemoteSynchronizer_1.RemoteSynchronizer(onHost, onPort, server, constants_1.CommonConstants.SYNC_BLOCKS_CHUNK, noShufflePeers === true, otherDAL, program.localsync !== undefined);
                    }
                    else {
                        strategy = new LocalPathSynchronizer_1.LocalPathSynchronizer(source, server, constants_1.CommonConstants.SYNC_BLOCKS_CHUNK);
                    }
                    if (program.onlypeers === true) {
                        return strategy.syncPeers(true);
                    }
                    else {
                        const remote = new sync_1.Synchroniser(server, strategy, interactive === true);
                        // If the sync fail, stop the program
                        process.on('unhandledRejection', (reason) => {
                            if (reason.message === errors_1.DataErrors[errors_1.DataErrors.NO_NODE_FOUND_TO_DOWNLOAD_CHUNK]) {
                                logger_1.NewLogger().error('Synchronization interrupted: no node was found to continue downloading after %s tries.', constants_2.CrawlerConstants.SYNC_MAX_FAIL_NO_NODE_FOUND);
                                process.exit(exit_codes_1.ExitCodes.SYNC_FAIL);
                            }
                        });
                        return remote.sync(upTo, chunkLength, askedCautious);
                    }
                }
            }, {
                name: 'peer [host] [port]',
                desc: 'Exchange peerings with another node',
                preventIfRunning: true,
                onDatabaseExecute: async (server, conf, program, params) => {
                    const host = params[0];
                    const port = params[1];
                    const logger = server.logger;
                    try {
                        const ERASE_IF_ALREADY_RECORDED = true;
                        logger.info('Fetching peering record at %s:%s...', host, port);
                        let peering = await contacter_1.Contacter.fetchPeer(host, port);
                        logger.info('Apply peering ...');
                        await server.PeeringService.submitP(peering, ERASE_IF_ALREADY_RECORDED, !program.nocautious, true);
                        logger.info('Applied');
                        let selfPeer = await server.dal.getPeer(server.PeeringService.pubkey);
                        if (!selfPeer) {
                            await server.PeeringService.generateSelfPeer(server.conf);
                            selfPeer = await server.dal.getPeer(server.PeeringService.pubkey);
                        }
                        logger.info('Send self peering ...');
                        const p = PeerDTO_1.PeerDTO.fromJSONObject(peering);
                        const contact = new contacter_1.Contacter(p.getHostPreferDNS(), p.getPort(), {});
                        await contact.postPeer(PeerDTO_1.PeerDTO.fromJSONObject(selfPeer));
                        logger.info('Sent.');
                        await server.disconnect();
                    }
                    catch (e) {
                        logger.error(e.code || e.message || e);
                        throw Error("Exiting");
                    }
                }
            }, {
                name: 'import <fromHost> <fromPort> <search> <toHost> <toPort>',
                desc: 'Import all pending data from matching <search>',
                onDatabaseExecute: async (server, conf, program, params) => {
                    const fromHost = params[0];
                    const fromPort = params[1];
                    const search = params[2];
                    const toHost = params[3];
                    const toPort = params[4];
                    const logger = server.logger;
                    try {
                        const peers = fromHost && fromPort ? [{ endpoints: [['BASIC_MERKLED_API', fromHost, fromPort].join(' ')] }] : await server.dal.peerDAL.withUPStatus();
                        // Memberships
                        for (const p of peers) {
                            const peer = PeerDTO_1.PeerDTO.fromJSONObject(p);
                            const fromHost = peer.getHostPreferDNS();
                            const fromPort = peer.getPort();
                            logger.info('Looking at %s:%s...', fromHost, fromPort);
                            try {
                                const node = new contacter_1.Contacter(fromHost, fromPort, { timeout: 10000 });
                                const requirements = await node.getRequirements(search);
                                await req2fwd_1.req2fwd(requirements, toHost, toPort, logger);
                            }
                            catch (e) {
                                logger.error(e);
                            }
                        }
                        await server.disconnect();
                    }
                    catch (e) {
                        logger.error(e);
                        throw Error("Exiting");
                    }
                }
            }, {
                name: 'sync-mempool <from>',
                desc: 'Import all pending data from matching <search>',
                onDatabaseExecute: async (server, conf, program, params) => {
                    const source = params[0];
                    if (!source || !(source.match(HOST_PATTERN) || source.match(FILE_PATTERN))) {
                        throw 'Source of sync is required. (host[:port])';
                    }
                    const logger = logger_1.NewLogger();
                    const from = params[0];
                    const { host, port } = extractHostPort(from);
                    try {
                        const peer = PeerDTO_1.PeerDTO.fromJSONObject({ endpoints: [['BASIC_MERKLED_API', host, port].join(' ')] });
                        const fromHost = peer.getHostPreferDNS();
                        const fromPort = peer.getPort();
                        logger.info('Looking at %s:%s...', fromHost, fromPort);
                        try {
                            const fromHost = await connect_1.connect(peer, 60 * 1000);
                            const api = new BMARemoteContacter_1.BMARemoteContacter(fromHost);
                            await sandbox_1.pullSandboxToLocalServer(server.conf.currency, api, server, logger);
                        }
                        catch (e) {
                            logger.error(e);
                        }
                        await server.disconnect();
                    }
                    catch (e) {
                        logger.error(e);
                        throw Error("Exiting");
                    }
                }
            }, {
                name: 'sync-mempool-search <from> <search>',
                desc: 'Import all pending data from matching <search>',
                onDatabaseExecute: async (server, conf, program, params) => {
                    const source = params[0];
                    const search = params[1];
                    if (!source || !(source.match(HOST_PATTERN) || source.match(FILE_PATTERN))) {
                        throw 'Source of sync is required. (host[:port])';
                    }
                    const logger = logger_1.NewLogger();
                    const from = params[0];
                    const { host, port } = extractHostPort(from);
                    try {
                        const peer = PeerDTO_1.PeerDTO.fromJSONObject({ endpoints: [['BASIC_MERKLED_API', host, port].join(' ')] });
                        const fromHost = peer.getHostPreferDNS();
                        const fromPort = peer.getPort();
                        logger.info('Looking at %s:%s...', fromHost, fromPort);
                        try {
                            const fromHost = await connect_1.connect(peer);
                            const res = await fromHost.getRequirements(search);
                            await sandbox_1.applyMempoolRequirements(server.conf.currency, res, server, logger);
                        }
                        catch (e) {
                            logger.error(e);
                        }
                        await server.disconnect();
                    }
                    catch (e) {
                        logger.error(e);
                        throw Error("Exiting");
                    }
                }
            }, {
                name: 'sync-mempool-fwd <from> <to> <search>',
                desc: 'Import all pending data from matching <search>',
                onDatabaseExecute: async (server, conf, program, params) => {
                    const source = params[0];
                    const target = params[1];
                    const search = params[2];
                    if (!source || !(source.match(HOST_PATTERN) || source.match(FILE_PATTERN))) {
                        throw 'Source of sync is required. (host[:port])';
                    }
                    if (!target || !(target.match(HOST_PATTERN) || target.match(FILE_PATTERN))) {
                        throw 'Target of sync is required. (host[:port])';
                    }
                    const logger = logger_1.NewLogger();
                    const { host, port } = extractHostPort(source);
                    const { host: toHost, port: toPort } = extractHostPort(target);
                    try {
                        const peer = PeerDTO_1.PeerDTO.fromJSONObject({ endpoints: [['BASIC_MERKLED_API', host, port].join(' ')] });
                        logger.info('Looking at %s...', source);
                        try {
                            const fromHost = await connect_1.connect(peer);
                            const res = await fromHost.getRequirements(search);
                            await sandbox_1.forwardToServer(server.conf.currency, res, toHost, toPort, logger);
                        }
                        catch (e) {
                            logger.error(e);
                        }
                        await server.disconnect();
                    }
                    catch (e) {
                        logger.error(e);
                        throw Error("Exiting");
                    }
                }
            }, {
                name: 'pull <from> [<number>]',
                desc: 'Pull blocks from <from> source up to block <number>',
                onDatabaseExecute: async (server, conf, program, params) => {
                    const source = params[0];
                    const to = parseInt(params[1]);
                    if (!source || !(source.match(HOST_PATTERN) || source.match(FILE_PATTERN))) {
                        throw 'Source of sync is required. (host[:port])';
                    }
                    const logger = logger_1.NewLogger();
                    const { host, port } = extractHostPort(source);
                    try {
                        const peer = PeerDTO_1.PeerDTO.fromJSONObject({ endpoints: [['BASIC_MERKLED_API', host, port].join(' ')] });
                        logger.info('Looking at %s...', source);
                        try {
                            const fromHost = await connect_1.connect(peer);
                            let current = await server.dal.getCurrentBlockOrNull();
                            // Loop until an error occurs
                            while (current && (isNaN(to) || current.number < to)) {
                                current = await fromHost.getBlock(current.number + 1);
                                await server.writeBlock(current, false);
                            }
                        }
                        catch (e) {
                            logger.error(e);
                        }
                    }
                    catch (e) {
                        logger.error(e);
                        throw Error("Exiting");
                    }
                }
            }, {
                name: 'forward <number> <fromHost> <fromPort> <toHost> <toPort>',
                desc: 'Forward existing block <number> from a host to another',
                onDatabaseExecute: async (server, conf, program, params) => {
                    const number = params[0];
                    const fromHost = params[1];
                    const fromPort = params[2];
                    const toHost = params[3];
                    const toPort = params[4];
                    const logger = server.logger;
                    try {
                        logger.info('Looking at %s:%s...', fromHost, fromPort);
                        try {
                            const source = new contacter_1.Contacter(fromHost, fromPort, { timeout: 10000 });
                            const target = new contacter_1.Contacter(toHost, toPort, { timeout: 10000 });
                            const block = await source.getBlock(number);
                            const raw = BlockDTO_1.BlockDTO.fromJSONObject(block).getRawSigned();
                            await target.postBlock(raw);
                        }
                        catch (e) {
                            logger.error(e);
                        }
                        await server.disconnect();
                    }
                    catch (e) {
                        logger.error(e);
                        throw Error("Exiting");
                    }
                }
            }, {
                name: 'import-lookup [search] [fromhost] [fromport] [tohost] [toport]',
                desc: 'Exchange peerings with another node',
                onDatabaseExecute: async (server, conf, program, params) => {
                    const search = params[0];
                    const fromhost = params[1];
                    const fromport = params[2];
                    const tohost = params[3];
                    const toport = params[4];
                    const logger = server.logger;
                    try {
                        logger.info('Looking for "%s" at %s:%s...', search, fromhost, fromport);
                        const sourcePeer = new contacter_1.Contacter(fromhost, fromport);
                        const targetPeer = new contacter_1.Contacter(tohost, toport);
                        const lookup = await sourcePeer.getLookup(search);
                        for (const res of lookup.results) {
                            for (const uid of res.uids) {
                                const rawIdty = index_1.rawer.getOfficialIdentity({
                                    currency: 'g1',
                                    issuer: res.pubkey,
                                    uid: uid.uid,
                                    buid: uid.meta.timestamp,
                                    sig: uid.self
                                });
                                logger.info('Success idty %s', uid.uid);
                                try {
                                    await targetPeer.postIdentity(rawIdty);
                                }
                                catch (e) {
                                    logger.error(e);
                                }
                                for (const received of uid.others) {
                                    const rawCert = index_1.rawer.getOfficialCertification({
                                        currency: 'g1',
                                        issuer: received.pubkey,
                                        idty_issuer: res.pubkey,
                                        idty_uid: uid.uid,
                                        idty_buid: uid.meta.timestamp,
                                        idty_sig: uid.self,
                                        buid: buid_1.Buid.format.buid(received.meta.block_number, received.meta.block_hash),
                                        sig: received.signature
                                    });
                                    try {
                                        logger.info('Success cert %s -> %s', received.pubkey.slice(0, 8), uid.uid);
                                        await targetPeer.postCert(rawCert);
                                    }
                                    catch (e) {
                                        logger.error(e);
                                    }
                                }
                            }
                        }
                        const certBy = await sourcePeer.getCertifiedBy(search);
                        const mapBlocks = {};
                        for (const signed of certBy.certifications) {
                            if (signed.written) {
                                logger.info('Already written cert %s -> %s', certBy.pubkey.slice(0, 8), signed.uid);
                            }
                            else {
                                const lookupIdty = await sourcePeer.getLookup(signed.pubkey);
                                let idty = null;
                                for (const result of lookupIdty.results) {
                                    for (const uid of result.uids) {
                                        if (uid.uid === signed.uid && result.pubkey === signed.pubkey && uid.meta.timestamp === signed.sigDate) {
                                            idty = uid;
                                        }
                                    }
                                }
                                let block = mapBlocks[signed.cert_time.block];
                                if (!block) {
                                    block = await sourcePeer.getBlock(signed.cert_time.block);
                                    mapBlocks[block.number] = block;
                                }
                                const rawCert = index_1.rawer.getOfficialCertification({
                                    currency: 'g1',
                                    issuer: certBy.pubkey,
                                    idty_issuer: signed.pubkey,
                                    idty_uid: signed.uid,
                                    idty_buid: idty.meta.timestamp,
                                    idty_sig: idty.self,
                                    buid: buid_1.Buid.format.buid(block.number, block.hash),
                                    sig: signed.signature
                                });
                                try {
                                    logger.info('Success cert %s -> %s', certBy.pubkey.slice(0, 8), signed.uid);
                                    await targetPeer.postCert(rawCert);
                                }
                                catch (e) {
                                    logger.error(e);
                                }
                            }
                        }
                        logger.info('Sent.');
                        await server.disconnect();
                    }
                    catch (e) {
                        logger.error(e);
                        throw Error("Exiting");
                    }
                }
            }, {
                name: 'crawl-lookup <toHost> <toPort> [<fromHost> [<fromPort>]]',
                desc: 'Make a full network scan and rebroadcast every WoT pending document (identity, certification, membership)',
                onDatabaseExecute: async (server, conf, program, params) => {
                    const toHost = params[0];
                    const toPort = params[1];
                    const fromHost = params[2];
                    const fromPort = params[3];
                    const logger = server.logger;
                    try {
                        const peers = fromHost && fromPort ? [{ endpoints: [['BASIC_MERKLED_API', fromHost, fromPort].join(' ')] }] : await server.dal.peerDAL.withUPStatus();
                        // Memberships
                        for (const p of peers) {
                            const peer = PeerDTO_1.PeerDTO.fromJSONObject(p);
                            const fromHost = peer.getHostPreferDNS();
                            const fromPort = peer.getPort();
                            logger.info('Looking at %s:%s...', fromHost, fromPort);
                            try {
                                const node = new contacter_1.Contacter(fromHost, fromPort, { timeout: 10000 });
                                const requirements = await node.getRequirementsPending(program.minsig || 5);
                                await req2fwd_1.req2fwd(requirements, toHost, toPort, logger);
                            }
                            catch (e) {
                                logger.error(e);
                            }
                        }
                        await server.disconnect();
                    }
                    catch (e) {
                        logger.error(e);
                        throw Error("Exiting");
                    }
                }
            }, {
                name: 'fwd-pending-ms',
                desc: 'Forwards all the local pending memberships to target node',
                onDatabaseExecute: async (server, conf, program, params) => {
                    const logger = server.logger;
                    try {
                        const pendingMSS = await server.dal.msDAL.getPendingIN();
                        const targetPeer = new contacter_1.Contacter('g1.cgeek.fr', 80, { timeout: 5000 });
                        // Membership
                        let rawMS;
                        for (const theMS of pendingMSS) {
                            console.log('New membership pending for %s', theMS.userid);
                            try {
                                rawMS = index_1.rawer.getMembership({
                                    currency: 'g1',
                                    issuer: theMS.issuer,
                                    block: theMS.block,
                                    membership: theMS.membership,
                                    userid: theMS.userid,
                                    certts: theMS.certts,
                                    signature: theMS.signature
                                });
                                await targetPeer.postRenew(rawMS);
                                logger.info('Success ms idty %s', theMS.userid);
                            }
                            catch (e) {
                                logger.warn(e);
                            }
                        }
                        await server.disconnect();
                    }
                    catch (e) {
                        logger.error(e);
                        throw Error("Exiting");
                    }
                }
            }]
    }
};
function extractHostPort(source) {
    const sp = source.split(':');
    const onHost = sp[0];
    const onPort = sp[1] ? sp[1] : '443'; // Defaults to 443
    return {
        host: onHost,
        port: onPort,
    };
}
//# sourceMappingURL=index.js.map