"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const constants_1 = require("./constants");
const rp = require('request-promise');
const sanitize = require('../../../modules/bma/lib/sanitize');
const dtos = require('../../../modules/bma').BmaDependency.duniter.methods.dtos;
class Contacter {
    constructor(host, port, opts = {}) {
        this.host = host;
        this.port = port;
        this.path = '';
        this.options = {
            timeout: (opts && opts.timeout) || constants_1.CrawlerConstants.DEFAULT_TIMEOUT
        };
        // We suppose that IPv6 is already wrapped by [], for example 'http://[::1]:80/index.html'
        this.fullyQualifiedHost = [host, port].join(':');
    }
    static fromHostPortPath(host, port, path, opts) {
        const contacter = new Contacter(host, port, opts);
        contacter.path = path;
        return contacter;
    }
    getSummary() {
        return this.get('/node/summary/', dtos.Summary);
    }
    getCertifiedBy(search) {
        return this.get('/wot/certified-by/' + search, dtos.Certifications);
    }
    getRequirements(search) {
        return this.get('/wot/requirements/' + search, dtos.Requirements);
    }
    getRequirementsPending(minsig) {
        return this.get('/wot/requirements-of-pending/' + minsig, dtos.Requirements);
    }
    getLookup(search) {
        return this.get('/wot/lookup/', dtos.Lookup, search);
    }
    getBlock(number) {
        return this.get('/blockchain/block/', dtos.Block, number);
    }
    getCurrent() {
        return this.get('/blockchain/current', dtos.Block);
    }
    getMilestonesPage() {
        return this.get('/blockchain/milestones', dtos.MilestonesPage);
    }
    getMilestones(page) {
        return this.get('/blockchain/milestones/' + page, dtos.Milestones);
    }
    getPeer() {
        return this.get('/network/peering', dtos.Peer);
    }
    getPeers(obj) {
        return this.get('/network/peering/peers', dtos.MerkleOfPeers, obj);
    }
    getPeersArray() {
        return this.get('/network/peers', dtos.Peers);
    }
    getSources(pubkey) {
        return this.get('/tx/sources/', dtos.Sources, pubkey);
    }
    getBlocks(count, fromNumber) {
        return this.get('/blockchain/blocks/', dtos.Blocks, [count, fromNumber].join('/'));
    }
    postPeer(peer) {
        return this.post('/network/peering/peers', dtos.Peer, { peer: peer });
    }
    postIdentity(raw) {
        return this.post('/wot/add', dtos.Identity, { identity: raw });
    }
    postCert(cert) {
        return this.post('/wot/certify', dtos.Cert, { cert: cert });
    }
    postRenew(ms) {
        return this.post('/blockchain/membership', dtos.Membership, { membership: ms });
    }
    postRevocation(rev) {
        return this.post('/wot/revoke', dtos.Identity, { revocation: rev });
    }
    wotPending() {
        return this.get('/wot/pending', dtos.MembershipList);
    }
    wotMembers() {
        return this.get('/wot/members', dtos.Members);
    }
    postBlock(rawBlock) {
        return this.post('/blockchain/block', dtos.Block, { block: rawBlock });
    }
    processTransaction(rawTX) {
        return this.post('/tx/process', dtos.Transaction, { transaction: rawTX });
    }
    async get(url, dtoContract, param) {
        if (typeof param === 'object') {
            // Classical URL params (a=1&b=2&...)
            param = '?' + Object.keys(param).map((k) => [k, param[k]].join('=')).join('&');
        }
        try {
            const path = this.path || '';
            const json = await rp.get({
                url: Contacter.protocol(this.port) + this.fullyQualifiedHost + path + url + (param !== undefined ? param : ''),
                json: true,
                timeout: this.options.timeout
            });
            // Prevent JSON injection
            return sanitize(json, dtoContract);
        }
        catch (e) {
            throw e.error;
        }
    }
    async post(url, dtoContract, data) {
        try {
            const path = this.path || '';
            const json = await rp.post({
                url: Contacter.protocol(this.port) + this.fullyQualifiedHost + path + url,
                body: data,
                json: true,
                timeout: this.options.timeout
            });
            // Prevent JSON injection
            return sanitize(json, dtoContract);
        }
        catch (e) {
            throw e.error;
        }
    }
    static protocol(port) {
        return port == 443 ? 'https://' : 'http://';
    }
    static async quickly(host, port, opts, callbackPromise) {
        const node = new Contacter(host, port, opts);
        return callbackPromise(node);
    }
    static async quickly2(peer, opts, callbackPromise) {
        const Peer = require('./entity/peer');
        const p = Peer.fromJSON(peer);
        const node = new Contacter(p.getHostPreferDNS(), p.getPort(), opts);
        return callbackPromise(node);
    }
    static fetchPeer(host, port, opts = {}) {
        return Contacter.quickly(host, port, opts, (node) => node.getPeer());
    }
    static fetchBlock(number, peer, opts = {}) {
        return Contacter.quickly2(peer, opts, (node) => node.getBlock(number));
    }
    static async isReachableFromTheInternet(peer, opts) {
        const Peer = require('./entity/peer');
        const p = Peer.fromJSON(peer);
        const node = new Contacter(p.getHostPreferDNS(), p.getPort(), opts);
        try {
            await node.getPeer();
            return true;
        }
        catch (e) {
            return false;
        }
    }
}
exports.Contacter = Contacter;
//# sourceMappingURL=contacter.js.map