// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const DBBlock_1 = require("../../../lib/db/DBBlock");
const underscore_1 = require("../../../lib/common-libs/underscore");
class PullingDao {
}
exports.PullingDao = PullingDao;
class AbstractDAO extends PullingDao {
    /**
     * Sugar function. Apply a bunch of blocks instead of one.
     * @param blocks
     */
    async applyBranch(blocks) {
        for (const block of blocks) {
            await this.applyMainBranch(block);
        }
        return true;
    }
    /**
     * Binary search algorithm to find the common root block between a local and a remote blockchain.
     * @param fork An object containing a peer, its current block and top fork block
     * @param forksize The maximum length we can look at to find common root block.
     * @returns {*|Promise}
     */
    async findCommonRoot(fork, forksize) {
        let commonRoot = null;
        let localCurrent = await this.localCurrent();
        if (!localCurrent) {
            throw Error('Local blockchain is empty, cannot find a common root');
        }
        // We look between the top block that is known as fork ...
        let topBlock = fork.block;
        // ... and the bottom which is bounded by `forksize`
        let bottomBlock = await this.getRemoteBlock(fork.peer, Math.max(0, localCurrent.number - forksize));
        let lookBlock = bottomBlock;
        let localEquivalent = await this.getLocalBlock(bottomBlock.number);
        let isCommonBlock = lookBlock.hash == localEquivalent.hash;
        if (isCommonBlock) {
            // Then common root can be found between top and bottom. We process.
            let position = -1, wrongRemotechain = false;
            do {
                isCommonBlock = lookBlock.hash == localEquivalent.hash;
                if (!isCommonBlock) {
                    // Too high, look downward
                    topBlock = lookBlock;
                    position = middle(topBlock.number, bottomBlock.number);
                }
                else {
                    let upperBlock = await this.getRemoteBlock(fork.peer, lookBlock.number + 1);
                    let localUpper = await this.getLocalBlock(upperBlock.number);
                    let isCommonUpper = upperBlock.hash == localUpper.hash;
                    if (isCommonUpper) {
                        // Too low, look upward
                        bottomBlock = lookBlock;
                        position = middle(topBlock.number, bottomBlock.number);
                    }
                    else {
                        // Spotted!
                        commonRoot = lookBlock;
                    }
                }
                let noSpace = topBlock.number == bottomBlock.number + 1;
                if (!commonRoot && noSpace) {
                    // Remote node have inconsistency blockchain, stop search
                    wrongRemotechain = true;
                }
                if (!wrongRemotechain) {
                    lookBlock = await this.getRemoteBlock(fork.peer, position);
                    localEquivalent = await this.getLocalBlock(position);
                }
            } while (!commonRoot && !wrongRemotechain);
        }
        // Otherwise common root is unreachable
        return commonRoot;
    }
    static defaultLocalBlock() {
        const localCurrent = new DBBlock_1.DBBlock();
        localCurrent.number = -1;
        return localCurrent;
    }
    /**
     * Pull algorithm. Look at given peers' blockchain and try to pull blocks from it.
     * May lead local blockchain to fork.
     * @param conf The local node configuration
     * @param dao An abstract layer to retrieve peers data (blocks).
     * @param logger Logger of the main application.
     */
    async pull(conf, logger) {
        let localCurrent = await this.localCurrent() || AbstractDAO.defaultLocalBlock();
        const forks = [];
        if (!localCurrent) {
            localCurrent = new DBBlock_1.DBBlock();
            localCurrent.number = -1;
        }
        const applyCoroutine = async (peer, blocks) => {
            if (blocks.length > 0) {
                let isFork = localCurrent
                    && localCurrent.number !== -1
                    && !(blocks[0].previousHash == localCurrent.hash
                        && blocks[0].number == localCurrent.number + 1);
                if (!isFork) {
                    await this.applyBranch(blocks);
                    const newLocalCurrent = await this.localCurrent();
                    localCurrent = newLocalCurrent || AbstractDAO.defaultLocalBlock();
                    const appliedSuccessfully = localCurrent.number == blocks[blocks.length - 1].number
                        && localCurrent.hash == blocks[blocks.length - 1].hash;
                    return appliedSuccessfully;
                }
                else {
                    let remoteCurrent = await this.remoteCurrent(peer);
                    forks.push({
                        peer: peer,
                        block: blocks[0],
                        current: remoteCurrent
                    });
                    return false;
                }
            }
            return true;
        };
        const downloadCoroutine = async (peer, number) => {
            return await this.downloadBlocks(peer, number);
        };
        const downloadChuncks = async (peer) => {
            let blocksToApply = [];
            const currentBlock = await this.localCurrent();
            let currentChunckStart;
            if (currentBlock) {
                currentChunckStart = currentBlock.number + 1;
            }
            else {
                currentChunckStart = 0;
            }
            let res = { applied: {}, downloaded: [] };
            do {
                let [applied, downloaded] = await Promise.all([
                    applyCoroutine(peer, blocksToApply),
                    downloadCoroutine(peer, currentChunckStart)
                ]);
                res.applied = applied;
                res.downloaded = downloaded;
                blocksToApply = downloaded;
                currentChunckStart += downloaded.length;
                if (!applied) {
                    logger && logger.info("Blocks were not applied.");
                }
            } while (res.downloaded.length > 0 && res.applied);
        };
        let peers = await this.remotePeers();
        // Try to get new legit blocks for local blockchain
        const downloadChuncksTasks = [];
        for (const peer of peers) {
            downloadChuncksTasks.push(downloadChuncks(peer));
        }
        await Promise.all(downloadChuncksTasks);
        // Filter forks: do not include mirror peers (non-member peers)
        let memberForks = [];
        for (const fork of forks) {
            let isMember = await this.isMemberPeer(fork.peer);
            if (isMember) {
                memberForks.push(fork);
            }
        }
        memberForks = memberForks.sort((f1, f2) => {
            let result = compare(f1, f2, "number");
            if (result == 0) {
                result = compare(f1, f2, "medianTime");
            }
            return result;
        });
        memberForks = underscore_1.Underscore.filter(memberForks, (fork) => {
            let blockDistanceInBlocks = (fork.current.number - localCurrent.number);
            let timeDistanceInBlocks = (fork.current.medianTime - localCurrent.medianTime) / conf.avgGenTime;
            const requiredTimeAdvance = conf.switchOnHeadAdvance;
            logger && logger.debug('Fork of %s has blockDistance %s ; timeDistance %s ; required is >= %s for both values to try to follow the fork', fork.peer.pubkey.substr(0, 6), blockDistanceInBlocks.toFixed(2), timeDistanceInBlocks.toFixed(2), requiredTimeAdvance);
            return blockDistanceInBlocks >= requiredTimeAdvance
                && timeDistanceInBlocks >= requiredTimeAdvance;
        });
        // Remove any previous fork block
        await this.removeForks();
        // Find the common root block
        let j = 0, successFork = false;
        while (!successFork && j < memberForks.length) {
            let fork = memberForks[j];
            let commonRootBlock = await this.findCommonRoot(fork, conf.forksize);
            if (commonRootBlock) {
                let blocksToApply = await this.downloadBlocks(fork.peer, commonRootBlock.number + 1, conf.forksize);
                successFork = await this.applyBranch(blocksToApply);
            }
            else {
                logger && logger.debug('No common root block with peer %s', fork.peer.pubkey.substr(0, 6));
            }
            j++;
        }
        return this.localCurrent();
    }
}
exports.AbstractDAO = AbstractDAO;
function compare(f1, f2, field) {
    if (f1[field] > f2[field]) {
        return 1;
    }
    if (f1[field] < f2[field]) {
        return -1;
    }
    return 0;
}
function middle(top, bottom) {
    let difference = top - bottom;
    if (difference % 2 == 1) {
        // We look one step below to not forget any block
        difference++;
    }
    return bottom + (difference / 2);
}
//# sourceMappingURL=pulling.js.map