// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const index_1 = require("../../../lib/common-libs/index");
const index_2 = require("../../../lib/common-libs/parsers/index");
const PeerDTO_1 = require("../../../lib/dto/PeerDTO");
const connect_1 = require("./connect");
exports.pullSandboxToLocalServer = async (currency, fromHost, toServer, logger, watcher = null, nbCertsMin = 1, notify = true) => {
    let res;
    try {
        res = await fromHost.getRequirementsPending(nbCertsMin || 1);
    }
    catch (e) {
        watcher && watcher.writeStatus('Sandbox pulling: could not fetch requirements on %s', fromHost.getName());
    }
    if (res) {
        await applyMempoolRequirements(currency, res, toServer);
    }
};
async function applyMempoolRequirements(currency, res, toServer, notify = true, logger, watcher) {
    const docs = getDocumentsTree(currency, res);
    let t = 0;
    let T = docs.identities.length + docs.certifications.length + docs.revocations.length + docs.memberships.length;
    for (let i = 0; i < docs.identities.length; i++) {
        const idty = docs.identities[i];
        watcher && watcher.writeStatus('Identity ' + (i + 1) + '/' + docs.identities.length);
        watcher && watcher.sbxPercent((t++) / T * 100);
        await submitIdentityToServer(idty, toServer, notify, logger);
    }
    for (let i = 0; i < docs.revocations.length; i++) {
        const idty = docs.revocations[i];
        watcher && watcher.writeStatus('Revocation ' + (i + 1) + '/' + docs.revocations.length);
        watcher && watcher.sbxPercent((t++) / T * 100);
        await submitRevocationToServer(idty, toServer, notify, logger);
    }
    for (let i = 0; i < docs.certifications.length; i++) {
        const cert = docs.certifications[i];
        watcher && watcher.writeStatus('Certification ' + (i + 1) + '/' + docs.certifications.length);
        watcher && watcher.sbxPercent((t++) / T * 100);
        await submitCertificationToServer(cert, toServer, notify, logger);
    }
    for (let i = 0; i < docs.memberships.length; i++) {
        const ms = docs.memberships[i];
        watcher && watcher.writeStatus('Membership ' + (i + 1) + '/' + docs.memberships.length);
        watcher && watcher.sbxPercent((t++) / T * 100);
        await submitMembershipToServer(ms, toServer, notify, logger);
    }
    watcher && watcher.sbxPercent(100);
}
exports.applyMempoolRequirements = applyMempoolRequirements;
async function forwardToServer(currency, res, toHost, toPort, logger, watcher) {
    const docs = getDocumentsTree(currency, res);
    const [port, path] = toPort.split('/');
    let ep = [port == '443' ? 'BMAS' : 'BASIC_MERKLED_API', toHost, port].join(' ');
    if (path) {
        ep += ' /' + path;
    }
    toPort = ':' + toPort;
    const peer = PeerDTO_1.PeerDTO.fromJSONObject({ endpoints: [ep] });
    logger.info('Forwarded to %s...', toHost + toPort);
    const target = await connect_1.connect(peer);
    let t = 0;
    let T = docs.identities.length + docs.certifications.length + docs.revocations.length + docs.memberships.length;
    for (let i = 0; i < docs.identities.length; i++) {
        const idty = docs.identities[i];
        watcher && watcher.writeStatus('Identity ' + (i + 1) + '/' + docs.identities.length);
        watcher && watcher.sbxPercent((t++) / T * 100);
        try {
            await target.postIdentity(idty);
            logger.info('Forwarded identity to %s...', toHost + toPort);
        }
        catch (e) {
            logger.warn(e);
        }
    }
    for (let i = 0; i < docs.revocations.length; i++) {
        const revo = docs.revocations[i];
        watcher && watcher.writeStatus('Revocation ' + (i + 1) + '/' + docs.revocations.length);
        watcher && watcher.sbxPercent((t++) / T * 100);
        try {
            await target.postRevocation(revo);
            logger.info('Forwarded revocation to %s...', toHost + toPort);
        }
        catch (e) {
            logger.warn(e);
        }
    }
    for (let i = 0; i < docs.certifications.length; i++) {
        const cert = docs.certifications[i];
        watcher && watcher.writeStatus('Certification ' + (i + 1) + '/' + docs.certifications.length);
        watcher && watcher.sbxPercent((t++) / T * 100);
        try {
            await target.postCert(cert);
            logger.info('Forwarded cert to %s...', toHost + toPort);
        }
        catch (e) {
            logger.warn(e);
        }
    }
    for (let i = 0; i < docs.memberships.length; i++) {
        const ms = docs.memberships[i];
        watcher && watcher.writeStatus('Membership ' + (i + 1) + '/' + docs.memberships.length);
        watcher && watcher.sbxPercent((t++) / T * 100);
        try {
            await target.postRenew(ms);
            logger.info('Forwarded membership to %s...', toHost + toPort);
        }
        catch (e) {
            logger.warn(e);
        }
    }
    watcher && watcher.sbxPercent(100);
}
exports.forwardToServer = forwardToServer;
function getDocumentsTree(currency, res) {
    const documents = {
        identities: [],
        certifications: [],
        memberships: [],
        revocations: []
    };
    for (const idty of res.identities) {
        const identity = index_1.rawer.getOfficialIdentity({
            currency,
            uid: idty.uid,
            pubkey: idty.pubkey,
            buid: idty.meta.timestamp,
            sig: idty.sig
        });
        if (idty.revocation_sig) {
            const revocation = index_1.rawer.getOfficialRevocation({
                currency,
                uid: idty.uid,
                issuer: idty.pubkey,
                buid: idty.meta.timestamp,
                sig: idty.sig,
                revocation: idty.revocation_sig
            });
            documents.revocations.push(revocation);
        }
        documents.identities.push(identity);
        for (const cert of idty.pendingCerts) {
            const certification = index_1.rawer.getOfficialCertification({
                currency,
                idty_issuer: idty.pubkey,
                idty_uid: idty.uid,
                idty_buid: idty.meta.timestamp,
                idty_sig: idty.sig,
                issuer: cert.from,
                buid: cert.blockstamp,
                sig: cert.sig
            });
            documents.certifications.push(certification);
        }
        for (const ms of idty.pendingMemberships) {
            const membership = index_1.rawer.getMembership({
                currency,
                userid: idty.uid,
                issuer: idty.pubkey,
                certts: idty.meta.timestamp,
                membership: ms.type,
                block: ms.blockstamp,
                signature: ms.sig
            });
            documents.memberships.push(membership);
        }
    }
    return documents;
}
async function submitIdentityToServer(idty, toServer, notify, logger) {
    try {
        const obj = index_2.parsers.parseIdentity.syncWrite(idty);
        await toServer.writeIdentity(obj, notify);
        logger && logger.trace('Sandbox pulling: success with identity \'%s\'', obj.uid);
    }
    catch (e) {
        // Silent error
    }
}
async function submitRevocationToServer(revocation, toServer, notify, logger) {
    try {
        const obj = index_2.parsers.parseRevocation.syncWrite(revocation);
        await toServer.writeRevocation(obj, notify);
        logger && logger.trace('Sandbox pulling: success with revocation \'%s\'', obj.uid);
    }
    catch (e) {
        // Silent error
    }
}
async function submitCertificationToServer(cert, toServer, notify, logger) {
    try {
        const obj = index_2.parsers.parseCertification.syncWrite(cert);
        await toServer.writeCertification(obj, notify);
        logger && logger.trace('Sandbox pulling: success with cert key %s => %s', cert.from.substr(0, 6), cert.idty_uid);
    }
    catch (e) {
        // Silent error
    }
}
async function submitMembershipToServer(ms, toServer, notify, logger) {
    try {
        const obj = index_2.parsers.parseMembership.syncWrite(ms);
        await toServer.writeMembership(obj, notify);
        logger && logger.trace('Sandbox pulling: success with membership \'%s\'', ms.uid);
    }
    catch (e) {
        // Silent error
    }
}
//# sourceMappingURL=sandbox.js.map