"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const PeerDTO_1 = require("../../../../lib/dto/PeerDTO");
const connect_1 = require("../connect");
const logger_1 = require("../../../../lib/logger");
const programOptions_1 = require("../../../../lib/common-libs/programOptions");
const DBPeer_1 = require("../../../../lib/db/DBPeer");
const underscore_1 = require("../../../../lib/common-libs/underscore");
const P2PSyncDownloader_1 = require("./P2PSyncDownloader");
const FsSyncDownloader_1 = require("./FsSyncDownloader");
const AbstractSynchronizer_1 = require("./AbstractSynchronizer");
const sandbox_1 = require("../sandbox");
const path = require("path");
const BMARemoteContacter_1 = require("./BMARemoteContacter");
const WS2PConnection_1 = require("../../../ws2p/lib/WS2PConnection");
const WS2PRequester_1 = require("../../../ws2p/lib/WS2PRequester");
const errors_1 = require("../../../../lib/common-libs/errors");
const keyring_1 = require("../../../../lib/common-libs/crypto/keyring");
const WS2PRemoteContacter_1 = require("./WS2PRemoteContacter");
const logger = logger_1.NewLogger();
class RemoteSynchronizer extends AbstractSynchronizer_1.AbstractSynchronizer {
    constructor(host, port, server, chunkSize, noShufflePeers = false, otherDAL, allowLocalSync = false) {
        super(chunkSize);
        this.host = host;
        this.port = port;
        this.server = server;
        this.noShufflePeers = noShufflePeers;
        this.otherDAL = otherDAL;
        this.allowLocalSync = allowLocalSync;
        this.endpoint = "";
        this.milestones = {};
        this.milestonesPerPage = 1;
        this.maxPage = 0;
    }
    get dal() {
        return this.server.dal;
    }
    get readDAL() {
        return this.otherDAL || this.dal;
    }
    get PeeringService() {
        return this.server.PeeringService;
    }
    getCurrency() {
        return this.currency || 'unknown-currency';
    }
    getPeer() {
        return this.node;
    }
    setWatcher(watcher) {
        this.watcher = watcher;
    }
    getChunksPath() {
        return this.getCurrency();
    }
    async init() {
        const syncApi = await RemoteSynchronizer.getSyncAPI([{ host: this.host, port: this.port }], this.server.conf.pair);
        if (!syncApi.api) {
            this.watcher.syncFailCannotConnectToRemote();
            throw Error(errors_1.DataErrors[errors_1.DataErrors.CANNOT_CONNECT_TO_REMOTE_FOR_SYNC]);
        }
        this.currency = syncApi.currency;
        this.endpoint = syncApi.endpoint;
        this.node = syncApi.api;
        this.peer = PeerDTO_1.PeerDTO.fromJSONObject(syncApi.peering);
        logger.info("Try with %s %s", this.peer.getURL(), this.peer.pubkey.substr(0, 6));
        // We save this peer as a trusted peer for future contact
        try {
            await this.server.PeeringService.submitP(DBPeer_1.DBPeer.fromPeerDTO(this.peer), false, false, true);
        }
        catch (e) {
            logger.debug(e);
        }
        ;
        this.node.pubkey = this.peer.pubkey;
    }
    static async getSyncAPI(hosts, keypair) {
        let api;
        let peering;
        let endpoint = "";
        for (const access of hosts) {
            const host = access.host;
            const port = access.port;
            const path = access.path;
            logger.info(`Connecting to address ${host} :${port}...`);
            // If we know this is a WS2P connection, don't try BMA
            if (access.isWS2P !== true) {
                try {
                    const contacter = await connect_1.connect(PeerDTO_1.PeerDTO.fromJSONObject({ endpoints: [`BASIC_MERKLED_API ${host} ${port}${path && ' ' + path || ''}`] }), 3000);
                    peering = await contacter.getPeer();
                    api = new BMARemoteContacter_1.BMARemoteContacter(contacter);
                    endpoint = 'BASIC_MERKLED_API ' + host + ' ' + port + ((path && ' ' + path) || '');
                }
                catch (e) {
                }
            }
            // If BMA is unreachable and the connection is not marked as strict BMA, let's try WS2P
            if (!api && access.isBMA !== true) {
                const pair = keyring_1.KeyGen(keypair.pub, keypair.sec);
                const connection = WS2PConnection_1.WS2PConnection.newConnectionToAddress(1, `ws://${host}:${port}${path && ' ' + path || ''}`, new (class SyncMessageHandler {
                    async answerToRequest(json, c) {
                        throw Error(errors_1.DataErrors[errors_1.DataErrors.WS2P_SYNC_PERIMETER_IS_LIMITED]);
                    }
                    async handlePushMessage(json, c) {
                        logger.warn('Receiving push messages, which are not allowed during a SYNC.', json);
                    }
                }), new WS2PConnection_1.WS2PPubkeySyncLocalAuth("", pair, '00000000'), new WS2PConnection_1.WS2PPubkeyRemoteAuth("", pair), // The currency will be set by the remote node
                undefined);
                try {
                    const requester = WS2PRequester_1.WS2PRequester.fromConnection(connection);
                    peering = await requester.getPeer();
                    api = new WS2PRemoteContacter_1.WS2PRemoteContacter(requester);
                    endpoint = 'WS2P 99999999 ' + host + ' ' + port + ((path && ' ' + path) || '');
                }
                catch (e) {
                }
            }
            // If we have a working API: stop!
            if (api && peering) {
                break;
            }
        }
        if (!api) {
            throw Error(errors_1.DataErrors[errors_1.DataErrors.CANNOT_CONNECT_TO_REMOTE_FOR_SYNC]);
        }
        if (!peering) {
            throw Error(errors_1.DataErrors[errors_1.DataErrors.NO_PEERING_AVAILABLE_FOR_SYNC]);
        }
        return {
            api,
            peering,
            endpoint,
            currency: peering.currency
        };
    }
    async initWithKnownLocalAndToAndCurrency(to, localNumber) {
        this.to = to;
        this.localNumber = localNumber;
        //=======
        // Peers (just for P2P download)
        //=======
        let peers = [];
        const p2psync = !programOptions_1.cliprogram.nop2p;
        if (p2psync) {
            this.watcher.writeStatus('Peers...');
            peers = await this.node.getPeers();
        }
        // Add current peer if it is not returned (example of a local node)
        peers.push({
            version: 1,
            currency: '',
            status: 'UP',
            first_down: null,
            last_try: null,
            pubkey: '',
            block: '',
            signature: '',
            endpoints: [this.endpoint]
        });
        peers = peers.filter(p => {
            if (!p)
                return false;
            let hasWS2P = false;
            let hasBMA = false;
            for (const e of p.endpoints) {
                if (e.indexOf('MERKLED')) {
                    hasBMA = true;
                }
                if (e.indexOf('WS2P') !== -1) {
                    hasWS2P = true;
                }
            }
            return (hasWS2P || hasBMA) && p.status === 'UP';
        });
        if (!peers.length) {
            peers.push(DBPeer_1.DBPeer.fromPeerDTO(this.peer));
        }
        this.shuffledPeers = (this.noShufflePeers ? peers : underscore_1.Underscore.shuffle(peers)).filter(p => !!(p));
    }
    p2pDownloader() {
        if (!this.theP2pDownloader) {
            this.theP2pDownloader = new P2PSyncDownloader_1.P2PSyncDownloader(this.currency, this.server.conf.pair, this.localNumber, this.to, this.shuffledPeers, this.watcher, logger, this.chunkSize, this.allowLocalSync);
        }
        return this.theP2pDownloader;
    }
    fsDownloader() {
        if (!this.theFsDownloader) {
            this.theFsDownloader = new FsSyncDownloader_1.FsSyncDownloader(this.readDAL.fs, path.join(this.readDAL.rootPath, this.getChunksPath()), this.getChunkName.bind(this), this.chunkSize);
        }
        return this.theFsDownloader;
    }
    getCurrent() {
        return this.node.getCurrent();
    }
    getBlock(number) {
        return this.node.getBlock(number);
    }
    async getMilestone(number) {
        if (this.hasMilestonesPages === undefined) {
            try {
                const mlPage = await this.node.getMilestonesPage();
                this.hasMilestonesPages = mlPage.chunkSize === this.chunkSize;
                this.milestonesPerPage = mlPage.milestonesPerPage;
                this.maxPage = mlPage.totalPages;
            }
            catch (e) {
                this.hasMilestonesPages = false;
            }
        }
        if (!this.hasMilestonesPages) {
            return this.getBlock(number);
        }
        if (this.milestones[number]) {
            return this.milestones[number];
        }
        if ((number + 1) % this.chunkSize !== 0) {
            // Something went wrong: we cannot rely on milestones method
            this.hasMilestonesPages = false;
            return this.getBlock(number);
        }
        const chunkNumber = (number + 1) / this.chunkSize;
        const pageNumber = (chunkNumber - (chunkNumber % this.milestonesPerPage)) / this.milestonesPerPage + 1;
        if (pageNumber > this.maxPage) {
            // The page is not available: we cannot rely on milestones method at this point
            this.hasMilestonesPages = false;
            return this.getBlock(number);
        }
        const mlPage = await this.node.getMilestones(pageNumber);
        mlPage.blocks.forEach(b => this.milestones[b.number] = b);
        if (this.milestones[number]) {
            return this.milestones[number];
        }
        // Even after the download, it seems we don't have our milestone. We will download normally.
        this.hasMilestonesPages = false;
        return this.getBlock(number);
    }
    static async test(host, port, keypair) {
        const syncApi = await RemoteSynchronizer.getSyncAPI([{ host, port }], keypair);
        const current = await syncApi.api.getCurrent();
        if (!current) {
            throw Error(errors_1.DataErrors[errors_1.DataErrors.REMOTE_HAS_NO_CURRENT_BLOCK]);
        }
        return current;
    }
    async syncPeers(fullSync, to) {
        const peers = await this.node.getPeers();
        for (let i = 0; i < peers.length; i++) {
            const peer = PeerDTO_1.PeerDTO.fromJSONObject(peers[i]);
            this.watcher.writeStatus('Peer ' + peer.pubkey);
            this.watcher.peersPercent(Math.ceil(i / peers.length * 100));
            try {
                await this.PeeringService.submitP(DBPeer_1.DBPeer.fromPeerDTO(peer));
            }
            catch (e) {
            }
        }
        this.watcher.peersPercent(100);
    }
    async syncSandbox() {
        this.watcher.writeStatus('Synchronizing the sandboxes...');
        await sandbox_1.pullSandboxToLocalServer(this.currency, this.node, this.server, this.server.logger, this.watcher, 1, false);
    }
}
exports.RemoteSynchronizer = RemoteSynchronizer;
//# sourceMappingURL=RemoteSynchronizer.js.map