"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const querable_1 = require("../../../../../lib/common-libs/querable");
const RemoteSynchronizer_1 = require("../RemoteSynchronizer");
const timeout_promise_1 = require("../../../../../lib/common-libs/timeout-promise");
class P2pCandidate {
    constructor(p, keypair, logger, allowLocalSync) {
        this.p = p;
        this.keypair = keypair;
        this.logger = logger;
        this.allowLocalSync = allowLocalSync;
        this.responseTimes = [];
        this.nbSuccess = 0;
        this.failures = 0;
        this.reserved = false;
        this.apiPromise = this.initAPI();
        this.dlPromise = querable_1.querablep(Promise.resolve(null));
    }
    addFailure() {
        this.failures++;
        if (this.failures >= 5 && !this.isExcluded) {
            this.isExcluded = true;
            this.logger.warn('Excluding node %s as it returned unchainable chunks %s times', this.hostName, this.failures);
        }
    }
    isReady() {
        return !this.reserved && this.apiPromise.isResolved() && this.dlPromise.isResolved() && this.api && !this.isExcluded;
    }
    async waitAvailability(maxWait) {
        return Promise.race([
            // Wait for availablity
            (async () => !this.isExcluded
                && !this.reserved
                && (this.apiPromise.isRejected() ? await timeout_promise_1.newResolveTimeoutPromise(maxWait, false) : !!(await this.apiPromise))
                && (this.dlPromise.isRejected() ? await timeout_promise_1.newResolveTimeoutPromise(maxWait, false) : !!(await this.dlPromise)))(),
            // Maximum wait trigger
            timeout_promise_1.newResolveTimeoutPromise(maxWait, false)
        ]);
    }
    hasAvailableApi() {
        return !!this.api;
    }
    avgResponseTime() {
        if (!this.responseTimes.length) {
            return 0;
        }
        return this.responseTimes.reduce((sum, rt) => sum + rt, 0) / this.responseTimes.length;
    }
    get hostName() {
        return (this.api && this.api.hostName) || 'NO_API';
    }
    async downloadBlocks(count, from) {
        const start = Date.now();
        let error;
        this.reserved = false;
        this.dlPromise = querable_1.querablep((async () => {
            // We try to download the blocks
            let blocks;
            try {
                blocks = await this.api.getBlocks(count, from);
            }
            catch (e) {
                // Unfortunately this can fail
                blocks = null;
                error = e;
            }
            this.responseTimes.push(Date.now() - start);
            // Only keep a flow of 5 ttas for the node
            if (this.responseTimes.length > 5)
                this.responseTimes.shift();
            this.nbSuccess++;
            if (error) {
                throw error;
            }
            return blocks;
        })());
        return this.dlPromise;
    }
    getRemoteAPIs() {
        const bmaAPI = this.p.getBMA();
        const ws2pAPI = this.p.getFirstNonTorWS2P();
        const apis = [];
        const bmaHost = bmaAPI.dns || bmaAPI.ipv4 || bmaAPI.ipv6;
        if (bmaAPI.port && bmaHost) {
            apis.push({
                isBMA: true,
                port: bmaAPI.port,
                host: bmaHost
            });
        }
        if (ws2pAPI) {
            apis.push({
                isWS2P: true,
                host: ws2pAPI.host,
                port: ws2pAPI.port,
                path: ws2pAPI.path,
            });
        }
        return apis;
    }
    initAPI() {
        return querable_1.querablep((async () => {
            try {
                const apis = this.getRemoteAPIs();
                const syncApi = await RemoteSynchronizer_1.RemoteSynchronizer.getSyncAPI(apis, this.keypair);
                if (!this.allowLocalSync && ((syncApi && syncApi.api.hostName || '').match(/^(localhost|192|127)/))) {
                    return null;
                }
                this.api = syncApi.api;
                return syncApi.api;
            }
            catch (e) {
                return null;
            }
        })());
    }
    reserve() {
        this.reserved = true;
    }
    get apiName() {
        return this.api && this.api.type;
    }
}
exports.P2pCandidate = P2pCandidate;
//# sourceMappingURL=p2p-candidate.js.map