"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const stream_1 = require("stream");
const querable_1 = require("../../../../../lib/common-libs/querable");
const errors_1 = require("../../../../../lib/common-libs/errors");
const logger_1 = require("../../../../../lib/logger");
const exit_codes_1 = require("../../../../../lib/common-libs/exit-codes");
class ValidatorStream extends stream_1.Readable {
    constructor(localNumber, to, toHash, syncStrategy, watcher) {
        super({ objectMode: true });
        this.localNumber = localNumber;
        this.to = to;
        this.toHash = toHash;
        this.syncStrategy = syncStrategy;
        this.watcher = watcher;
        this.currentChunkNumber = 0;
        this.bestDownloaded = -1;
        const nbBlocksToDownload = Math.max(0, to - localNumber);
        this.numberOfChunksToDownload = Math.ceil(nbBlocksToDownload / syncStrategy.chunkSize);
        this.chunks = Array.from({ length: this.numberOfChunksToDownload });
        this.dowloading = Array.from({ length: this.numberOfChunksToDownload });
        this.fsSynchronizer = syncStrategy.fsDownloader();
        this.downloadBlock(0);
    }
    async downloadBlock(i, forceDownload = false) {
        const maximumCacheNumber = forceDownload ? -1 : await this.validateCacheLevel();
        if (i + 1 > this.numberOfChunksToDownload) {
            return Promise.resolve();
        }
        if (!this.dowloading[i] && !this.chunks[i]) {
            this.dowloading[i] = querable_1.querablep((async () => {
                let failures = 0;
                let block;
                do {
                    try {
                        const bNumber = Math.min(this.to, (i + 1) * this.syncStrategy.chunkSize - 1);
                        if (bNumber > maximumCacheNumber) {
                            block = await this.syncStrategy.getMilestone(bNumber);
                        }
                        else {
                            block = await this.getBlockFromCache(bNumber);
                        }
                        if (!forceDownload && i > this.bestDownloaded) {
                            this.watcher.storagePercent(Math.round((i + 1) / this.numberOfChunksToDownload * 100));
                            this.bestDownloaded = i;
                        }
                        if (!block) {
                            throw Error(errors_1.DataErrors[errors_1.DataErrors.CANNOT_GET_VALIDATION_BLOCK_FROM_REMOTE]);
                        }
                    }
                    catch (e) {
                        failures++;
                        await new Promise((res) => setTimeout(res, 3000));
                        if (failures >= 15) {
                            logger_1.NewLogger().error('Could not get a validation from remote blockchain after %s trials. Stopping sync.', failures);
                            process.exit(exit_codes_1.ExitCodes.SYNC_FAIL);
                        }
                        block = null;
                    }
                } while (!block);
                return block;
            })());
            this.dowloading[i]
                .then(chunk => {
                this.chunks[i] = chunk;
                delete this.dowloading[i];
                // this.push(chunk)
            })
                .catch(err => {
                throw err;
            });
            return this.dowloading[i] || this.chunks[i];
        }
        return this.dowloading[i] || this.chunks[i];
    }
    async getBlockFromCache(bNumber) {
        return this.fsSynchronizer.getBlock(bNumber);
    }
    _read(size) {
        if (this.currentChunkNumber == this.numberOfChunksToDownload) {
            this.push(null);
        }
        else {
            // Asks for next chunk: do we have it?
            if (this.chunks[this.currentChunkNumber]) {
                this.push(this.chunks[this.currentChunkNumber]);
                delete this.chunks[this.currentChunkNumber];
                // Let's start the download of next chunk
                this.currentChunkNumber++;
                this.downloadBlock(this.currentChunkNumber)
                    .then(() => this.downloadBlock(this.currentChunkNumber + 1))
                    .then(() => this.downloadBlock(this.currentChunkNumber + 2))
                    .then(() => this.downloadBlock(this.currentChunkNumber + 3))
                    .then(() => this.downloadBlock(this.currentChunkNumber + 4))
                    .then(() => this.downloadBlock(this.currentChunkNumber + 5))
                    .then(() => this.downloadBlock(this.currentChunkNumber + 6));
            }
            else {
                // We don't have it yet
                this.push(undefined);
            }
        }
    }
    async validateCacheLevel() {
        if (!this.cacheLevelValidationPromise) {
            this.cacheLevelValidationPromise = (async () => {
                // Find the best common chunk with remote
                let topChunk = this.numberOfChunksToDownload - 1; // We ignore the top chunk, which is special (most unlikely to be full)
                let botChunk = -1; // In the worst case, this is the good index
                let current;
                do {
                    current = topChunk - ((topChunk - botChunk) % 2 == 0 ? (topChunk - botChunk) / 2 : ((topChunk - botChunk + 1) / 2) - 1);
                    if (current === 0) {
                        // we have no compliant cache
                        return -1;
                    }
                    const bNumber = current * this.syncStrategy.chunkSize - 1;
                    const remoteBlock = await this.downloadBlock(current - 1, true);
                    const localBlock = await this.fsSynchronizer.getBlock(bNumber);
                    if (remoteBlock && localBlock && remoteBlock.hash === localBlock.hash) {
                        // Success! Let's look forward
                        botChunk = current;
                    }
                    else {
                        // Fail: let's look backward
                        topChunk = current - 1;
                    }
                } while (botChunk !== topChunk);
                // retur topChunk or botChunk, it is the same
                return topChunk === -1 ? -1 : (topChunk * this.syncStrategy.chunkSize) - 1;
            })();
        }
        return this.cacheLevelValidationPromise;
    }
}
exports.ValidatorStream = ValidatorStream;
//# sourceMappingURL=ValidatorStream.js.map