"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const blockGenerator_1 = require("./lib/blockGenerator");
const constants_1 = require("./lib/constants");
const blockProver_1 = require("./lib/blockProver");
const prover_1 = require("./lib/prover");
const contacter_1 = require("../crawler/lib/contacter");
const index_1 = require("../../lib/common-libs/parsers/index");
const PeerDTO_1 = require("../../lib/dto/PeerDTO");
const async = require('async');
exports.ProverDependency = {
    duniter: {
        /*********** Permanent prover **************/
        config: {
            onLoading: async (conf) => {
                if (conf.cpu === null || conf.cpu === undefined) {
                    conf.cpu = constants_1.ProverConstants.DEFAULT_CPU;
                }
                if (conf.nbCores === null || conf.nbCores === undefined) {
                    conf.nbCores = Math.min(constants_1.ProverConstants.CORES_MAXIMUM_USE_IN_PARALLEL, require('os').cpus().length);
                }
                else if (conf.nbCores <= 0) {
                    conf.nbCores = 1;
                }
                if (conf.prefix === null || conf.prefix === undefined) {
                    conf.prefix = constants_1.ProverConstants.DEFAULT_PEER_ID;
                }
                conf.powSecurityRetryDelay = constants_1.ProverConstants.POW_SECURITY_RETRY_DELAY;
                conf.powMaxHandicap = constants_1.ProverConstants.POW_MAXIMUM_ACCEPTABLE_HANDICAP;
            },
            beforeSave: async (conf) => {
                delete conf.powSecurityRetryDelay;
                delete conf.powMaxHandicap;
            }
        },
        service: {
            output: (server) => {
                const generator = new blockGenerator_1.BlockGenerator(server);
                server.generatorGetJoinData = generator.getSinglePreJoinData.bind(generator);
                server.generatorComputeNewCerts = generator.computeNewCerts.bind(generator);
                server.generatorNewCertsToLinks = generator.newCertsToLinks.bind(generator);
                return new prover_1.Prover(server);
            }
        },
        methods: {
            hookServer: (server) => {
                const generator = new blockGenerator_1.BlockGenerator(server);
                server.generatorGetJoinData = generator.getSinglePreJoinData.bind(generator);
                server.generatorComputeNewCerts = generator.computeNewCerts.bind(generator);
                server.generatorNewCertsToLinks = generator.newCertsToLinks.bind(generator);
            },
            prover: (server) => new prover_1.Prover(server),
            blockGenerator: (server, prover) => new blockGenerator_1.BlockGeneratorWhichProves(server, prover),
            generateTheNextBlock: async (server, manualValues) => {
                const prover = new blockProver_1.BlockProver(server);
                const generator = new blockGenerator_1.BlockGeneratorWhichProves(server, prover);
                return generator.nextBlock(manualValues);
            },
            generateAndProveTheNext: async (server, block, trial, manualValues) => {
                const prover = new blockProver_1.BlockProver(server);
                const generator = new blockGenerator_1.BlockGeneratorWhichProves(server, prover);
                let res = await generator.makeNextBlock(block, trial, manualValues);
                return res;
            }
        },
        /*********** CLI gen-next + gen-root **************/
        cliOptions: [
            { value: '--show', desc: 'With gen-* commands, displays the generated block.' },
            { value: '--check', desc: 'With gen-* commands: just check validity of generated block.' },
            { value: '--submit-local', desc: 'With gen-* commands: the generated block is submitted to this node only.' },
            { value: '--submit-host <host>', desc: 'With gen-* commands: the generated block is submitted to `submit-host` node.' },
            { value: '--submit-port <port>', desc: 'With gen-* commands: the generated block is submitted to `submit-host` node with port `submit-port`.' },
            { value: '--at <medianTime>', desc: 'With gen-next --show --check: allows to try in a future time.', parser: parseInt }
        ],
        cli: [{
                name: 'gen-next [difficulty]',
                desc: 'Tries to generate the next block of the blockchain.',
                onDatabaseExecute: async (server, conf, program, params) => {
                    const difficulty = params[0];
                    const generator = new blockGenerator_1.BlockGeneratorWhichProves(server, null);
                    return generateAndSend(program, difficulty, server, () => () => generator.nextBlock());
                }
            }, {
                name: 'bc-resolve',
                desc: 'Tries to resolve next blocks or forks.',
                onDatabaseExecute: async () => { }
            }, {
                name: 'gen-root [difficulty]',
                desc: 'Tries to generate the next block of the blockchain.',
                preventIfRunning: true,
                onDatabaseExecute: async (server, conf, program, params) => {
                    const difficulty = params[0];
                    const generator = new blockGenerator_1.BlockGeneratorWhichProves(server, null);
                    let toDelete = [], catched = true;
                    do {
                        try {
                            await generateAndSend(program, difficulty, server, () => () => generator.nextBlock());
                            catched = false;
                        }
                        catch (e) {
                            toDelete = await server.dal.idtyDAL.query('SELECT * FROM idty i WHERE 5 > (SELECT count(*) from cert c where c.`to` = i.pubkey)');
                            console.log('Deleting', toDelete.map((i) => i.pubkey));
                            await server.dal.idtyDAL.exec('DELETE FROM idty WHERE pubkey IN (' + toDelete.map((i) => "'" + i.pubkey + "'").join(',') + ')');
                            await server.dal.idtyDAL.exec('DELETE FROM cert WHERE `to` IN (' + toDelete.map((i) => "'" + i.pubkey + "'").join(',') + ')');
                            await server.dal.idtyDAL.exec('DELETE FROM cert WHERE `from` IN (' + toDelete.map((i) => "'" + i.pubkey + "'").join(',') + ')');
                        }
                    } while (catched && toDelete.length);
                    console.log('Done');
                }
            }, {
                name: 'gen-root-choose [difficulty]',
                desc: 'Tries to generate root block, with choice of root members.',
                preventIfRunning: true,
                onDatabaseExecute: async (server, conf, program, params) => {
                    const difficulty = params[0];
                    if (!difficulty) {
                        throw 'Difficulty is required.';
                    }
                    const generator = new blockGenerator_1.BlockGenerator(server);
                    return generateAndSend(program, difficulty, server, () => () => generator.manualRoot());
                }
            }]
    }
};
function generateAndSend(program, difficulty, server, getGenerationMethod) {
    const logger = server.logger;
    return new Promise((resolve, reject) => {
        if (!program.submitLocal) {
            if (!program.submitHost) {
                throw 'Option --submit-host is required.';
            }
            if (!program.submitPort) {
                throw 'Option --submit-port is required.';
            }
            if (isNaN(parseInt(program.submitPort))) {
                throw 'Option --submit-port must be a number.';
            }
        }
        async.waterfall([
            function (next) {
                const method = getGenerationMethod(server);
                (async () => {
                    const simulationValues = {};
                    if (program.show && program.check) {
                        if (program.at && !isNaN(program.at)) {
                            simulationValues.medianTime = program.at;
                        }
                    }
                    const block = await method(null, simulationValues);
                    next(null, block);
                })();
            },
            function (block, next) {
                if (program.check) {
                    block.time = block.medianTime;
                    program.show && console.log(block.getRawSigned());
                    (async () => {
                        try {
                            const parsed = index_1.parsers.parseBlock.syncWrite(block.getRawSigned());
                            await server.BlockchainService.checkBlock(parsed, false);
                            logger.info('Acceptable block');
                            next();
                        }
                        catch (e) {
                            next(e);
                        }
                    })();
                }
                else {
                    logger.debug('Block to be sent: %s', block.getRawInnerPart());
                    async.waterfall([
                        function (subNext) {
                            proveAndSend(program, server, block, server.conf.pair.pub, parseInt(difficulty), subNext);
                        }
                    ], next);
                }
            }
        ], (err, data) => {
            err && reject(err);
            !err && resolve(data);
        });
    });
}
function proveAndSend(program, server, block, issuer, difficulty, done) {
    const logger = server.logger;
    async.waterfall([
        function (next) {
            block.issuer = issuer;
            program.show && console.log(block.getRawSigned());
            (async () => {
                try {
                    const host = program.submitHost;
                    const port = program.submitPort;
                    const trialLevel = isNaN(difficulty) ? await server.getBcContext().getIssuerPersonalizedDifficulty(server.PeeringService.selfPubkey) : difficulty;
                    const prover = new blockProver_1.BlockProver(server);
                    const proven = await prover.prove(block, trialLevel);
                    if (program.submitLocal) {
                        await server.writeBlock(proven);
                        next();
                    }
                    else {
                        const peer = PeerDTO_1.PeerDTO.fromJSONObject({
                            endpoints: [['BASIC_MERKLED_API', host, port].join(' ')]
                        });
                        program.show && console.log(proven.getRawSigned());
                        logger.info('Posted block ' + proven.getRawSigned());
                        const p = PeerDTO_1.PeerDTO.fromJSONObject(peer);
                        const contact = new contacter_1.Contacter(p.getHostPreferDNS(), p.getPort());
                        await contact.postBlock(proven.getRawSigned());
                        next();
                    }
                }
                catch (e) {
                    next(e);
                }
            })();
        }
    ], done);
}
//# sourceMappingURL=index.js.map