"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const constants_1 = require("./constants");
const engine_1 = require("./engine");
const constants_2 = require("../../../lib/common-libs/constants");
const BlockDTO_1 = require("../../../lib/dto/BlockDTO");
const os = require('os');
const querablep = require('querablep');
const POW_FOUND = true;
const POW_NOT_FOUND_YET = false;
class WorkerFarm {
    constructor(server, logger) {
        this.server = server;
        this.logger = logger;
        this.onAlmostPoW = null;
        this.powPromise = null;
        this.stopPromise = null;
        this.checkPoWandNotify = null;
        this.theEngine = new engine_1.PowEngine(server.conf, server.logger, server.dal);
        // An utility method to filter the pow notifications
        this.checkPoWandNotify = (hash, block, found) => {
            const matches = hash.match(/^(0{2,})[^0]/);
            if (matches && this.onAlmostPoW) {
                this.onAlmostPoW(hash, matches, block, found);
            }
        };
        // Keep track of PoW advancement
        this.theEngine.setOnInfoMessage((message) => {
            if (message.error) {
                this.logger.error('Error in engine#%s:', this.theEngine.id, message.error);
            }
            else if (message.pow) {
                // A message about the PoW
                const msg = message.pow;
                this.checkPoWandNotify(msg.pow, msg.block, POW_NOT_FOUND_YET);
            }
        });
    }
    get nbWorkers() {
        return this.theEngine.getNbWorkers();
    }
    changeCPU(cpu) {
        return this.theEngine.setConf({ cpu });
    }
    changePoWPrefix(prefix) {
        return this.theEngine.setConf({ prefix });
    }
    isComputing() {
        return this.powPromise !== null && !this.powPromise.isResolved();
    }
    isStopping() {
        return this.stopPromise !== null && !this.stopPromise.isResolved();
    }
    /**
     * Eventually stops the engine PoW if one was computing
     */
    stopPoW() {
        this.stopPromise = querablep(this.theEngine.cancel());
        return this.stopPromise;
    }
    shutDownEngine() {
        return this.theEngine.shutDown();
    }
    /**
     * Starts a new computation of PoW
     * @param stuff The necessary data for computing the PoW
     */
    async askNewProof(stuff) {
        // Starts the PoW
        this.powPromise = querablep(this.theEngine.prove(stuff));
        const res = await this.powPromise;
        if (res) {
            this.checkPoWandNotify(res.pow.pow, res.pow.block, POW_FOUND);
        }
        return res && res.pow;
    }
    setOnAlmostPoW(onPoW) {
        this.onAlmostPoW = onPoW;
    }
}
exports.WorkerFarm = WorkerFarm;
class BlockProver {
    constructor(server) {
        this.server = server;
        this.logger = server.logger;
        const debug = process.execArgv.toString().indexOf('--debug') !== -1;
        if (debug) {
            //Set an unused port number.
            process.execArgv = [];
        }
    }
    get conf() {
        return this.server.conf;
    }
    get pair() {
        return this.conf.pair;
    }
    getWorker() {
        if (!this.workerFarmPromise) {
            this.workerFarmPromise = (async () => {
                return new WorkerFarm(this.server, this.logger);
            })();
        }
        return this.workerFarmPromise;
    }
    async cancel() {
        // If no farm was instanciated, there is nothing to do yet
        if (this.workerFarmPromise) {
            let farm = await this.getWorker();
            await farm.stopPoW();
            if (this.waitResolve) {
                this.waitResolve();
                this.waitResolve = null;
            }
        }
    }
    prove(block, difficulty, forcedTime = null) {
        if (this.waitResolve) {
            this.waitResolve();
            this.waitResolve = null;
        }
        const remainder = difficulty % 16;
        const nbZeros = (difficulty - remainder) / 16;
        const highMark = constants_2.CommonConstants.PROOF_OF_WORK.UPPER_BOUND[remainder];
        const notifyVersionJumpReady = block.version === 10 && constants_2.CommonConstants.BLOCK_NEW_GENERATED_VERSION === 11;
        return (async () => {
            let powFarm = await this.getWorker();
            if (block.number == 0) {
                // On initial block, difficulty is the one given manually
                block.powMin = difficulty;
            }
            // Start
            powFarm.setOnAlmostPoW((pow, matches, aBlock, found) => {
                this.powEvent(found, pow);
                if (matches && matches[1].length >= constants_1.ProverConstants.MINIMAL_ZEROS_TO_SHOW_IN_LOGS) {
                    this.logger.info('Matched %s zeros %s with Nonce = %s for block#%s by %s', matches[1].length, pow, aBlock.nonce, aBlock.number, aBlock.issuer.slice(0, 6));
                }
            });
            block.nonce = 0;
            this.logger.info('Generating proof-of-work with %s leading zeros followed by [0-' + highMark + ']... (CPU usage set to %s%) for block#%s', nbZeros, (this.conf.cpu * 100).toFixed(0), block.number, block.issuer.slice(0, 6));
            const start = Date.now();
            let result = await powFarm.askNewProof({
                newPoW: {
                    conf: {
                        powNoSecurity: this.conf.powNoSecurity,
                        cpu: this.conf.cpu,
                        prefix: this.conf.prefix ? String(this.conf.prefix) : '',
                        avgGenTime: this.conf.avgGenTime,
                        medianTimeBlocks: this.conf.medianTimeBlocks
                    },
                    block: block,
                    zeros: nbZeros,
                    highMark: highMark,
                    forcedTime: forcedTime,
                    pair: this.pair
                },
                specialNonce: notifyVersionJumpReady ? 999 * (constants_1.ProverConstants.NONCE_RANGE / 1000) : 0,
            });
            if (!result) {
                this.logger.info('GIVEN proof-of-work for block#%s with %s leading zeros followed by [0-' + highMark + ']! stop PoW for %s', block.number, nbZeros, this.pair && this.pair.pub.slice(0, 6));
                throw 'Proof-of-work computation canceled because block received';
            }
            else {
                const proof = result.block;
                const testsCount = result.testsCount * powFarm.nbWorkers;
                const duration = (Date.now() - start);
                const testsPerSecond = testsCount / (duration / 1000);
                this.logger.info('Done: #%s, %s in %ss (~%s tests, ~%s tests/s, using %s cores, CPU %s%)', block.number, proof.hash, (duration / 1000).toFixed(2), testsCount, testsPerSecond.toFixed(2), powFarm.nbWorkers, Math.floor(100 * this.conf.cpu));
                this.logger.info('FOUND proof-of-work with %s leading zeros followed by [0-' + highMark + ']!', nbZeros);
                return BlockDTO_1.BlockDTO.fromJSONObject(proof);
            }
        })();
    }
    ;
    async changeCPU(cpu) {
        this.conf.cpu = cpu;
        const farm = await this.getWorker();
        return farm.changeCPU(cpu);
    }
    async changePoWPrefix(prefix) {
        this.conf.prefix = prefix;
        const farm = await this.getWorker();
        return farm.changePoWPrefix(prefix);
    }
    powEvent(found, hash) {
        this.server && this.server.push({ pow: { found, hash } });
    }
}
exports.BlockProver = BlockProver;
//# sourceMappingURL=blockProver.js.map