"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const constants_1 = require("./constants");
const proof_1 = require("./proof");
const PowWorker_1 = require("./PowWorker");
const underscore_1 = require("../../../lib/common-libs/underscore");
const exit_codes_1 = require("../../../lib/common-libs/exit-codes");
const nuuid = require('node-uuid');
const cluster = require('cluster');
const querablep = require('querablep');
let clusterId = 0;
cluster.setMaxListeners(3);
/**
 * Cluster controller, handles the messages between the main program and the PoW cluster.
 */
class Master {
    constructor(nbCores, logger, dal) {
        this.nbCores = nbCores;
        this.dal = dal;
        this.nbCancels = 0;
        this.currentPromise = null;
        this.slaves = [];
        this.slavesMap = {};
        this.conf = {};
        this.clusterId = clusterId++;
        this.logger = logger || Master.defaultLogger();
        this.onInfoMessage = (message) => {
            this.logger.info(`${message.pow.pow} nonce = ${message.pow.block.nonce}`);
        };
    }
    get nbWorkers() {
        return this.slaves.length;
    }
    set onInfoMessage(callback) {
        this.onInfoCallback = callback;
    }
    onWorkerMessage(workerIndex, message) {
        // this.logger.info(`worker#${this.slavesMap[worker.id].index} sent message:${message}`)
        if (message && message.pow) {
            this.onInfoCallback && this.onInfoCallback(message);
        }
        if (this.currentPromise && message.uuid && !this.currentPromise.isResolved() && message.answer) {
            this.logger.info(`ENGINE c#${this.clusterId}#${workerIndex} HAS FOUND A PROOF #${message.answer.pow.pow}`);
        }
        else if (message.canceled) {
            this.nbCancels++;
        }
        // this.logger.debug(`ENGINE c#${this.clusterId}#${this.slavesMap[worker.id].index}:`, message)
    }
    /*****************
     * CLUSTER METHODS
     ****************/
    initCluster() {
        // Setup master
        cluster.setupMaster({
            exec: __filename,
            execArgv: [] // Do not try to debug forks
        });
        const nbCores = this.nbCores !== undefined && this.nbCores !== null ? this.nbCores : 1;
        this.slaves = Array.from({ length: nbCores }).map((value, index) => {
            const nodejsWorker = cluster.fork();
            const worker = new PowWorker_1.PowWorker(nodejsWorker, message => {
                this.onWorkerMessage(index, message);
            }, () => {
                this.logger.info(`[online] worker c#${this.clusterId}#w#${index}`);
                worker.sendConf({
                    rootPath: this.dal ? this.dal.rootPath : '',
                    command: 'conf',
                    value: this.conf
                });
            }, (code, signal) => {
                this.logger.info(`worker ${worker.pid} died with code ${code} and signal ${signal}`);
            });
            this.logger.info(`Creating worker c#${this.clusterId}#w#${nodejsWorker.id}`);
            const slave = {
                // The Node.js worker
                worker,
                // Inner identifier
                index,
                // Worker ready
                online: worker.online,
                // Each worker has his own chunk of possible nonces
                nonceBeginning: this.nbCores === 1 ? 0 : (index + 1) * constants_1.ProverConstants.NONCE_RANGE
            };
            this.slavesMap[nodejsWorker.id] = slave;
            return slave;
        });
        this.workersOnline = this.slaves.map((s) => s.online);
        return Promise.all(this.workersOnline);
    }
    changeConf(conf) {
        this.logger.info(`Changing conf to: ${JSON.stringify(conf)} on PoW cluster`);
        this.conf.cpu = conf.cpu || this.conf.cpu;
        this.conf.prefix = this.conf.prefix || conf.prefix;
        this.slaves.forEach(s => {
            s.worker.sendConf({
                rootPath: '',
                command: 'conf',
                value: this.conf
            });
        });
        return Promise.resolve(underscore_1.Underscore.clone(conf));
    }
    cancelWorkersWork() {
        this.slaves.forEach(s => {
            s.worker.sendCancel();
        });
        if (this.dal) {
            this.dal.powDAL.writeCurrent("");
        }
    }
    async cancelWork() {
        const workEnded = this.currentPromise;
        // Don't await the cancellation!
        this.cancelWorkersWork();
        // Current promise is done
        this.currentPromise = null;
        return await workEnded;
    }
    async shutDownWorkers() {
        if (this.workersOnline) {
            await Promise.all(this.workersOnline);
            await Promise.all(this.slaves.map(async (s) => {
                s.worker.kill();
            }));
        }
        this.slaves = [];
    }
    async proveByWorkers(stuff) {
        // Eventually spawn the workers
        if (this.slaves.length === 0) {
            this.initCluster();
        }
        if (this.dal) {
            await this.dal.powDAL.writeCurrent([stuff.newPoW.block.number - 1, stuff.newPoW.block.previousHash].join('-'));
        }
        // Register the new proof uuid
        const uuid = nuuid.v4();
        this.currentPromise = querablep((async () => {
            await Promise.all(this.workersOnline);
            if (!this.currentPromise) {
                this.logger.info(`Proof canceled during workers' initialization`);
                return null;
            }
            // Start the salves' job
            const asks = this.slaves.map(async (s, index) => {
                const nonceBeginning = stuff.specialNonce || s.nonceBeginning;
                const proof = await s.worker.askProof({
                    uuid,
                    command: 'newPoW',
                    value: {
                        rootPath: this.dal ? this.dal.rootPath : '',
                        initialTestsPerRound: stuff.initialTestsPerRound,
                        maxDuration: stuff.maxDuration,
                        block: stuff.newPoW.block,
                        nonceBeginning,
                        zeros: stuff.newPoW.zeros,
                        highMark: stuff.newPoW.highMark,
                        pair: underscore_1.Underscore.clone(stuff.newPoW.pair),
                        forcedTime: stuff.newPoW.forcedTime,
                        conf: {
                            powNoSecurity: stuff.newPoW.conf.powNoSecurity,
                            medianTimeBlocks: stuff.newPoW.conf.medianTimeBlocks,
                            avgGenTime: stuff.newPoW.conf.avgGenTime,
                            cpu: stuff.newPoW.conf.cpu,
                            prefix: stuff.newPoW.conf.prefix
                        }
                    }
                });
                this.logger.info(`[done] worker c#${this.clusterId}#w#${index}`);
                return {
                    workerID: index,
                    proof
                };
            });
            // Find a proof
            const result = await Promise.race(asks);
            // Don't await the cancellation!
            this.cancelWorkersWork();
            // Wait for all workers to have stopped looking for a proof
            await Promise.all(asks);
            if (!result.proof || !result.proof.message.answer) {
                this.logger.info('No engine found the proof. It was probably cancelled.');
                return null;
            }
            else {
                this.logger.info(`ENGINE c#${this.clusterId}#${result.workerID} HAS FOUND A PROOF #${result.proof.message.answer.pow.pow}`);
                return result.proof.message.answer;
            }
        })());
        return this.currentPromise;
    }
    static defaultLogger() {
        return {
            info: (message) => { }
        };
    }
}
exports.Master = Master;
if (cluster.isMaster) {
    // Super important for Node.js debugging
    const debug = process.execArgv.toString().indexOf('--debug') !== -1;
    if (debug) {
        //Set an unused port number.
        process.execArgv = [];
    }
}
else {
    process.on("SIGTERM", function () {
        process.exit(exit_codes_1.ExitCodes.OK);
    });
    proof_1.createPowWorker();
}
//# sourceMappingURL=powCluster.js.map