"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const os = require("os");
const underscore_1 = require("../lib/common-libs/underscore");
const upnp = require('nat-upnp');
class UpnpProvider {
    constructor(portStart, portEnd, identifier, upnpInterval = 300, ttl = 600, logger, host = '') {
        this.portStart = portStart;
        this.portEnd = portEnd;
        this.identifier = identifier;
        this.upnpInterval = upnpInterval;
        this.ttl = ttl;
        this.logger = logger;
        this.host = host;
        this.client = upnp.createClient();
        if (!host) {
            this.host = this.getBestHostForUPnP();
        }
    }
    async checkUPnPisAvailable() {
        try {
            await new Promise((resolve, reject) => {
                this.client.externalIp((err, res) => {
                    if (err || !res) {
                        reject();
                    }
                    else {
                        resolve();
                    }
                });
            });
            return true;
        }
        catch (err) {
            return false;
        }
    }
    getCurrentConfig() {
        return this.currentConfig;
    }
    getUpnpDescription() {
        return 'duniter:' + this.identifier;
    }
    /**
     * Always open the same port during an execution of Duniter.
     * @returns { host:string, port:number }
     */
    openPort() {
        return new Promise(async (resolve, reject) => {
            if (!this.currentConfig) {
                this.currentConfig = await this.getAvailablePort(this.client);
            }
            this.logger && this.logger.trace('WS2P: mapping external port %s to local %s using UPnP...', this.currentConfig.port, [this.currentConfig.host, this.currentConfig.port].join(':'));
            const client = upnp.createClient();
            client.portMapping({
                'public': this.currentConfig.port,
                'private': {
                    host: this.currentConfig.host,
                    port: this.currentConfig.port,
                },
                'ttl': this.ttl,
                'description': this.getUpnpDescription()
            }, (err) => {
                client.close();
                if (err) {
                    this.logger && this.logger.warn(err);
                    return reject(err);
                }
                resolve(this.currentConfig);
            });
        });
    }
    async startRegular() {
        this.stopRegular();
        const available = await this.checkUPnPisAvailable();
        if (available) {
            // Update UPnP IGD every INTERVAL seconds
            this.interval = setInterval(() => this.openPort(), 1000 * this.upnpInterval);
            const { host, port } = await this.openPort();
            return { host, port, available };
        }
        return { host: '', port: 0, available: false };
    }
    stopRegular() {
        if (this.interval) {
            clearInterval(this.interval);
        }
    }
    static async getLocalIP(client) {
        return await new Promise((resolve, reject) => {
            client.findGateway((err, res, localIP) => {
                if (err)
                    return reject(err);
                resolve(localIP);
            });
        });
    }
    static async getRemoteIP(client) {
        return await new Promise((resolve, reject) => {
            client.externalIp((err, externalIP) => {
                if (err)
                    return reject(err);
                resolve(externalIP);
            });
        });
    }
    async getAvailablePort(client) {
        const localIP = this.host || await UpnpProvider.getLocalIP(client);
        const remoteIP = await UpnpProvider.getRemoteIP(client);
        const mappings = await UpnpProvider.getUPnPMappings(client);
        const thisDesc = this.getUpnpDescription();
        const externalPortsUsed = mappings.filter((m) => m.description !== thisDesc).map((m) => m.public.port);
        let availablePort = this.portStart;
        while (externalPortsUsed.indexOf(availablePort) !== -1
            && availablePort <= this.portEnd) {
            availablePort++;
        }
        if (availablePort > this.portEnd) {
            throw "No port available for UPnP";
        }
        return {
            remotehost: remoteIP,
            host: localIP,
            port: availablePort
        };
    }
    static async getUPnPMappings(client) {
        return new Promise((resolve, reject) => {
            client.getMappings((err, res) => {
                if (err) {
                    reject(err);
                }
                else {
                    resolve(res);
                }
            });
        });
    }
    /**
     * Apparently on computers with both Wi-Fi and Ethernet, UPnP fails to open the port.
     * Giving the priority to Ethernet seems to fix the problem.
     * @param family
     */
    getBestHostForUPnP(family = '') {
        let netInterfaces = os.networkInterfaces();
        let keys = underscore_1.Underscore.keys(netInterfaces);
        let res = [];
        for (const name of keys) {
            let addresses = netInterfaces[name];
            for (const addr of addresses) {
                if (!family || addr.family == family) {
                    res.push({
                        name: name,
                        value: addr.address
                    });
                }
            }
        }
        const interfacePriorityRegCatcher = [
            /^tun\d+/,
            /^enp\d+s\d+/,
            /^enp\d+s\d+f\d+/,
            /^eth\d+/,
            /^Ethernet/,
            /^wlp\d+s\d+/,
            /^wlan\d+/,
            /^Wi-Fi/,
            /^lo/,
            /^Loopback/,
            /^None/
        ];
        const best = underscore_1.Underscore.sortBy(res, function (entry) {
            for (let i = 0; i < interfacePriorityRegCatcher.length; i++) {
                // `i` is the priority (0 is the better, 1 is the second, ...)
                if (entry.name.match(interfacePriorityRegCatcher[i]))
                    return i;
            }
            return interfacePriorityRegCatcher.length;
        })[0];
        return (best && best.value) || "";
    }
}
exports.UpnpProvider = UpnpProvider;
//# sourceMappingURL=upnp-provider.js.map