"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const keyring_1 = require("../../../lib/common-libs/crypto/keyring");
const constants_1 = require("./constants");
const websocket_1 = require("../../../lib/common-libs/websocket");
const manual_promise_1 = require("../../../lib/common-libs/manual-promise");
const errors_1 = require("../../../lib/common-libs/errors");
const SocksProxyAgent = require('socks-proxy-agent');
const nuuid = require('node-uuid');
const logger = require('../../../lib/logger').NewLogger('ws2p');
const MAXIMUM_ERRORS_COUNT = 5;
var WS2P_ERR;
(function (WS2P_ERR) {
    WS2P_ERR[WS2P_ERR["REJECTED_PUBKEY_OR_INCORRECT_ASK_SIGNATURE_FROM_REMOTE"] = 0] = "REJECTED_PUBKEY_OR_INCORRECT_ASK_SIGNATURE_FROM_REMOTE";
    WS2P_ERR[WS2P_ERR["AUTH_INVALID_ASK_FIELDS"] = 1] = "AUTH_INVALID_ASK_FIELDS";
    WS2P_ERR[WS2P_ERR["AUTH_INVALID_ACK_FIELDS"] = 2] = "AUTH_INVALID_ACK_FIELDS";
    WS2P_ERR[WS2P_ERR["AUTH_INVALID_OK_FIELDS"] = 3] = "AUTH_INVALID_OK_FIELDS";
    WS2P_ERR[WS2P_ERR["INCORRECT_ACK_SIGNATURE_FROM_REMOTE"] = 4] = "INCORRECT_ACK_SIGNATURE_FROM_REMOTE";
    WS2P_ERR[WS2P_ERR["INCORRECT_ASK_SIGNATURE_FROM_REMOTE"] = 5] = "INCORRECT_ASK_SIGNATURE_FROM_REMOTE";
    WS2P_ERR[WS2P_ERR["UNKNOWN_AUTH_MESSAGE"] = 6] = "UNKNOWN_AUTH_MESSAGE";
    WS2P_ERR[WS2P_ERR["ALREADY_AUTHENTICATED_AND_CONFIRMED_BY_REMOTE"] = 7] = "ALREADY_AUTHENTICATED_AND_CONFIRMED_BY_REMOTE";
    WS2P_ERR[WS2P_ERR["ALREADY_AUTHENTICATED_REMOTE"] = 8] = "ALREADY_AUTHENTICATED_REMOTE";
    WS2P_ERR[WS2P_ERR["ALREADY_AUTHENTICATED_BY_REMOTE"] = 9] = "ALREADY_AUTHENTICATED_BY_REMOTE";
    WS2P_ERR[WS2P_ERR["INCORRECT_PUBKEY_FOR_REMOTE"] = 10] = "INCORRECT_PUBKEY_FOR_REMOTE";
    WS2P_ERR[WS2P_ERR["MUST_BE_AUTHENTICATED_FIRST"] = 11] = "MUST_BE_AUTHENTICATED_FIRST";
    WS2P_ERR[WS2P_ERR["REQUEST_FAILED"] = 12] = "REQUEST_FAILED";
    WS2P_ERR[WS2P_ERR["MESSAGE_MUST_BE_AN_OBJECT"] = 13] = "MESSAGE_MUST_BE_AN_OBJECT";
    WS2P_ERR[WS2P_ERR["ANSWER_TO_UNDEFINED_REQUEST"] = 14] = "ANSWER_TO_UNDEFINED_REQUEST";
})(WS2P_ERR || (WS2P_ERR = {}));
var WS2P_PUSH;
(function (WS2P_PUSH) {
    WS2P_PUSH[WS2P_PUSH["PEER"] = 0] = "PEER";
    WS2P_PUSH[WS2P_PUSH["TRANSACTION"] = 1] = "TRANSACTION";
    WS2P_PUSH[WS2P_PUSH["MEMBERSHIP"] = 2] = "MEMBERSHIP";
    WS2P_PUSH[WS2P_PUSH["CERTIFICATION"] = 3] = "CERTIFICATION";
    WS2P_PUSH[WS2P_PUSH["IDENTITY"] = 4] = "IDENTITY";
    WS2P_PUSH[WS2P_PUSH["BLOCK"] = 5] = "BLOCK";
    WS2P_PUSH[WS2P_PUSH["HEAD"] = 6] = "HEAD";
})(WS2P_PUSH = exports.WS2P_PUSH || (exports.WS2P_PUSH = {}));
/**
 * A passive authenticator based on our keyring.
 */
class WS2PPubkeyRemoteAuth {
    constructor(currency, pair, tellIsAuthorizedPubkey = () => Promise.resolve(true)) {
        this.currency = currency;
        this.pair = pair;
        this.tellIsAuthorizedPubkey = tellIsAuthorizedPubkey;
        this.authenticatedByRemote = false;
        this.remotePub = "";
        this.remoteWs2pId = "";
        this.remoteVersion = 1;
        this.isSyncConnection = false;
        this.challenge = nuuid.v4() + nuuid.v4();
        this.serverAuth = new Promise((resolve, reject) => {
            this.serverAuthResolve = resolve;
            this.serverAuthReject = reject;
        });
        this.givenCurrency = manual_promise_1.newManualPromise();
        // If the currency is already provided, resolve the promise immediately
        if (currency) {
            this.givenCurrency.resolve(currency);
        }
    }
    getVersion() {
        return this.remoteVersion;
    }
    getPubkey() {
        return this.remotePub;
    }
    isSync() {
        return this.isSyncConnection;
    }
    async sendACK(ws) {
        const challengeMessage = `WS2P:ACK:${this.currency}:${this.pair.pub}:${this.challenge}`;
        Logger.log('sendACK >>> ' + challengeMessage);
        const sig = this.pair.signSync(challengeMessage);
        await ws.send(JSON.stringify({
            auth: 'ACK',
            pub: this.pair.pub,
            sig
        }));
    }
    async registerCONNECT(type, ws2pVersion, challenge, sig, pub, currency, ws2pId = "") {
        this.isSyncConnection = type === 'SYNC';
        const allow = await this.tellIsAuthorizedPubkey(pub, this.isSyncConnection);
        if (!allow) {
            return false;
        }
        // If the connection was not aware of the currency beforehand, let's give the remote node's value
        if (!this.currency && currency) {
            this.currency = currency;
        }
        else if (this.currency && this.currency !== currency && currency) {
            throw Error(errors_1.DataErrors[errors_1.DataErrors.WRONG_CURRENCY_DETECTED]);
        }
        this.givenCurrency.resolve(this.currency);
        const challengeMessage = (ws2pVersion > 1) ? `WS2P:${type}:${this.currency}:${pub}:${ws2pId}:${challenge}` : `WS2P:${type}:${this.currency}:${pub}:${challenge}`;
        Logger.log('registerCONNECT >>> ' + challengeMessage);
        const verified = keyring_1.verify(challengeMessage, sig, pub);
        if (verified) {
            this.remoteVersion = ws2pVersion;
            this.challenge = challenge;
            this.remotePub = pub;
            this.remoteWs2pId = ws2pId;
        }
        return verified;
    }
    async registerOK(sig) {
        const challengeMessage = `WS2P:OK:${this.currency}:${this.remotePub}:${this.challenge}`;
        Logger.log('registerOK >>> ' + challengeMessage);
        this.authenticatedByRemote = keyring_1.verify(challengeMessage, sig, this.remotePub);
        if (!this.authenticatedByRemote) {
            this.serverAuthReject("Wrong signature from remote OK");
        }
        else {
            this.serverAuthResolve();
        }
        return this.authenticatedByRemote;
    }
    isAuthenticatedByRemote() {
        return this.authenticatedByRemote;
    }
    authenticationIsDone() {
        return this.serverAuth;
    }
}
exports.WS2PPubkeyRemoteAuth = WS2PPubkeyRemoteAuth;
/**
 * An connecting authenticator based on our keyring.
 */
class WS2PPubkeyLocalAuth {
    constructor(currency, pair, ws2pId, tellIsAuthorizedPubkey = () => Promise.resolve(true)) {
        this.currency = currency;
        this.pair = pair;
        this.ws2pId = ws2pId;
        this.tellIsAuthorizedPubkey = tellIsAuthorizedPubkey;
        this.authenticated = false;
        this.challenge = nuuid.v4() + nuuid.v4();
        this.serverAuth = new Promise((resolve, reject) => {
            this.serverAuthResolve = resolve;
            this.serverAuthReject = reject;
        });
        this.isSync = false;
    }
    async sendCONNECT(ws, ws2pVersion) {
        const connectWord = this.isSync ? 'SYNC' : 'CONNECT';
        if (ws2pVersion > 1) {
            const challengeMessage = `WS2P:${ws2pVersion}:${connectWord}:${this.currency}:${this.pair.pub}:${this.ws2pId}:${this.challenge}`;
            Logger.log('sendCONNECT >>> ' + challengeMessage);
            const sig = this.pair.signSync(challengeMessage);
            await ws.send(JSON.stringify({
                auth: `${connectWord}`,
                version: ws2pVersion,
                pub: this.pair.pub,
                ws2pid: this.ws2pId,
                challenge: this.challenge,
                sig,
                currency: this.currency,
            }));
            return this.serverAuth;
        }
        else if (ws2pVersion == 1) {
            const challengeMessage = `WS2P:${connectWord}:${this.currency}:${this.pair.pub}:${this.challenge}`;
            Logger.log('sendCONNECT >>> ' + challengeMessage);
            const sig = this.pair.signSync(challengeMessage);
            await ws.send(JSON.stringify({
                auth: `${connectWord}`,
                pub: this.pair.pub,
                challenge: this.challenge,
                sig,
                currency: this.currency,
            }));
            return this.serverAuth;
        }
    }
    async registerACK(sig, pub) {
        const allow = await this.tellIsAuthorizedPubkey(pub);
        if (!allow) {
            return false;
        }
        const challengeMessage = `WS2P:ACK:${this.currency}:${pub}:${this.challenge}`;
        Logger.log('registerACK >>> ' + challengeMessage);
        this.authenticated = keyring_1.verify(challengeMessage, sig, pub);
        if (!this.authenticated) {
            this.serverAuthReject("Wrong signature from server ACK");
        }
        else {
            this.serverAuthResolve();
        }
        return this.authenticated;
    }
    async sendOK(ws) {
        const challengeMessage = `WS2P:OK:${this.currency}:${this.pair.pub}:${this.challenge}`;
        Logger.log('sendOK >>> ' + challengeMessage);
        const sig = this.pair.signSync(challengeMessage);
        await ws.send(JSON.stringify({
            auth: 'OK',
            sig
        }));
        return this.serverAuth;
    }
    isRemoteAuthenticated() {
        return this.authenticated;
    }
    authenticationIsDone() {
        return this.serverAuth;
    }
}
exports.WS2PPubkeyLocalAuth = WS2PPubkeyLocalAuth;
class WS2PPubkeySyncLocalAuth extends WS2PPubkeyLocalAuth {
    constructor(currency, // Only here for function signature purpose
    pair, ws2pId, tellIsAuthorizedPubkey = () => Promise.resolve(true)) {
        super("", pair, ws2pId, tellIsAuthorizedPubkey);
        this.pair = pair;
        this.ws2pId = ws2pId;
        this.tellIsAuthorizedPubkey = tellIsAuthorizedPubkey;
        this.isSync = true;
    }
}
exports.WS2PPubkeySyncLocalAuth = WS2PPubkeySyncLocalAuth;
class WS2PMessageExchange {
    constructor(extras, race) {
        this.promise = race;
        this.extras = extras;
    }
}
exports.WS2PMessageExchange = WS2PMessageExchange;
/**
 * The handler of a WS2P connection.
 *
 * Goal: operating an authenticated bidirectionnal communication over a WebSocket connection.
 * Requires an established WebSocket connection in order to work.
 */
class WS2PConnection {
    constructor(ws2pVersion, ws, onWsOpened, onWsClosed, messageHandler, localAuth, remoteAuth, options = {
        connectionTimeout: constants_1.WS2PConstants.REQUEST_TIMEOUT,
        requestTimeout: constants_1.WS2PConstants.REQUEST_TIMEOUT
    }, expectedPub = "", expectedWS2PUID = "") {
        this.ws2pVersion = ws2pVersion;
        this.ws = ws;
        this.onWsOpened = onWsOpened;
        this.onWsClosed = onWsClosed;
        this.messageHandler = messageHandler;
        this.localAuth = localAuth;
        this.remoteAuth = remoteAuth;
        this.options = options;
        this.expectedPub = expectedPub;
        this.expectedWS2PUID = expectedWS2PUID;
        this.nbErrors = 0;
        this.nbRequestsCount = 0;
        this.nbResponsesCount = 0;
        this.nbPushsToRemoteCount = 0;
        this.nbPushsByRemoteCount = 0;
        this.exchanges = {};
        this.connectedp = new Promise((resolve, reject) => {
            this.connectedResolve = resolve;
            this.connectedReject = reject;
        });
    }
    static newConnectionToAddress(ws2pVersion, address, messageHandler, localAuth, remoteAuth, proxySocksAddress = undefined, options = {
        connectionTimeout: constants_1.WS2PConstants.REQUEST_TIMEOUT,
        requestTimeout: constants_1.WS2PConstants.REQUEST_TIMEOUT
    }, expectedPub = "", expectedWS2PUID = "") {
        if (address.match(constants_1.WS2PConstants.FULL_ADDRESS_ONION_REGEX)) {
            options = {
                connectionTimeout: constants_1.WS2PConstants.CONNEXION_TOR_TIMEOUT,
                requestTimeout: constants_1.WS2PConstants.REQUEST_TOR_TIMEOUT
            };
        }
        const websocket = (proxySocksAddress !== undefined) ? new websocket_1.WebSocket(address, { agent: SocksProxyAgent("socks://" + proxySocksAddress) }) : new websocket_1.WebSocket(address);
        const onWsOpened = new Promise(res => {
            websocket.on('open', () => res());
        });
        const onWsClosed = new Promise(res => {
            websocket.on('close', () => res());
        });
        websocket.on('error', () => websocket.close());
        return new WS2PConnection(ws2pVersion, websocket, onWsOpened, onWsClosed, messageHandler, localAuth, remoteAuth, options, expectedPub, expectedWS2PUID);
    }
    static newConnectionFromWebSocketServer(websocket, messageHandler, localAuth, remoteAuth, options = {
        connectionTimeout: constants_1.WS2PConstants.REQUEST_TIMEOUT,
        requestTimeout: constants_1.WS2PConstants.REQUEST_TIMEOUT
    }, expectedPub = "") {
        const onWsOpened = Promise.resolve();
        const onWsClosed = new Promise(res => {
            websocket.on('close', () => res());
        });
        return new WS2PConnection(constants_1.WS2PConstants.WS2P_DEFAULT_API_VERSION, websocket, onWsOpened, onWsClosed, messageHandler, localAuth, remoteAuth, options, expectedPub);
    }
    get version() {
        return Math.min(constants_1.WS2PConstants.WS2P_HEAD_VERSION, this.remoteAuth.getVersion());
    }
    get pubkey() {
        return this.remoteAuth.getPubkey();
    }
    get uuid() {
        return this.expectedWS2PUID;
    }
    get isSync() {
        return this.remoteAuth.isSync();
    }
    get nbRequests() {
        return this.nbRequestsCount;
    }
    get nbResponses() {
        return this.nbResponsesCount;
    }
    get nbPushsToRemote() {
        return this.nbPushsToRemoteCount;
    }
    get nbPushsByRemote() {
        return this.nbPushsByRemoteCount;
    }
    get connected() {
        return this.connectedp;
    }
    get closed() {
        return this.onWsClosed;
    }
    close() {
        return this.ws.close();
    }
    async connectAsInitiator() {
        return this.connect(true);
    }
    async connect(initiator = false) {
        const whoIs = initiator ? 'INITIATOR' : 'SERVER';
        if (!this.connectp) {
            this.connectp = (async () => {
                const connectionTimeout = new Promise((res, rej) => {
                    setTimeout(() => {
                        rej("WS2P connection timeout");
                    }, this.options.connectionTimeout);
                });
                try {
                    await Promise.race([connectionTimeout, new Promise((resolve, reject) => {
                            (async () => {
                                await this.onWsOpened;
                                try {
                                    // First: wait to know about the currency name. It can be given spontaneously by the remote node,
                                    // or be already given if we know it before any exchange.
                                    const currency = await this.remoteAuth.givenCurrency;
                                    // Then make some checks about its value
                                    if (this.localAuth.currency && this.localAuth.currency !== currency) {
                                        throw Error(errors_1.DataErrors[errors_1.DataErrors.WRONG_CURRENCY_DETECTED]);
                                    }
                                    // Eventually give the information to localAuth, which will now be able to start the connection (currency is required for WS2P messages)
                                    if (!this.localAuth.currency) {
                                        this.localAuth.currency = currency;
                                    }
                                    await this.localAuth.sendCONNECT(this.ws, this.ws2pVersion);
                                    await Promise.all([
                                        this.localAuth.authenticationIsDone(),
                                        this.remoteAuth.authenticationIsDone()
                                    ]);
                                    resolve();
                                }
                                catch (e) {
                                    reject(e);
                                }
                            })();
                            this.ws.on('message', async (msg) => {
                                const data = JSON.parse(msg);
                                // Incorrect data
                                if (typeof data !== 'object') {
                                    // We only accept JSON objects
                                    await this.errorDetected(WS2P_ERR.MESSAGE_MUST_BE_AN_OBJECT);
                                }
                                // OK: JSON object
                                else {
                                    /************************
                                     * CONNECTION STUFF
                                     ************************/
                                    if (data.auth && typeof data.auth === "string") {
                                        if (data.auth === "CONNECT" || data.auth === "SYNC") {
                                            if (data.version) {
                                                if (typeof data.version !== "number") {
                                                    await this.errorDetected(WS2P_ERR.AUTH_INVALID_ASK_FIELDS);
                                                }
                                                else {
                                                    this.ws2pVersion = data.version;
                                                }
                                            }
                                            if (this.remoteAuth.isAuthenticatedByRemote()) {
                                                return this.errorDetected(WS2P_ERR.ALREADY_AUTHENTICATED_BY_REMOTE);
                                            }
                                            else if (typeof data.pub !== "string" || typeof data.sig !== "string" || typeof data.challenge !== "string" || (this.ws2pVersion > 1 && typeof data.ws2pId !== "string")) {
                                                await this.errorDetected(WS2P_ERR.AUTH_INVALID_ASK_FIELDS);
                                            }
                                            else {
                                                if (this.expectedPub && data.pub !== this.expectedPub) {
                                                    await this.errorDetected(WS2P_ERR.INCORRECT_PUBKEY_FOR_REMOTE);
                                                }
                                                else {
                                                    const valid = await this.remoteAuth.registerCONNECT(data.auth, this.ws2pVersion, data.challenge, data.sig, data.pub, data.currency, (this.ws2pVersion > 1) ? data.ws2pID : "");
                                                    if (valid) {
                                                        await this.remoteAuth.sendACK(this.ws);
                                                    }
                                                    else {
                                                        await this.errorDetected(WS2P_ERR.REJECTED_PUBKEY_OR_INCORRECT_ASK_SIGNATURE_FROM_REMOTE);
                                                    }
                                                }
                                            }
                                        }
                                        else if (data.auth === "ACK") {
                                            if (this.localAuth.isRemoteAuthenticated()) {
                                                return this.errorDetected(WS2P_ERR.ALREADY_AUTHENTICATED_REMOTE);
                                            }
                                            if (typeof data.pub !== "string" || typeof data.sig !== "string") {
                                                await this.errorDetected(WS2P_ERR.AUTH_INVALID_ACK_FIELDS);
                                            }
                                            else {
                                                if (this.expectedPub && data.pub !== this.expectedPub) {
                                                    await this.errorDetected(WS2P_ERR.INCORRECT_PUBKEY_FOR_REMOTE);
                                                }
                                                else {
                                                    try {
                                                        const valid = await this.localAuth.registerACK(data.sig, data.pub);
                                                        if (valid) {
                                                            await this.localAuth.sendOK(this.ws);
                                                        }
                                                    }
                                                    catch (e) {
                                                        await this.errorDetected(WS2P_ERR.INCORRECT_ACK_SIGNATURE_FROM_REMOTE);
                                                    }
                                                }
                                            }
                                        }
                                        else if (data.auth === "OK") {
                                            if (this.remoteAuth.isAuthenticatedByRemote()) {
                                                return this.errorDetected(WS2P_ERR.ALREADY_AUTHENTICATED_AND_CONFIRMED_BY_REMOTE);
                                            }
                                            if (typeof data.sig !== "string") {
                                                await this.errorDetected(WS2P_ERR.AUTH_INVALID_OK_FIELDS);
                                            }
                                            else {
                                                await this.remoteAuth.registerOK(data.sig);
                                            }
                                        }
                                        else {
                                            await this.errorDetected(WS2P_ERR.UNKNOWN_AUTH_MESSAGE);
                                        }
                                    }
                                    /************************
                                     * APPLICATION STUFF
                                     ************************/
                                    else {
                                        if (!this.localAuth.isRemoteAuthenticated()) {
                                            await this.errorDetected(WS2P_ERR.MUST_BE_AUTHENTICATED_FIRST);
                                        }
                                        // Request message
                                        else if (data.reqId && typeof data.reqId === "string") {
                                            try {
                                                const answer = await this.messageHandler.answerToRequest(data.body, this);
                                                this.ws.send(JSON.stringify({ resId: data.reqId, body: answer }));
                                            }
                                            catch (e) {
                                                this.ws.send(JSON.stringify({ resId: data.reqId, err: e }));
                                            }
                                        }
                                        // Answer message
                                        else if (data.resId && typeof data.resId === "string") {
                                            // An answer
                                            const request = this.exchanges[data.resId];
                                            this.nbResponsesCount++;
                                            if (request !== undefined) {
                                                request.extras.resolve(data.body);
                                            }
                                            else {
                                                await this.errorDetected(WS2P_ERR.ANSWER_TO_UNDEFINED_REQUEST);
                                            }
                                        }
                                        // Push message
                                        else {
                                            this.nbPushsByRemoteCount++;
                                            await this.messageHandler.handlePushMessage(data, this);
                                        }
                                    }
                                }
                            });
                        })]);
                    this.connectedResolve(this.remoteAuth.getPubkey());
                }
                catch (e) {
                    this.connectedReject(e);
                    await this.connectedp;
                }
            })();
        }
        return this.connectp;
    }
    async request(body) {
        await this.connectAsInitiator();
        const uuid = nuuid.v4();
        return new Promise((resolve, reject) => {
            this.nbRequestsCount++;
            this.ws.send(JSON.stringify({
                reqId: uuid,
                body
            }), async (err) => {
                if (err) {
                    return reject(err);
                }
                else {
                    // The request was successfully sent. Now wait for the answer.
                    const extras = {
                        resolve: () => { console.error('resolution not implemented'); },
                        reject: () => { console.error('rejection not implemented'); }
                    };
                    this.exchanges[uuid] = new WS2PMessageExchange(extras, Promise.race([
                        // The answer
                        new Promise((res, rej) => {
                            extras.resolve = res;
                            extras.reject = () => {
                                this.errorDetected(WS2P_ERR.REQUEST_FAILED);
                                rej();
                            };
                        }),
                        // Timeout
                        new Promise((res, rej) => {
                            setTimeout(() => {
                                rej("WS2P request timeout");
                            }, this.options.requestTimeout);
                        })
                    ]));
                    try {
                        resolve(await this.exchanges[uuid].promise);
                    }
                    catch (e) {
                        reject(e);
                    }
                    finally {
                        delete this.exchanges[uuid];
                    }
                }
            });
        });
    }
    async pushBlock(block) {
        return this.pushData(WS2P_PUSH.BLOCK, 'block', block);
    }
    async pushIdentity(idty) {
        return this.pushData(WS2P_PUSH.IDENTITY, 'identity', idty);
    }
    async pushCertification(cert) {
        return this.pushData(WS2P_PUSH.CERTIFICATION, 'certification', cert);
    }
    async pushMembership(ms) {
        return this.pushData(WS2P_PUSH.MEMBERSHIP, 'membership', ms);
    }
    async pushTransaction(tx) {
        return this.pushData(WS2P_PUSH.TRANSACTION, 'transaction', tx);
    }
    async pushPeer(peer) {
        return this.pushData(WS2P_PUSH.PEER, 'peer', peer);
    }
    async pushHeads(heads) {
        return this.pushData(WS2P_PUSH.HEAD, 'heads', heads);
    }
    async pushData(type, key, data) {
        await this.connectAsInitiator();
        return new Promise((resolve, reject) => {
            this.nbPushsToRemoteCount++;
            try {
                this.ws.send(JSON.stringify({
                    body: {
                        name: WS2P_PUSH[type],
                        [key]: data
                    }
                }), async (err) => {
                    if (err) {
                        return reject(err);
                    }
                    resolve();
                });
            }
            catch (e) {
                reject(e);
            }
        });
    }
    errorDetected(cause) {
        this.nbErrors++;
        Logger.error('>>> WS ERROR: %s', WS2P_ERR[cause]);
        if (this.nbErrors >= MAXIMUM_ERRORS_COUNT) {
            this.ws.terminate();
        }
    }
    get hostName() {
        return this.ws.url
            .replace('ws://', '')
            .replace('wss://', '');
    }
}
exports.WS2PConnection = WS2PConnection;
class Logger {
    static log(message) {
        logger.trace('WS2P >>> ' + message);
    }
    static error(message, obj) {
        logger.error('WS2P >>> ' + message, obj);
    }
}
//# sourceMappingURL=WS2PConnection.js.map