"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const WS2PReqMapperByServer_1 = require("../impl/WS2PReqMapperByServer");
const BlockDTO_1 = require("../../../../lib/dto/BlockDTO");
const IdentityDTO_1 = require("../../../../lib/dto/IdentityDTO");
const CertificationDTO_1 = require("../../../../lib/dto/CertificationDTO");
const MembershipDTO_1 = require("../../../../lib/dto/MembershipDTO");
const TransactionDTO_1 = require("../../../../lib/dto/TransactionDTO");
const PeerDTO_1 = require("../../../../lib/dto/PeerDTO");
const WS2PRequester_1 = require("../WS2PRequester");
const constants_1 = require("../constants");
const constants_2 = require("../../../../lib/common-libs/constants");
const errors_1 = require("../../../../lib/common-libs/errors");
var WS2P_REQERROR;
(function (WS2P_REQERROR) {
    WS2P_REQERROR[WS2P_REQERROR["UNKNOWN_REQUEST"] = 0] = "UNKNOWN_REQUEST";
})(WS2P_REQERROR = exports.WS2P_REQERROR || (exports.WS2P_REQERROR = {}));
class WS2PServerMessageHandler {
    constructor(server, cluster) {
        this.server = server;
        this.cluster = cluster;
        this.errors = {};
        this.mapper = new WS2PReqMapperByServer_1.WS2PReqMapperByServer(server);
    }
    async handlePushMessage(json, c) {
        if (c.isSync) {
            // Push messages are forbidden on sync connection
            c.close();
            return;
        }
        let documentHash = '';
        try {
            if (json.body) {
                if (json.body.block) {
                    const dto = BlockDTO_1.BlockDTO.fromJSONObject(json.body.block);
                    const raw = dto.getRawSigned();
                    documentHash = dto.getHash();
                    await this.server.writeRawBlock(raw);
                }
                else if (json.body.identity) {
                    const dto = IdentityDTO_1.IdentityDTO.fromJSONObject(json.body.identity);
                    const raw = dto.getRawSigned();
                    documentHash = dto.getHash();
                    await this.server.writeRawIdentity(raw);
                }
                else if (json.body.certification) {
                    const dto = CertificationDTO_1.CertificationDTO.fromJSONObject(json.body.certification);
                    const raw = dto.getRawSigned();
                    documentHash = dto.getHash();
                    await this.server.writeRawCertification(raw);
                }
                else if (json.body.membership) {
                    const dto = MembershipDTO_1.MembershipDTO.fromJSONObject(json.body.membership);
                    const raw = dto.getRawSigned();
                    documentHash = dto.getHash();
                    await this.server.writeRawMembership(raw);
                }
                else if (json.body.transaction) {
                    const dto = TransactionDTO_1.TransactionDTO.fromJSONObject(json.body.transaction);
                    const raw = dto.getRaw();
                    documentHash = dto.getHash();
                    await this.server.writeRawTransaction(raw);
                }
                else if (json.body.peer) {
                    const dto = PeerDTO_1.PeerDTO.fromJSONObject(json.body.peer);
                    const raw = dto.getRawSigned();
                    documentHash = dto.getHash();
                    await this.server.writeRawPeer(raw);
                }
                else if (json.body.heads && typeof json.body.heads === "object" && json.body.heads.length !== undefined) {
                    if (!json.body.heads.length) {
                        documentHash = 'HEADs';
                        throw "Heads empty HEADs received";
                    }
                    await this.cluster.headsReceived(json.body.heads || []);
                }
            }
        }
        catch (e) {
            if (documentHash
                && this.errors[documentHash]
                && this.errors[documentHash].pubkeys[c.pubkey] !== undefined
                && this.server.conf.pair.pub !== c.pubkey) { // We do not want to ban ourselves
                this.errors[documentHash].pubkeys[c.pubkey].push(json.body);
                if (this.errors[documentHash].pubkeys[c.pubkey].length >= constants_1.WS2PConstants.BAN_ON_REPEAT_THRESHOLD) {
                    let message = "peer " + (c.pubkey || '--unknown--') + " sent " + constants_1.WS2PConstants.BAN_ON_REPEAT_THRESHOLD + " times a same wrong document: " + (e && (e.message || (e.uerr && e.uerr.message)) || JSON.stringify(e));
                    this.cluster.banConnection(c, message);
                    for (const body of this.errors[documentHash].pubkeys[c.pubkey]) {
                        message += '\n => ' + JSON.stringify(body);
                    }
                }
                else {
                    let message = "WS2P IN => " + (c.pubkey || '--unknown--') + " sent " + this.errors[documentHash].pubkeys[c.pubkey].length + " times a same wrong document: " + (e && (e.message || (e.uerr && e.uerr.message)) || JSON.stringify(e));
                    for (const body of this.errors[documentHash].pubkeys[c.pubkey]) {
                        message += '\n => ' + JSON.stringify(body);
                    }
                    this.server.logger.warn(message);
                }
                setTimeout(() => {
                    if (this.errors[documentHash]) {
                        delete this.errors[documentHash];
                    }
                }, 1000 * constants_1.WS2PConstants.ERROR_RECALL_DURATION_IN_SECONDS);
            }
            else {
                // Remember the error for some time
                if (!this.errors[documentHash]) {
                    this.errors[documentHash] = {
                        createdOn: Date.now(),
                        pubkeys: {}
                    };
                }
                this.errors[documentHash].pubkeys[c.pubkey] = [json.body];
                setTimeout(() => {
                    if (this.errors[documentHash]) {
                        delete this.errors[documentHash];
                    }
                }, 1000 * constants_1.WS2PConstants.ERROR_RECALL_DURATION_IN_SECONDS);
            }
            if (e !== "Block already known"
                && (!e.uerr
                    || !(e.uerr.ucode == constants_2.CommonConstants.ERRORS.DOCUMENT_BEING_TREATED.uerr.ucode
                        || e.uerr.ucode == constants_2.CommonConstants.ERRORS.PEER_DOCUMENT_ALREADY_KNOWN.uerr.ucode))) {
                this.server.logger.warn(e);
            }
        }
    }
    async answerToRequest(data, c) {
        /**********
         * REQUEST
         *********/
        let body = {};
        const forbiddenRequestsForSync = []; // For now, no WS2P requests are forbidden
        if (c.isSync && (!data || !data.name || forbiddenRequestsForSync.indexOf(data.name) !== -1)) {
            // Some messages are forbidden on sync connection
            c.close();
            throw Error(errors_1.DataErrors[errors_1.DataErrors.WS2P_SYNC_PERIMETER_IS_LIMITED]);
        }
        if (data && data.name) {
            switch (data.name) {
                case WS2PRequester_1.WS2P_REQ[WS2PRequester_1.WS2P_REQ.CURRENT]:
                    body = await this.mapper.getCurrent();
                    break;
                case WS2PRequester_1.WS2P_REQ[WS2PRequester_1.WS2P_REQ.BLOCK_BY_NUMBER]:
                    if (isNaN(data.params.number)) {
                        throw "Wrong param `number`";
                    }
                    const number = data.params.number;
                    body = await this.mapper.getBlock(number);
                    break;
                case WS2PRequester_1.WS2P_REQ[WS2PRequester_1.WS2P_REQ.BLOCKS_CHUNK]:
                    if (isNaN(data.params.count)) {
                        throw "Wrong param `count`";
                    }
                    if (isNaN(data.params.fromNumber)) {
                        throw "Wrong param `fromNumber`";
                    }
                    const count = data.params.count;
                    const fromNumber = data.params.fromNumber;
                    body = await this.mapper.getBlocks(count, fromNumber);
                    break;
                case WS2PRequester_1.WS2P_REQ[WS2PRequester_1.WS2P_REQ.WOT_REQUIREMENTS_OF_PENDING]:
                    if (isNaN(data.params.minCert)) {
                        throw "Wrong param `minCert`";
                    }
                    const minCert = data.params.minCert;
                    body = await this.mapper.getRequirementsOfPending(minCert);
                    break;
                case WS2PRequester_1.WS2P_REQ[WS2PRequester_1.WS2P_REQ.PEER_DOCUMENT]:
                    body = await this.mapper.getPeer();
                    break;
                case WS2PRequester_1.WS2P_REQ[WS2PRequester_1.WS2P_REQ.KNOWN_PEERS]:
                    body = await this.mapper.getKnownPeers();
                    break;
                case WS2PRequester_1.WS2P_REQ[WS2PRequester_1.WS2P_REQ.MILESTONES_PAGE]:
                    body = await this.mapper.getMilestonesPage();
                    break;
                case WS2PRequester_1.WS2P_REQ[WS2PRequester_1.WS2P_REQ.MILESTONES]:
                    if (isNaN(data.params.page)) {
                        throw "Wrong param `page`";
                    }
                    const page = data.params.page;
                    body = await this.mapper.getMilestones(page);
                    break;
                default:
                    throw Error(WS2P_REQERROR[WS2P_REQERROR.UNKNOWN_REQUEST]);
            }
        }
        return body;
    }
}
exports.WS2PServerMessageHandler = WS2PServerMessageHandler;
//# sourceMappingURL=WS2PServerMessageHandler.js.map