"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const global_rules_1 = require("../lib/rules/global_rules");
const BlockDTO_1 = require("../lib/dto/BlockDTO");
const RevocationDTO_1 = require("../lib/dto/RevocationDTO");
const IdentityDTO_1 = require("../lib/dto/IdentityDTO");
const CertificationDTO_1 = require("../lib/dto/CertificationDTO");
const keyring_1 = require("../lib/common-libs/crypto/keyring");
const FIFOService_1 = require("./FIFOService");
const errors_1 = require("../lib/common-libs/errors");
"use strict";
const constants = require('../lib/constants');
const BY_ABSORPTION = true;
class IdentityService extends FIFOService_1.FIFOService {
    constructor(fifoPromiseHandler) {
        super(fifoPromiseHandler);
    }
    setConfDAL(newConf, newDAL) {
        this.dal = newDAL;
        this.conf = newConf;
        this.logger = require('../lib/logger').NewLogger(this.dal.profile);
    }
    searchIdentities(search) {
        return this.dal.searchJustIdentities(search);
    }
    async findMember(search) {
        let idty = null;
        if (search.match(constants.PUBLIC_KEY)) {
            idty = await this.dal.getWrittenIdtyByPubkeyForHashing(search);
        }
        else {
            idty = await this.dal.getWrittenIdtyByUidForHashing(search);
        }
        if (!idty) {
            throw constants.ERRORS.NO_MEMBER_MATCHING_PUB_OR_UID;
        }
        let memberships = [];
        if (idty) {
            const mss = await this.dal.msDAL.getMembershipsOfIssuer(idty.pub);
            const mssFromMindex = await this.dal.mindexDAL.reducable(idty.pub);
            memberships = mss.map(m => {
                return {
                    blockstamp: [m.blockNumber, m.blockHash].join('-'),
                    membership: m.membership,
                    number: m.blockNumber,
                    fpr: m.blockHash,
                    written_number: m.written_number
                };
            });
            memberships = memberships.concat(mssFromMindex.map((ms) => {
                const sp = ms.created_on.split('-');
                return {
                    blockstamp: ms.created_on,
                    membership: ms.leaving ? 'OUT' : 'IN',
                    number: parseInt(sp[0]),
                    fpr: sp[1],
                    written_number: parseInt(ms.written_on)
                };
            }));
        }
        return {
            idty: {
                pubkey: idty.pub,
                uid: idty.uid,
                buid: idty.created_on
            },
            memberships
        };
    }
    getPendingFromPubkey(pubkey) {
        return this.dal.getNonWritten(pubkey);
    }
    submitIdentity(idty, byAbsorption = false) {
        const idtyObj = IdentityDTO_1.IdentityDTO.fromJSONObject(idty);
        const toSave = IdentityDTO_1.IdentityDTO.fromBasicIdentity(idty);
        // Force usage of local currency name, do not accept other currencies documents
        idtyObj.currency = this.conf.currency;
        const createIdentity = idtyObj.rawWithoutSig();
        const hash = idtyObj.getHash();
        return this.pushFIFO(hash, async () => {
            this.logger.info('⬇ IDTY %s %s', idty.pubkey, idty.uid);
            // Check signature's validity
            let verified = keyring_1.verify(createIdentity, idty.sig, idty.pubkey);
            if (!verified) {
                throw constants.ERRORS.SIGNATURE_DOES_NOT_MATCH;
            }
            let existing = await this.dal.getGlobalIdentityByHashForExistence(toSave.hash);
            if (existing) {
                throw constants.ERRORS.ALREADY_UP_TO_DATE;
            }
            else {
                // Create if not already written uid/pubkey
                let used = await global_rules_1.GLOBAL_RULES_HELPERS.checkExistsPubkey(idty.pubkey, this.dal);
                if (used) {
                    throw constants.ERRORS.PUBKEY_ALREADY_USED;
                }
                used = await global_rules_1.GLOBAL_RULES_HELPERS.checkExistsUserID(idty.uid, this.dal);
                if (used) {
                    throw constants.ERRORS.UID_ALREADY_USED;
                }
                const current = await this.dal.getCurrentBlockOrNull();
                if (idty.buid == '0-E3B0C44298FC1C149AFBF4C8996FB92427AE41E4649B934CA495991B7852B855' && current) {
                    throw constants.ERRORS.BLOCKSTAMP_DOES_NOT_MATCH_A_BLOCK;
                }
                else if (current) {
                    let basedBlock = await this.dal.getAbsoluteValidBlockInForkWindowByBlockstamp(idty.buid);
                    if (!basedBlock) {
                        throw constants.ERRORS.BLOCKSTAMP_DOES_NOT_MATCH_A_BLOCK;
                    }
                    toSave.expires_on = basedBlock.medianTime + this.conf.idtyWindow;
                }
                await global_rules_1.GLOBAL_RULES_FUNCTIONS.checkIdentitiesAreWritable({ identities: [idtyObj.inline()], version: (current && current.version) || constants.BLOCK_GENERATED_VERSION }, this.conf, this.dal);
                if (byAbsorption !== BY_ABSORPTION) {
                    if (!(await this.dal.idtyDAL.sandbox.acceptNewSandBoxEntry({
                        certsCount: 0,
                        issuers: [idty.pubkey],
                        ref_block: parseInt(idty.buid.split('-')[0])
                    }, this.conf.pair && this.conf.pair.pub))) {
                        throw constants.ERRORS.SANDBOX_FOR_IDENTITY_IS_FULL;
                    }
                }
                await this.dal.savePendingIdentity(toSave);
                this.logger.info('✔ IDTY %s %s', idty.pubkey, idty.uid);
                return toSave;
            }
        });
    }
    async submitCertification(obj) {
        const current = await this.dal.getCurrentBlockOrNull();
        // Prepare validator for certifications
        const potentialNext = BlockDTO_1.BlockDTO.fromJSONObject({ currency: this.conf.currency, identities: [], number: current ? current.number + 1 : 0 });
        // Force usage of local currency name, do not accept other currencies documents
        obj.currency = this.conf.currency || obj.currency;
        const cert = CertificationDTO_1.CertificationDTO.fromJSONObject(obj);
        const targetHash = cert.getTargetHash();
        let possiblyNullIdty = await this.dal.getGlobalIdentityByHashForHashingAndSig(targetHash);
        let idtyAbsorbed = false;
        const idty = possiblyNullIdty !== null ? possiblyNullIdty : await this.submitIdentity({
            pubkey: cert.idty_issuer,
            uid: cert.idty_uid,
            buid: cert.idty_buid,
            sig: cert.idty_sig
        }, BY_ABSORPTION);
        if (possiblyNullIdty === null) {
            idtyAbsorbed = true;
        }
        let anErr;
        const hash = cert.getHash();
        return this.pushFIFO(hash, async () => {
            this.logger.info('⬇ CERT %s block#%s -> %s', cert.from, cert.block_number, idty.uid);
            try {
                await global_rules_1.GLOBAL_RULES_HELPERS.checkCertificationIsValidInSandbox(cert, potentialNext, () => Promise.resolve(idty), this.conf, this.dal);
            }
            catch (e) {
                anErr = e;
            }
            if (!anErr) {
                try {
                    let basedBlock = await this.dal.getTristampOf(cert.block_number);
                    if (cert.block_number == 0 && !basedBlock) {
                        basedBlock = {
                            number: 0,
                            hash: 'E3B0C44298FC1C149AFBF4C8996FB92427AE41E4649B934CA495991B7852B855',
                            medianTime: 0
                        };
                    }
                    if (!basedBlock) {
                        throw Error(errors_1.DataErrors[errors_1.DataErrors.CERT_BASED_ON_UNKNOWN_BLOCK]);
                    }
                    const mCert = {
                        issuers: [cert.from],
                        from: cert.from,
                        sig: cert.sig,
                        block_number: cert.block_number,
                        block_hash: basedBlock.hash,
                        target: targetHash,
                        to: idty.pubkey,
                        expires_on: basedBlock.medianTime + this.conf.sigWindow,
                        linked: false,
                        written: false,
                        expired: false,
                        written_block: null,
                        written_hash: null,
                        block: cert.block_number
                    };
                    let existingCert = await this.dal.existsCert(mCert, current);
                    if (!existingCert) {
                        if (!(await this.dal.certDAL.getSandboxForKey(cert.from).acceptNewSandBoxEntry(mCert, this.conf.pair && this.conf.pair.pub))) {
                            throw constants.ERRORS.SANDBOX_FOR_CERT_IS_FULL;
                        }
                        await this.dal.registerNewCertification(mCert);
                        this.logger.info('✔ CERT %s', mCert.from);
                    }
                    else {
                        throw constants.ERRORS.ALREADY_UP_TO_DATE;
                    }
                }
                catch (e) {
                    anErr = e;
                }
            }
            if (anErr) {
                if (idtyAbsorbed) {
                    await this.dal.idtyDAL.deleteByHash(targetHash);
                }
                const err = anErr;
                const errMessage = (err.uerr && err.uerr.message) || err.message || err;
                this.logger.info('✘ CERT %s %s', cert.from, errMessage);
                throw anErr;
            }
            return cert;
        });
    }
    submitRevocation(obj) {
        // Force usage of local currency name, do not accept other currencies documents
        obj.currency = this.conf.currency || obj.currency;
        const revoc = RevocationDTO_1.RevocationDTO.fromJSONObject(obj);
        const raw = revoc.rawWithoutSig();
        const hash = revoc.getHash();
        return this.pushFIFO(hash, async () => {
            try {
                this.logger.info('⬇ REVOCATION %s %s', revoc.pubkey, revoc.idty_uid);
                let verified = keyring_1.verify(raw, revoc.revocation, revoc.pubkey);
                if (!verified) {
                    throw 'Wrong signature for revocation';
                }
                const existing = await this.dal.getGlobalIdentityByHashForRevocation(obj.hash);
                if (existing) {
                    // Modify
                    if (existing.revoked) {
                        throw 'Already revoked';
                    }
                    else if (existing.revocation_sig) {
                        throw 'Revocation already registered';
                    }
                    else {
                        await this.dal.setRevocating({
                            pubkey: existing.pub,
                            buid: existing.created_on,
                            sig: existing.sig,
                            uid: existing.uid,
                            expires_on: existing.expires_on,
                            member: existing.member,
                            wasMember: existing.wasMember,
                        }, revoc.revocation);
                        this.logger.info('✔ REVOCATION %s %s', revoc.pubkey, revoc.idty_uid);
                        return revoc;
                    }
                }
                else {
                    // Create identity given by the revocation
                    const idty = IdentityDTO_1.IdentityDTO.fromRevocation(revoc);
                    idty.revocation_sig = revoc.revocation;
                    if (!(await this.dal.idtyDAL.sandbox.acceptNewSandBoxEntry({
                        issuers: [idty.pubkey],
                        ref_block: parseInt(idty.buid.split('-')[0]),
                        certsCount: 0
                    }, this.conf.pair && this.conf.pair.pub))) {
                        throw constants.ERRORS.SANDBOX_FOR_IDENTITY_IS_FULL;
                    }
                    await this.dal.savePendingIdentity(idty);
                    this.logger.info('✔ REVOCATION %s %s', revoc.pubkey, revoc.idty_uid);
                    return revoc;
                }
            }
            catch (e) {
                this.logger.info('✘ REVOCATION %s %s', revoc.pubkey, revoc.idty_uid);
                throw e;
            }
        });
    }
}
exports.IdentityService = IdentityService;
//# sourceMappingURL=IdentityService.js.map