"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const test_framework_1 = require("../tools/test-framework");
const constants_1 = require("../../../app/lib/common-libs/constants");
describe('Block revert with transaction sources', () => test_framework_1.writeBasicTestWithConfAnd2Users({
    dt: 10,
    udTime0: constants_1.CommonConstants.BLOCK_TX_CHAINING_ACTIVATION_MT + 10,
    ud0: 1000,
    switchOnHeadAdvance: 0,
}, (test) => {
    const now = constants_1.CommonConstants.BLOCK_TX_CHAINING_ACTIVATION_MT;
    test('(b#1) should init with a Dividend at block#1', async (s1, cat, tac, toc) => {
        await cat.createIdentity();
        await tac.createIdentity();
        await cat.cert(tac);
        await tac.cert(cat);
        await cat.join();
        await tac.join();
        await s1.commit({ time: now + 11 });
        await s1.commit({ time: now + 11 });
        await assertBlock1(s1, cat, tac, toc);
    });
    test('(b#2) tac sends to both tac and toc', async (s1, cat, tac, toc) => {
        // Using transaction chaining to also test this case
        const current = await s1._server.dal.getCurrentBlockOrNull();
        const tx1 = await cat.prepareITX(1000, cat);
        const tx2 = await cat.prepareUTX(tx1, ['SIG(0)'], [
            { qty: 100, base: 0, lock: 'SIG(' + tac.pub + ')' },
            { qty: 200, base: 0, lock: 'SIG(' + toc.pub + ')' },
            { qty: 700, base: 0, lock: 'SIG(' + cat.pub + ')' },
        ], {
            comment: 'CHAINED TX to 2 recipients', blockstamp: [current.number, current.hash].join('-')
        });
        await cat.sendTX(tx1);
        await cat.sendTX(tx2);
        await s1.commit({ time: now + 11 });
        await assertBlock2(s1, cat, tac, toc);
    });
    test('(b#3) tac gives all to cat', async (s1, cat, tac, toc) => {
        await tac.sendMoney(1100, cat);
        await s1.commit({ time: now + 11 });
        await assertBlock3(s1, cat, tac, toc);
    });
    test('(b#4) toc spends some received money', async (s1, cat, tac, toc) => {
        // Using normal transaction
        await toc.sendMoney(100, cat);
        await s1.commit({ time: now + 11 });
        await assertBlock4(s1, cat, tac, toc);
    });
    test('revert b#3-4 and re-commit block#3 should be ok', async (s1, cat, tac, toc) => {
        await s1.revert();
        await s1.revert();
        await s1.resolve(b => b.number === 3);
        await assertBlock3(s1, cat, tac, toc);
    });
    test('re-commit block#4 should be ok', async (s1, cat, tac, toc) => {
        await s1.resolve(b => b.number === 4);
        await assertBlock4(s1, cat, tac, toc);
    });
}));
async function assertBlock1(s1, cat, tac, toc) {
    test_framework_1.assertEqual((await s1._server.dal.dividendDAL.getUDSources(cat.pub)).length, 1);
    test_framework_1.assertEqual((await s1._server.dal.dividendDAL.getUDSources(tac.pub)).length, 1);
    test_framework_1.assertEqual((await s1._server.dal.dividendDAL.getUDSources(toc.pub)).length, 0); // toc is not a member
    test_framework_1.assertEqual((await s1._server.dal.sindexDAL.getAvailableForPubkey(cat.pub)).length, 0);
    test_framework_1.assertEqual((await s1._server.dal.sindexDAL.getAvailableForPubkey(tac.pub)).length, 0);
    test_framework_1.assertEqual((await s1._server.dal.sindexDAL.getAvailableForPubkey(toc.pub)).length, 0);
    test_framework_1.assertEqual((await s1._server.dal.walletDAL.getWallet('SIG(' + cat.pub + ')')).balance, 1000);
    test_framework_1.assertEqual((await s1._server.dal.walletDAL.getWallet('SIG(' + tac.pub + ')')).balance, 1000);
    test_framework_1.assertNull(await s1._server.dal.walletDAL.getWallet('SIG(' + toc.pub + ')'));
}
async function assertBlock2(s1, cat, tac, toc) {
    test_framework_1.assertEqual((await s1._server.dal.dividendDAL.getUDSources(cat.pub)).length, 0); // <-- The UD gets consumed
    test_framework_1.assertEqual((await s1._server.dal.dividendDAL.getUDSources(tac.pub)).length, 1);
    test_framework_1.assertEqual((await s1._server.dal.dividendDAL.getUDSources(toc.pub)).length, 0); // toc is not a member
    test_framework_1.assertEqual((await s1._server.dal.sindexDAL.getAvailableForPubkey(cat.pub)).length, 1);
    test_framework_1.assertEqual((await s1._server.dal.sindexDAL.getAvailableForPubkey(tac.pub)).length, 1);
    test_framework_1.assertEqual((await s1._server.dal.sindexDAL.getAvailableForPubkey(toc.pub)).length, 1);
    test_framework_1.assertEqual((await s1._server.dal.walletDAL.getWallet('SIG(' + cat.pub + ')')).balance, 700); // <-- -300 here
    test_framework_1.assertEqual((await s1._server.dal.walletDAL.getWallet('SIG(' + tac.pub + ')')).balance, 1100); // <-- +100 here
    test_framework_1.assertEqual((await s1._server.dal.walletDAL.getWallet('SIG(' + toc.pub + ')')).balance, 200); // <-- +200 here
}
async function assertBlock3(s1, cat, tac, toc) {
    test_framework_1.assertEqual((await s1._server.dal.dividendDAL.getUDSources(cat.pub)).length, 0);
    test_framework_1.assertEqual((await s1._server.dal.dividendDAL.getUDSources(tac.pub)).length, 0); // <-- The UD gets consumed
    test_framework_1.assertEqual((await s1._server.dal.dividendDAL.getUDSources(toc.pub)).length, 0); // toc is not a member
    test_framework_1.assertEqual((await s1._server.dal.sindexDAL.getAvailableForPubkey(cat.pub)).length, 2); // <-- Cat receives a new source
    test_framework_1.assertEqual((await s1._server.dal.sindexDAL.getAvailableForPubkey(tac.pub)).length, 0); // <-- Every TX source gets consumed
    test_framework_1.assertEqual((await s1._server.dal.sindexDAL.getAvailableForPubkey(toc.pub)).length, 1);
    test_framework_1.assertEqual((await s1._server.dal.walletDAL.getWallet('SIG(' + cat.pub + ')')).balance, 1800); // <-- +1100 here
    test_framework_1.assertEqual((await s1._server.dal.walletDAL.getWallet('SIG(' + tac.pub + ')')).balance, 0); // <-- -1100 here
    test_framework_1.assertEqual((await s1._server.dal.walletDAL.getWallet('SIG(' + toc.pub + ')')).balance, 200);
}
async function assertBlock4(s1, cat, tac, toc) {
    test_framework_1.assertEqual((await s1._server.dal.dividendDAL.getUDSources(cat.pub)).length, 0);
    test_framework_1.assertEqual((await s1._server.dal.dividendDAL.getUDSources(tac.pub)).length, 0);
    test_framework_1.assertEqual((await s1._server.dal.dividendDAL.getUDSources(toc.pub)).length, 0); // toc is not a member
    test_framework_1.assertEqual((await s1._server.dal.sindexDAL.getAvailableForPubkey(cat.pub)).length, 3); // <-- Cat receives a new source
    test_framework_1.assertEqual((await s1._server.dal.sindexDAL.getAvailableForPubkey(tac.pub)).length, 0);
    test_framework_1.assertEqual((await s1._server.dal.sindexDAL.getAvailableForPubkey(toc.pub)).length, 1); // <-- Consume everything + Create a rest
    test_framework_1.assertEqual((await s1._server.dal.walletDAL.getWallet('SIG(' + cat.pub + ')')).balance, 1900); // <-- +100 here
    test_framework_1.assertEqual((await s1._server.dal.walletDAL.getWallet('SIG(' + tac.pub + ')')).balance, 0);
    test_framework_1.assertEqual((await s1._server.dal.walletDAL.getWallet('SIG(' + toc.pub + ')')).balance, 100); // <-- -100 here
}
//# sourceMappingURL=block-with-transaction-revert.js.map