"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const WS2PConnection_1 = require("../../../app/modules/ws2p/lib/WS2PConnection");
const keyring_1 = require("../../../app/lib/common-libs/crypto/keyring");
const toolbox_1 = require("../tools/toolbox");
const constants_1 = require("../../../app/modules/ws2p/lib/constants");
const unit_tools_1 = require("../../unit-tools");
const logger_1 = require("../../../app/lib/logger");
const websocket_1 = require("../../../app/lib/common-libs/websocket");
const assert = require('assert');
const logger = logger_1.NewLogger();
const gtest = "gtest";
describe('WS2P', () => {
    constants_1.WS2PConstants.CONNEXION_TIMEOUT = 100;
    constants_1.WS2PConstants.REQUEST_TIMEOUT = 100;
    describe("WS2P client connection", function () {
        describe("no auth", () => {
            let wss, portA;
            before(async () => {
                portA = toolbox_1.getNewTestingPort();
                wss = new websocket_1.WebSocketServer({ port: portA });
                wss.on('connection', (ws) => {
                    ws.on('message', (data) => {
                        const obj = JSON.parse(data);
                        if (obj.reqId) {
                            ws.send(JSON.stringify({ resId: obj.reqId, body: { bla: 'aa' } }));
                        }
                    });
                });
            });
            after((done) => {
                wss.close(done);
            });
            it('should be able to create a connection', async () => {
                const ws2p = WS2PConnection_1.WS2PConnection.newConnectionToAddress(1, 'ws://localhost:' + portA, new WS2PMutedHandler(), new WS2PNoLocalAuth(), new WS2PNoRemoteAuth(gtest));
                const res = await ws2p.request({ name: 'head' });
                assert.deepEqual({ bla: 'aa' }, res);
            });
        });
        describe("pubkey auth", () => {
            let wss, clientAskError = "";
            before(async () => {
                const serverKeypair = new keyring_1.Key('DKpQPUL4ckzXYdnDRvCRKAm1gNvSdmAXnTrJZ7LvM5Qo', '64EYRvdPpTfLGGmaX5nijLXRqWXaVz8r1Z1GtaahXwVSJGQRn7tqkxLb288zwSYzELMEG5ZhXSBYSxsTsz1m9y8F');
                let nbAsk = 0;
                wss = new websocket_1.WebSocketServer({ port: 20903 });
                wss.on('connection', (ws) => {
                    ws.on('message', (data) => {
                        const obj = JSON.parse(data);
                        if (obj.reqId) {
                            ws.send(JSON.stringify({ resId: obj.reqId, body: { bla: 'aa' } }));
                        }
                        if (obj.auth) {
                            if (nbAsk == 1 || nbAsk == 3) {
                                const challengeMessage = `WS2P:ACK:gtest:${serverKeypair.pub}:${obj.challenge}`;
                                const sig = serverKeypair.signSync(challengeMessage);
                                if (nbAsk == 1) {
                                    ws.send(JSON.stringify({ auth: 'ACK', pub: serverKeypair.pub, sig: 'hiohoihio' }));
                                }
                                if (nbAsk == 3) {
                                    ws.send(JSON.stringify({ auth: 'ACK', pub: serverKeypair.pub, sig }));
                                }
                            }
                            if (nbAsk == 2) {
                                // We do like if the key was wrong
                                const clientPub = 'GgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd';
                                const challengeMessage = `WS2P:CONNECT:${clientPub}:${obj.challenge}`;
                                if (!keyring_1.verify(challengeMessage, obj.sig, clientPub)) {
                                    clientAskError = 'Wrong signature from client CONNECT';
                                }
                            }
                            nbAsk++;
                        }
                    });
                });
            });
            after((done) => {
                wss.close(done);
            });
            it('should refuse the connection if the server does not answer', async () => {
                const keypair = new keyring_1.Key('HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd', '51w4fEShBk1jCMauWu4mLpmDVfHksKmWcygpxriqCEZizbtERA6de4STKRkQBpxmMUwsKXRjSzuQ8ECwmqN1u2DP');
                const ws2p = WS2PConnection_1.WS2PConnection.newConnectionToAddress(1, 'ws://localhost:20903', new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, keypair, ""), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, keypair), undefined, {
                    connectionTimeout: 100,
                    requestTimeout: 100
                });
                await unit_tools_1.assertThrows(ws2p.request({ name: 'a' }), "WS2P connection timeout");
            });
            it('should refuse the connection if the server answers with a wrong signature', async () => {
                const keypair = new keyring_1.Key('HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd', '51w4fEShBk1jCMauWu4mLpmDVfHksKmWcygpxriqCEZizbtERA6de4STKRkQBpxmMUwsKXRjSzuQ8ECwmqN1u2DP');
                const ws2p = WS2PConnection_1.WS2PConnection.newConnectionToAddress(1, 'ws://localhost:20903', new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, keypair, ""), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, keypair), undefined, {
                    connectionTimeout: 100,
                    requestTimeout: 100
                });
                await unit_tools_1.assertThrows(ws2p.request({ name: 'a' }), "Wrong signature from server ACK");
            });
            it('should refuse the connection if the server refuses our signature', async () => {
                const keypair = new keyring_1.Key('HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd', '51w4fEShBk1jCMauWu4mLpmDVfHksKmWcygpxriqCEZizbtERA6de4STKRkQBpxmMUwsKXRjSzuQ8ECwmqN1u2DP');
                const ws2p = WS2PConnection_1.WS2PConnection.newConnectionToAddress(1, 'ws://localhost:20903', new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, keypair, ""), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, keypair), undefined, {
                    connectionTimeout: 100,
                    requestTimeout: 100
                });
                await unit_tools_1.assertThrows(ws2p.request({ name: 'a' }), "WS2P connection timeout");
                assert.equal('Wrong signature from client CONNECT', clientAskError);
            });
            it('should accept the connection if the server answers with a good signature', async () => {
                const keypair = new keyring_1.Key('HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd', '51w4fEShBk1jCMauWu4mLpmDVfHksKmWcygpxriqCEZizbtERA6de4STKRkQBpxmMUwsKXRjSzuQ8ECwmqN1u2DP');
                const ws2p = WS2PConnection_1.WS2PConnection.newConnectionToAddress(1, 'ws://localhost:20903', new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, keypair, ""), new WS2PNoRemoteAuth(gtest), undefined, {
                    connectionTimeout: 1000,
                    requestTimeout: 1000
                });
                const res = await ws2p.request({ name: 'head' });
                assert.deepEqual({ bla: 'aa' }, res);
            });
        });
    });
    describe("WS2P server connection", function () {
        describe("no auth", () => {
            let wss;
            let s1;
            let s2;
            let portB;
            before(async () => {
                let i = 0;
                portB = toolbox_1.getNewTestingPort();
                wss = new websocket_1.WebSocketServer({ port: portB });
                wss.on('connection', (ws) => {
                    switch (i) {
                        case 0:
                            s1 = WS2PConnection_1.WS2PConnection.newConnectionFromWebSocketServer(ws, new (class TmpHandler {
                                async handlePushMessage(json) {
                                }
                                async answerToRequest(json) {
                                    return { answer: 'world' };
                                }
                            }), new WS2PNoLocalAuth(), new WS2PNoRemoteAuth(gtest));
                            s1.connect().catch(e => logger.error('WS2P: newConnectionFromWebSocketServer connection error'));
                            break;
                        case 1:
                            let j = 0;
                            s2 = WS2PConnection_1.WS2PConnection.newConnectionFromWebSocketServer(ws, new (class TmpHandler {
                                async handlePushMessage(json) {
                                }
                                async answerToRequest(json) {
                                    return { answer: 'this is s2![j = ' + (j++) + ']' };
                                }
                            }), new WS2PNoLocalAuth(), new WS2PNoRemoteAuth(gtest));
                            s2.connect().catch(e => logger.error('WS2P: newConnectionFromWebSocketServer connection error'));
                            break;
                    }
                    i++;
                });
            });
            after((done) => {
                wss.close(done);
            });
            it('should be able to create connections and make several requests', async () => {
                // connection 1
                const c1 = WS2PConnection_1.WS2PConnection.newConnectionToAddress(1, 'ws://localhost:' + portB, new WS2PMutedHandler(), new WS2PNoLocalAuth(), new WS2PNoRemoteAuth(gtest));
                assert.deepEqual({ answer: 'world' }, await c1.request({ name: 'hello!' }));
                assert.deepEqual({ answer: 'world' }, await c1.request({ name: 'hello2!' }));
                assert.equal(s1.nbRequests, 0);
                assert.equal(c1.nbRequests, 2);
                assert.equal(s1.nbResponses, 0);
                assert.equal(c1.nbResponses, 2);
                assert.equal(s1.nbPushsToRemote, 0);
                assert.equal(c1.nbPushsToRemote, 0);
                assert.equal(s1.nbPushsByRemote, 0);
                assert.equal(c1.nbPushsByRemote, 0);
                // connection 2
                const c2 = WS2PConnection_1.WS2PConnection.newConnectionToAddress(1, 'ws://localhost:' + portB, new WS2PMutedHandler(), new WS2PNoLocalAuth(), new WS2PNoRemoteAuth(gtest));
                assert.deepEqual({ answer: 'this is s2![j = 0]' }, await c2.request({ name: 'test?' }));
                assert.deepEqual({ answer: 'this is s2![j = 1]' }, await c2.request({ name: 'test!' }));
                assert.deepEqual({ answer: 'this is s2![j = 2]' }, await c2.request({ name: 'test!!!' }));
                assert.equal(s1.nbRequests, 0);
                assert.equal(c2.nbRequests, 3);
                assert.equal(s1.nbResponses, 0);
                assert.equal(c2.nbResponses, 3);
                assert.equal(s1.nbPushsToRemote, 0);
                assert.equal(c2.nbPushsToRemote, 0);
                assert.equal(s1.nbPushsByRemote, 0);
                assert.equal(c2.nbPushsByRemote, 0);
            });
        });
        describe("pubkey auth", () => {
            let wss;
            let resolveS1;
            let resolveS2;
            let resolveS3;
            let resolveS4;
            let resolveS5;
            let resolveS6;
            let s1p = new Promise(res => resolveS1 = res);
            let s2p = new Promise(res => resolveS2 = res);
            let s3p = new Promise(res => resolveS3 = res);
            let s4p = new Promise(res => resolveS4 = res);
            let s5p = new Promise(res => resolveS5 = res);
            let s6p = new Promise(res => resolveS6 = res);
            before(async () => {
                let i = 1;
                wss = new websocket_1.WebSocketServer({ port: 20903 });
                const serverKeypair = new keyring_1.Key('DKpQPUL4ckzXYdnDRvCRKAm1gNvSdmAXnTrJZ7LvM5Qo', '64EYRvdPpTfLGGmaX5nijLXRqWXaVz8r1Z1GtaahXwVSJGQRn7tqkxLb288zwSYzELMEG5ZhXSBYSxsTsz1m9y8F');
                wss.on('connection', async (ws) => {
                    switch (i) {
                        case 1:
                            resolveS1(WS2PConnection_1.WS2PConnection.newConnectionFromWebSocketServer(ws, new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, serverKeypair, ""), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, serverKeypair), {
                                connectionTimeout: 100,
                                requestTimeout: 100
                            }));
                            (await s1p).connect().catch((e) => logger.error('WS2P: newConnectionFromWebSocketServer connection error'));
                            break;
                        case 2:
                            class WS2PPubkeyNotAnsweringWithOKAuth extends WS2PConnection_1.WS2PPubkeyRemoteAuth {
                                async registerOK(sig) {
                                    return Promise.resolve(true);
                                }
                            }
                            resolveS2(WS2PConnection_1.WS2PConnection.newConnectionFromWebSocketServer(ws, new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, serverKeypair, ""), new WS2PPubkeyNotAnsweringWithOKAuth(gtest, serverKeypair), {
                                connectionTimeout: 100,
                                requestTimeout: 100
                            }));
                            (await s2p).connect().catch((e) => logger.error('WS2P: newConnectionFromWebSocketServer connection error'));
                            break;
                        case 3:
                            resolveS3(WS2PConnection_1.WS2PConnection.newConnectionFromWebSocketServer(ws, new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, serverKeypair, ""), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, serverKeypair), {
                                connectionTimeout: 100,
                                requestTimeout: 100
                            }));
                            (await s3p).connect().catch((e) => logger.error('WS2P: newConnectionFromWebSocketServer connection error'));
                            break;
                        case 4:
                            resolveS4(WS2PConnection_1.WS2PConnection.newConnectionFromWebSocketServer(ws, new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, serverKeypair, ""), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, serverKeypair), {
                                connectionTimeout: 100,
                                requestTimeout: 100
                            }));
                            (await s4p).connect().catch((e) => logger.error('WS2P: newConnectionFromWebSocketServer connection error'));
                            break;
                        case 5:
                            resolveS5(WS2PConnection_1.WS2PConnection.newConnectionFromWebSocketServer(ws, new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, serverKeypair, ""), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, serverKeypair)));
                            (await s5p).connect().catch((e) => logger.error('WS2P: newConnectionFromWebSocketServer connection error'));
                            break;
                        case 6:
                            resolveS6(WS2PConnection_1.WS2PConnection.newConnectionFromWebSocketServer(ws, new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, serverKeypair, ""), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, serverKeypair), {
                                connectionTimeout: 100,
                                requestTimeout: 100
                            }));
                            (await s6p).connect().catch((e) => logger.error('WS2P: newConnectionFromWebSocketServer connection error'));
                            break;
                    }
                    i++;
                });
            });
            after((done) => {
                wss.close(done);
            });
            it('should refuse the connection if the client does not send ACK', async () => {
                class WS2PPubkeyNotAnsweringWithACKAuth extends WS2PConnection_1.WS2PPubkeyRemoteAuth {
                    async sendACK(ws) {
                        return Promise.resolve();
                    }
                }
                const keypair = new keyring_1.Key('HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd', '51w4fEShBk1jCMauWu4mLpmDVfHksKmWcygpxriqCEZizbtERA6de4STKRkQBpxmMUwsKXRjSzuQ8ECwmqN1u2DP');
                const c1 = WS2PConnection_1.WS2PConnection.newConnectionToAddress(1, 'ws://localhost:20903', new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, keypair, ""), new WS2PPubkeyNotAnsweringWithACKAuth(gtest, keypair));
                c1.connect().catch((e) => logger.error('WS2P: connection error'));
                const s1 = await s1p;
                await unit_tools_1.assertThrows(s1.request({ name: 'something' }), "WS2P connection timeout");
            });
            it('should refuse the connection if the client not confirm with OK', async () => {
                const keypair = new keyring_1.Key('HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd', '51w4fEShBk1jCMauWu4mLpmDVfHksKmWcygpxriqCEZizbtERA6de4STKRkQBpxmMUwsKXRjSzuQ8ECwmqN1u2DP');
                WS2PConnection_1.WS2PConnection.newConnectionToAddress(1, 'ws://localhost:20903', new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, keypair, ""), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, keypair));
                const s2 = await s2p;
                await unit_tools_1.assertThrows(s2.request({ name: 'something' }), "WS2P connection timeout");
            });
            it('should refuse the connection if the client answers with a wrong signature', async () => {
                class WS2PPubkeyAnsweringWithWrongSigForACK extends WS2PConnection_1.WS2PPubkeyRemoteAuth {
                    async sendACK(ws) {
                        const challengeMessage = `WS2P:WRONG:${this.pair.pub}:${this.challenge}`;
                        const sig = this.pair.signSync(challengeMessage);
                        await ws.send(JSON.stringify({
                            auth: 'ACK',
                            pub: this.pair.pub,
                            sig
                        }));
                    }
                }
                const keypair = new keyring_1.Key('HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd', '51w4fEShBk1jCMauWu4mLpmDVfHksKmWcygpxriqCEZizbtERA6de4STKRkQBpxmMUwsKXRjSzuQ8ECwmqN1u2DP');
                const c3 = WS2PConnection_1.WS2PConnection.newConnectionToAddress(1, 'ws://localhost:20903', new WS2PMutedHandler(), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, keypair, ""), new WS2PPubkeyAnsweringWithWrongSigForACK(gtest, keypair));
                c3.connect().catch((e) => logger.error('WS2P: connection error'));
                const s3 = await s3p;
                await unit_tools_1.assertThrows(s3.request({ name: 'something' }), "Wrong signature from server ACK");
            });
            it('should refuse the connection if the client refuses our signature', async () => {
                class WS2PPubkeyRefusingACKSignature extends WS2PConnection_1.WS2PPubkeyLocalAuth {
                    async registerACK(sig, pub) {
                        const challengeMessage = `WS2P:BLABLA:${pub}:${this.challenge}`;
                        this.authenticated = keyring_1.verify(challengeMessage, sig, pub);
                        if (!this.authenticated) {
                            this.serverAuthReject("Wrong signature from server ACK");
                        }
                        else {
                            this.serverAuthResolve();
                        }
                        return this.authenticated;
                    }
                }
                const keypair = new keyring_1.Key('HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd', '51w4fEShBk1jCMauWu4mLpmDVfHksKmWcygpxriqCEZizbtERA6de4STKRkQBpxmMUwsKXRjSzuQ8ECwmqN1u2DP');
                const c4 = WS2PConnection_1.WS2PConnection.newConnectionToAddress(1, 'ws://localhost:20903', new WS2PMutedHandler(), new WS2PPubkeyRefusingACKSignature(gtest, keypair, ""), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, keypair));
                const s4 = await s4p;
                await unit_tools_1.assertThrows(c4.connect(), "Wrong signature from server ACK");
            });
            it('should accept the connection if everything is OK on both side', async () => {
                const keypair = new keyring_1.Key('HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd', '51w4fEShBk1jCMauWu4mLpmDVfHksKmWcygpxriqCEZizbtERA6de4STKRkQBpxmMUwsKXRjSzuQ8ECwmqN1u2DP');
                const c5 = WS2PConnection_1.WS2PConnection.newConnectionToAddress(1, 'ws://localhost:20903', new (class TmpHandler {
                    async handlePushMessage(json) {
                    }
                    async answerToRequest(json) {
                        return { answer: 'success!' };
                    }
                }), new WS2PConnection_1.WS2PPubkeyLocalAuth(gtest, keypair, ""), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, keypair));
                await c5.connect().catch((e) => logger.error('WS2P: connection error'));
                const s5 = await s5p;
                assert.deepEqual({ answer: 'success!' }, await s5.request({ name: 'connection?' }));
            });
            it('should refuse the connection if the client does not send OK', async () => {
                class WS2PPubkeyNotAnsweringWithOKAuth extends WS2PConnection_1.WS2PPubkeyLocalAuth {
                    async sendOK(ws) {
                    }
                }
                const keypair = new keyring_1.Key('HgTTJLAQ5sqfknMq7yLPZbehtuLSsKj9CxWN7k8QvYJd', '51w4fEShBk1jCMauWu4mLpmDVfHksKmWcygpxriqCEZizbtERA6de4STKRkQBpxmMUwsKXRjSzuQ8ECwmqN1u2DP');
                const c6 = WS2PConnection_1.WS2PConnection.newConnectionToAddress(1, 'ws://localhost:20903', new WS2PMutedHandler(), new WS2PPubkeyNotAnsweringWithOKAuth(gtest, keypair, ""), new WS2PConnection_1.WS2PPubkeyRemoteAuth(gtest, keypair));
                c6.connect().catch((e) => logger.error('WS2P: connection error'));
                const s6 = await s6p;
                await unit_tools_1.assertThrows(s6.request({ name: 'something' }), "WS2P connection timeout");
            });
        });
    });
});
/**************
 * TEST CLASSES
 *************/
class WS2PNoLocalAuth {
    async sendCONNECT(ws) {
    }
    async registerACK(sig, pub) {
        return true;
    }
    isRemoteAuthenticated() {
        return true;
    }
    async isAuthorizedPubkey(pub) {
        return true;
    }
    async sendOK(ws) {
    }
    async authenticationIsDone() {
    }
}
class WS2PNoRemoteAuth {
    constructor(currency) {
        this.givenCurrency = Promise.resolve(currency);
    }
    getVersion() {
        return 1;
    }
    getPubkey() {
        return "";
    }
    async sendACK(ws) {
    }
    async registerCONNECT(type, version, challenge, sig, pub, currency, ws2pId) {
        this.givenCurrency = Promise.resolve(currency);
        return true;
    }
    async registerOK(sig) {
        return true;
    }
    isAuthenticatedByRemote() {
        return true;
    }
    async isAuthorizedPubkey(pub) {
        return true;
    }
    async authenticationIsDone() {
    }
    isSync() {
        return false;
    }
}
class WS2PMutedHandler {
    async handlePushMessage(json) {
    }
    async answerToRequest(json) {
        throw "Does not answer";
    }
}
//# sourceMappingURL=ws2p_connection.js.map